/*
 * nrf24.c
 *
 *  Created on: 27.02.2013
 *      Author: AndreR
 */

#include "nrf24.h"

#include "errors.h"

#include "util/console.h"
#include "util/critical.h"
#include "FreeRTOS.h"
#include "task.h"

#include <string.h>

// commands
#define R_REGISTER			0			// | register address
#define W_REGISTER			0b00100000	// | register address
#define R_RX_PAYLOAD		0b01100001
#define W_TX_PAYLOAD		0b10100000
#define W_TX_PAYLOAD_NO_ACK	0b10110000
#define FLUSH_TX			0b11100001
#define FLUSH_RX			0b11100010
#define REUSE_TX_PL			0b11100011
#define R_RX_PL_WID			0b01100000
#define W_ACK_PAYLOAD		0b10101000	// | PPP = pipe
#define W_TX_PAYLOAD_NO_ACK	0b10110000
#define NOP					0b11111111

// register addresses
#define REG_CONFIG 		0x00
#define REG_EN_AA 		0x01
#define REG_EN_RXADDR	0x02
#define REG_SETUP_AW	0x03
#define REG_SETUP_RETR	0x04
#define REG_RF_CH		0x05
#define REG_RF_SETUP	0x06
#define REG_STATUS		0x07
#define REG_OBSERVE_TX	0x08
#define REG_RPD			0x09
#define REG_ADDR_P0		0x0A
#define REG_ADDR_P1		0x0B
// ...
#define REG_TX_ADDR		0x10
#define REG_RX_PW_P0	0x11
#define REG_RX_PW_P1	0x12
// ...
#define REG_FIFO_STATUS	0x17
#define REG_DYNPD		0x1C
#define REG_FEATURE		0x1D

#define REG_STATUS_RX_DR	0x40
#define REG_STATUS_TX_DS	0x20
#define REG_STATUS_MAX_RT	0x10

#define W_CONFIG (W_REGISTER | REG_CONFIG)
#define R_CONFIG (R_REGISTER | REG_CONFIG)
#define W_EN_AA (W_REGISTER | REG_EN_AA)
#define R_EN_AA (R_REGISTER | REG_EN_AA)

// macros
#ifdef STM32F10X_CL
#define CE_LOW(nrf)		(nrf->pCEPort->BRR = nrf->cePin)
#define CE_HIGH(nrf)	(nrf->pCEPort->BSRR = nrf->cePin)
#else
#define CE_LOW(nrf)		(nrf->pCEPort->BSRRH = nrf->cePin)
#define CE_HIGH(nrf)	(nrf->pCEPort->BSRRL = nrf->cePin)
#endif

// events
#define EVENT_DATA_READY	0x01
#define EVENT_DATA_SEND		0x02
#define EVENT_DATA_TXERR	0x04
#define EVENT_IRQ			0x08
#define EVENT_REG_READ		0x10

// states
#define STATE_IDLE			0x00
#define STATE_WAIT_STATUS	0x01	// wait for status
#define STATE_WAIT_TX		0x02
#define STATE_EXIT			0xFF

// packet flags
#define FLAG_SIZE 1
#define FLAG_END 0x66

// TX, ACK, NO-ACK Payload
#define TYPE_SIZE 1


#define MAX_PAYLOAD_SIZE (NRF24_MAX_LENGTH-FLAG_SIZE)

static uint8_t addresses[] = {0xF1, 0xE2, 0xD3, 0xC4, 0xB5, 0xA6, 0x97, 0x88, 0x79, 0x6A, 0x5B, 0x4C, 0x3D, 0x2E, 0x1F};

static __inline__ void enable(NRF24* pNRF24)
{
	pNRF24->enabled = 1;
	CE_HIGH(pNRF24);
}

static __inline__ void disable(NRF24* pNRF24)
{
	pNRF24->enabled = 0;
	CE_LOW(pNRF24);
}

static int16_t readRxLength(NRF24* pNRF24, uint8_t* pValue)
{
	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.cmd = R_RX_PL_WID;
	header.length = 2;

	int16_t result = SPIPutPacket(pNRF24->pSPI, &header, 1, &header.cmd);
	if(result)
		return result;

	uint8_t content;
	if(xQueueReceive(pNRF24->hRegisterQueue, &content, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	*pValue = content;

	return 0;
}

static int16_t writeSpecialCmd(NRF24* pNRF24, uint8_t cmd)
{
	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.cmd = cmd;
	header.length = 1;

	int16_t result = SPIPutPacket(pNRF24->pSPI, &header, 1, &header.cmd);
	if(result)
		return result;

	return 0;
}

static int16_t writeConfig(NRF24* pNRF24, uint8_t powerUp, uint8_t rx)
{
	return NRF24WriteRegister(pNRF24, REG_CONFIG, (powerUp << 1) | rx | 0x08);
}

static int16_t setId(NRF24* pNRF24, uint8_t id)
{
	int16_t result = 0;

	if(id > NRF24_MAX_BOT_ID)
		return ERROR_NRF24_BAD_ADDRESS;

	uint8_t address = addresses[id];

	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.length = 6;
	header.cmd = W_REGISTER | REG_ADDR_P0;
	uint8_t data[6];

	data[0] = header.cmd;

	// address is repeated 5 times
	for(uint8_t i = 0; i < 5; i++)
		data[i+1] = address;

	// set RX address
	result = SPIPutPacket(pNRF24->pSPI, &header, 6, data);
	if(result)
		return result;

	header.cmd = W_REGISTER | REG_TX_ADDR;
	data[0] = header.cmd;

	// set TX address
	return SPIPutPacket(pNRF24->pSPI, &header, 6, data);
}

static int16_t waitForIRQ(NRF24* pNRF24)
{
	if(xSemaphoreTake(pNRF24->hIRQ, 20/portTICK_RATE_MS) == pdFALSE)
		return 1;

	return 0;
}

static int16_t fetchRxData(NRF24* pNRF24)
{
	int16_t result = 0;
	uint8_t status;
	uint8_t rxLen;

	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.cmd = R_RX_PAYLOAD;
	header.length = 1;

	while(1)
	{
		result = NRF24ReadRegister(pNRF24, REG_FIFO_STATUS, &status);	// this will block until data fetch and status read is done
		if(result)
			break;

		if((status & 0x01) == 0x01)	// RX empty?
			break;

		result = readRxLength(pNRF24, &rxLen);
		if(result)
			break;

//		ConsolePrint("RX Len: %u, result 0x%X\r\n", rxLen, result);

		if(rxLen > 32)
		{
			writeSpecialCmd(pNRF24, FLUSH_RX);
		}
		else if(rxLen > 0)
		{
			// data to fetch
			header.length = rxLen+1;

			result = SPIPutPacket(pNRF24->pSPI, &header, 1, &header.cmd);
			if(result)
			{
				ConsolePrint("WARN: Reading RX data failed\r\n");
				break;
			}
		}
	}

	NRF24WriteRegister(pNRF24, REG_STATUS, 0x40);	// clear RX_DR status flag

	return result;
}

static int16_t pushTxData(NRF24* pNRF24)
{
	int16_t result = 0;
	uint8_t status;
	uint16_t length;

	SPIHeader header;
	header.slaveId = pNRF24->slaveId;

	if(FifoPacketSyncedNumPackets(&pNRF24->txFifo) == 0)
		return 0;

	while(1)
	{
		result = NRF24ReadRegister(pNRF24, REG_FIFO_STATUS, &status);	// this will block until data fetch and status read is done
		if(result)
			break;

		if((status & 0x20) == 0x20)	// Tx fifo full?
			break;

		// there is space in the TX fifo and we want to transmit something, that is good :)
		result = FifoPacketSyncedPop(&pNRF24->txFifo, &length, pNRF24->procBuf, NRF24_MAX_LENGTH+TYPE_SIZE);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&pNRF24->txFifo);
		}
		if(result)
			return result;
		if(length == 0)
			continue;

		header.cmd = pNRF24->procBuf[0];

		header.length = length;

		result = SPIPutPacket(pNRF24->pSPI, &header, length, pNRF24->procBuf);
		if(result)
			return result;

		break;
	}

	return result;
}

static int16_t send(NRF24* pNRF24, uint16_t dataLength, uint8_t* pData, uint8_t dataType)
{
	int16_t result = 0;
	uint16_t numPackets = (dataLength+MAX_PAYLOAD_SIZE-1)/MAX_PAYLOAD_SIZE;
	uint16_t totalSize = numPackets*(FIFO_PACKET_SYNCED_HEADER_SIZE+FLAG_SIZE+TYPE_SIZE) + dataLength;

	TASK_MUTEX_RECURSIVE(pNRF24->hTxMutex)
	{
		if(pNRF24->txFifo.fifo.EmptyBytes < totalSize)
			return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

		uint16_t bytesSent = 0;
		uint8_t flag = 0;

		while(bytesSent < dataLength)
		{
			result = FifoPacketSyncedBegin(&pNRF24->txFifo);
			if(result)
			{
				return result;
			}

			uint16_t length = MAX_PAYLOAD_SIZE;
			if(length > (dataLength-bytesSent))
				length = dataLength-bytesSent;

			if(bytesSent+length == dataLength)	// last packet?
				flag = FLAG_END;
			else
				flag = 0;

			result = FifoPacketSyncedPush(&pNRF24->txFifo, TYPE_SIZE, &dataType);
			if(result)
			{
				FifoPacketSyncedCancel(&pNRF24->txFifo);
				return result;
			}

			result = FifoPacketSyncedPush(&pNRF24->txFifo, FLAG_SIZE, &flag);	// push flag first in every packet
			if(result)
			{
				FifoPacketSyncedCancel(&pNRF24->txFifo);
				return result;
			}

			result = FifoPacketSyncedPush(&pNRF24->txFifo, length, pData+bytesSent);	// push up to MAX_PAYLOAD_SIZE bytes of data
			if(result)
			{
				FifoPacketSyncedCancel(&pNRF24->txFifo);
				return result;
			}

			FifoPacketSyncedComplete(&pNRF24->txFifo);	// packet complete

			bytesSent += length;
		}
	}

	return 0;
}

static void spiCallback(SPIHeader* pHeader, uint8_t* pData, void* pUserData)
{
	NRF24* pNRF24 = (NRF24*)pUserData;

	pNRF24->lastStatus = pData[0];

//	ConsolePrint("NRF SPI %hhu. Cmd: 0x%02hhX, Status 0x%02hhX\r\n", pHeader->slaveId, pHeader->cmd, pNRF24->lastStatus);

	if((pHeader->cmd & 0xE0) == R_REGISTER)
	{
		xQueueSend(pNRF24->hRegisterQueue, &pData[1], 0);
	}

	if(pHeader->cmd == R_RX_PL_WID)
	{
		xQueueSend(pNRF24->hRegisterQueue, &pData[1], 0);
	}

	if(pHeader->cmd == R_RX_PAYLOAD)
	{
		uint16_t payloadLength = pHeader->length-1-FLAG_SIZE;
		uint8_t* pPayload = pData+1+FLAG_SIZE;
		uint8_t flag = pData[1];	// first byte is flag

		pNRF24->stats.rxBytesRecv += pHeader->length-1;
		pNRF24->stats.rxPacketsRecv++;

		if(pNRF24->rxCancelled)
		{
			pNRF24->stats.rxBytesLost += pHeader->length-1;
		}
		else
		{
			int16_t result = FifoPacketSyncedPush(&pNRF24->rxFifo, payloadLength, pPayload);
			if(result)
			{
				pNRF24->stats.rxBytesLost += pNRF24->rxFifo.currentLength;
				pNRF24->stats.rxPacketsLost++;

				FifoPacketSyncedCancel(&pNRF24->rxFifo);
				pNRF24->rxCancelled = 1;
			}
		}

		if(flag == FLAG_END)
		{
			// this is the last chunk of a packet
			if(!pNRF24->rxCancelled)
			{
				FifoPacketSyncedComplete(&pNRF24->rxFifo);

				xSemaphoreGive(pNRF24->hIRQ);
			}

			pNRF24->rxCancelled = 0;
		}
	}

	if((pHeader->cmd & 0xE0) == W_REGISTER)
	{
		xSemaphoreGive(pNRF24->hWriteReg);
	}

	if(pHeader->cmd == W_ACK_PAYLOAD || pHeader->cmd == W_TX_PAYLOAD || pHeader->cmd == W_TX_PAYLOAD_NO_ACK)
	{
		pNRF24->stats.txBytesSent += pHeader->length-1;
		pNRF24->stats.txPacketsSent++;
	}
}

int16_t NRF24Init(NRF24* pNRF24, NRF24Data* pData)
{
	pNRF24->rxCallback = 0;
	pNRF24->cePin = pData->cePin;
	pNRF24->pCEPort = pData->pCEPort;
	pNRF24->pSPI = pData->pSPI;
	pNRF24->enabled = 0;
	pNRF24->rx = pData->rx;
	pNRF24->id = pData->id;
	pNRF24->channel = pData->channel;
	pNRF24->maxRt = pData->maxRt;
	pNRF24->ackDelay = ((pData->ackDelayUs/250)-1) << 4;
	pNRF24->speed = pData->speed;
	pNRF24->rxCancelled = 0;
	pNRF24->hRegisterQueue = xQueueCreate(1, sizeof(uint8_t));
	pNRF24->hTxMutex = xSemaphoreCreateRecursiveMutex();
	pNRF24->hRxMutex = xSemaphoreCreateRecursiveMutex();
	vSemaphoreCreateBinary(pNRF24->hIRQ);
	vSemaphoreCreateBinary(pNRF24->hWriteReg);

	memset(&pNRF24->stats, 0, sizeof(NRF24Stats));

	FifoPacketSyncedInit(&pNRF24->rxFifo, pData->pDataRx, pData->rxSize, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, pNRF24->hRxMutex);
	FifoPacketSyncedInit(&pNRF24->txFifo, pData->pDataTx, pData->txSize, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, pNRF24->hTxMutex);

	// init CE pin
	GPIO_InitTypeDef gpioInit;
#ifdef STM32F10X_CL
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
#else
	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
#endif
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = pData->cePin;
	GPIO_Init(pData->pCEPort, &gpioInit);

	CE_LOW(pNRF24);

	// init SPI slave
	SPISlave spiSlave;
	spiSlave.csPin = pData->csPin;
	spiSlave.pCSPort = pData->pCSPort;
	spiSlave.cpha = 0;
	spiSlave.cpol = 0;
	spiSlave.prescaler = SPI_BaudRatePrescaler_16;
	spiSlave.callback = &spiCallback;
	spiSlave.pUserData = pNRF24;

	int16_t result = SPIAddSlave(pData->pSPI, spiSlave, &pNRF24->slaveId);
	if(result)
		return result;

	// first write is lost on STM32F4, don't know why
	result = NRF24WriteRegister(pNRF24, REG_STATUS, 0x70);
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_FEATURE, 0x07); // enable dynamic packet payload and ack payload and no-ack
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_EN_RXADDR, 0x01); // enable data pipe 0
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_DYNPD, 0x01);	// enable dynamic payload length on pipe 0
	if(result)
		return result;

	result = setId(pNRF24, pNRF24->id);
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_RF_SETUP, 0b00000110 | pNRF24->speed);	// speed, 0dBm output power
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_RF_CH, pNRF24->channel);
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_SETUP_RETR, pNRF24->ackDelay | pNRF24->maxRt);	// x ACK wait, y auto-retransmit
	if(result)
		return result;

	result = writeSpecialCmd(pNRF24, FLUSH_RX);
	if(result)
		return result;

	result = writeSpecialCmd(pNRF24, FLUSH_TX);
	if(result)
		return result;

	result = NRF24WriteRegister(pNRF24, REG_STATUS, 0x70);
	if(result)
		return result;

	// finally power up
	writeConfig(pNRF24, 1, pNRF24->rx);

	if(pNRF24->rx)
		enable(pNRF24);

	return 0;
}

int16_t NRF24CheckChannel(NRF24* pNRF24, uint8_t* pRPD)
{
	int16_t result;
	uint8_t wasEnabled = pNRF24->enabled;

	if(pNRF24->enabled)
		disable(pNRF24);

	xSemaphoreTake(pNRF24->hWriteReg, 0);	// drain semaphore

	result = writeConfig(pNRF24, 0, pNRF24->rx);	// power down
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	vTaskDelay(2/portTICK_RATE_MS);

	result = writeConfig(pNRF24, 1, 1);	// power up as rx
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	uint8_t rpd;
	*pRPD = 0;

	for(uint8_t i = 0; i < 100; i++)
	{
		enable(pNRF24);

		vTaskDelay(2/portTICK_RATE_MS);

		disable(pNRF24);

		result = NRF24ReadRegister(pNRF24, REG_RPD, &rpd);
		if(result)
			return result;

		if(rpd & 0x01)
			(*pRPD)++;
	}

	result = writeConfig(pNRF24, 0, pNRF24->rx);	// power down
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	// power up
	result = writeConfig(pNRF24, 1, pNRF24->rx);
	if(result)
		return result;

	// power up delay
	vTaskDelay(20/portTICK_RATE_MS);

	if(wasEnabled && pNRF24->rx)
		enable(pNRF24);

	return 0;
}

int16_t NRF24ChangeId(NRF24* pNRF24, uint8_t id)
{
	int16_t result = 0;

	result = setId(pNRF24, id);
	if(result)
		return result;

	pNRF24->id = id;

	return 0;
}

int16_t NRF24ChangeChannel(NRF24* pNRF24, uint8_t channel)
{
	int16_t result = 0;

	if(channel > 127)
	{
		return ERROR_NRF24_INVALID_CHANNEL;
	}

	pNRF24->channel = channel;

	// set channel
	result = NRF24WriteRegister(pNRF24, REG_RF_CH, channel);
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	return result;
}

int16_t NRF24ChangeSpeed(NRF24* pNRF24, uint8_t speed)
{
	int16_t result = 0;

	pNRF24->speed = speed;

	// set speed
	result = NRF24WriteRegister(pNRF24, REG_RF_SETUP, 0b00000110 | speed);
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	return result;
}

int16_t NRF24ChangeMaxRt(NRF24* pNRF24, uint8_t maxRt)
{
	int16_t result = 0;

	pNRF24->maxRt = maxRt;

	// set max rt
	result = NRF24WriteRegister(pNRF24, REG_SETUP_RETR, 0b00110000 | maxRt);
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	return result;
}

int16_t NRF24ChangeAckDelay(NRF24* pNRF24, uint16_t delayUs)
{
	int16_t result = 0;

	if(delayUs%250 != 0)
		return 1;

	pNRF24->ackDelay = ((delayUs/250)-1) << 4;

	// set max rt and ack delay
	result = NRF24WriteRegister(pNRF24, REG_SETUP_RETR, pNRF24->ackDelay | pNRF24->maxRt);
	if(result)
		return result;

	if(xSemaphoreTake(pNRF24->hWriteReg, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	return result;
}


void NRF24SetRxCallback(NRF24* pNRF24, NRF24RxCallback cb)
{
	pNRF24->rxCallback = cb;
}

// IRQL: wifi handler
void NRF24IRQ(NRF24* pNRF24)
{
	if(!pNRF24->enabled)
		return;

	int32_t yield = pdFALSE;
	xSemaphoreGiveFromISR(pNRF24->hIRQ, &yield);
	if(yield)
		portYIELD();
}

int16_t NRF24ReadRegister(NRF24* pNRF24, uint8_t reg, uint8_t* pValue)
{
	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.cmd = R_REGISTER | reg;
	header.length = 2;

	xQueueReset(pNRF24->hRegisterQueue);

	int16_t result = SPIPutPacket(pNRF24->pSPI, &header, 1, &header.cmd);
	if(result)
		return result;

	uint8_t content;
	if(xQueueReceive(pNRF24->hRegisterQueue, &content, 1000/portTICK_RATE_MS) == pdFALSE)
		return ERROR_NRF24_TIMEOUT;

	*pValue = content;

	return 0;
}

int16_t NRF24WriteRegister(NRF24* pNRF24, uint8_t reg, uint8_t value)
{
	SPIHeader header;
	header.slaveId = pNRF24->slaveId;
	header.length = 2;
	header.cmd = W_REGISTER | reg;
	uint8_t data[2];

	data[0] = header.cmd;
	data[1] = value;

	int16_t result = SPIPutPacket(pNRF24->pSPI, &header, 2, data);
	if(result)
		return result;

	return 0;
}

void NRF24RxTask(void* params)
{
	if(params == 0)
	{
		ConsolePrint("nRF24 Rx Task params == 0!\r\n");
		vTaskSuspend(NULL);
	}

	NRF24* pNRF24 = (NRF24*)params;

	if(!pNRF24->rx)
	{
		ConsolePrint("nRF24 Rx Task started for TX module!\r\n");
		vTaskSuspend(NULL);
	}

	while(1)
	{
		waitForIRQ(pNRF24);

		NRF24Process(pNRF24);

		if(pNRF24->rxCallback && FifoPacketSyncedNumPackets(&pNRF24->rxFifo) > 0)
			(*pNRF24->rxCallback)(pNRF24);
	}
}

int16_t NRF24Process(NRF24* pNRF24)
{
	int16_t result;

	if(pNRF24->rx)
	{
		// this module operates as receiver

		result = fetchRxData(pNRF24);
		if(result)
			return result;

		result = pushTxData(pNRF24);
		if(result)
			return result;

		if(pNRF24->lastStatus & REG_STATUS_TX_DS)
			NRF24WriteRegister(pNRF24, REG_STATUS, REG_STATUS_TX_DS);
	}
	else
	{
		// this module operates as transmitter
		while(1)
		{
			if(FifoPacketSyncedNumPackets(&pNRF24->txFifo) == 0)
			{
				// Nothing to transmit, send a dummy packet to check if there is ACK data

				uint8_t dummy[1] = {0};
				NRF24Send(pNRF24, 1, dummy);
			}

			uint16_t length = 0;
			FifoPacketSyncedGetNextPacketDataLength(&pNRF24->txFifo, &length);

			xSemaphoreTake(pNRF24->hIRQ, 0);	// clear semaphore

			result = pushTxData(pNRF24);
			if(result)
				return result;

			enable(pNRF24);

			if(waitForIRQ(pNRF24))
			{
				disable(pNRF24);
				return ERROR_NRF24_IRQ_MALFUNCTION;
			}

			disable(pNRF24);

			uint8_t observeTx;

			result = NRF24ReadRegister(pNRF24, REG_OBSERVE_TX, &observeTx);
			if(result)
				ConsolePrint("WARN: OBSERVE_TX could not be read\r\n");

			if((pNRF24->lastStatus & 0x70) ==  0 && (GPIOD->IDR & GPIO_Pin_2) == 0)
				ConsolePrint("No RxTx flags set! (0x%hhx)\r\n", pNRF24->lastStatus);

			if(pNRF24->lastStatus & REG_STATUS_MAX_RT)
			{
//				ConsolePrint("WARN: Sending data failed\r\n");

				writeSpecialCmd(pNRF24, FLUSH_TX);	// flush TX queue

				NRF24WriteRegister(pNRF24, REG_STATUS, REG_STATUS_MAX_RT);	// clear MAX_RT flag

				pNRF24->stats.txBytesSent -= length;
				pNRF24->stats.txPacketsSent--;
				pNRF24->stats.txBytesLost += length;
				pNRF24->stats.txPacketsLost++;

				return ERROR_NRF24_NO_LINK;
			}

			if(pNRF24->lastStatus & REG_STATUS_TX_DS)
			{
				// data transmitted and acknowledged

				NRF24WriteRegister(pNRF24, REG_STATUS, REG_STATUS_TX_DS);	// clear flags

				pNRF24->stats.retransmissions += ((observeTx & 0x0F)) * length;
			}


			if(pNRF24->lastStatus & REG_STATUS_RX_DR)
			{
				// data in ACK payload, read it
				result = fetchRxData(pNRF24);
				if(result)
					return result;
			}
			else if(FifoPacketSyncedNumPackets(&pNRF24->txFifo) == 0)
			{
				// nothing more to transmit and no more ACK data => quit
				break;
			}
		}
	}

	return 0;
}

int16_t NRF24Send(NRF24* pNRF24, uint16_t dataLength, uint8_t* pData)
{
	if(pNRF24->rx)
		return send(pNRF24, dataLength, pData, W_ACK_PAYLOAD);
	else
		return send(pNRF24, dataLength, pData, W_TX_PAYLOAD);
}

int16_t NRF24SendNoAck(NRF24* pNRF24, uint16_t dataLength, uint8_t* pData)
{
	return send(pNRF24, dataLength, pData, W_TX_PAYLOAD_NO_ACK);
}

