/*
 * nrf24.h
 *
 *  Created on: 27.02.2013
 *      Author: AndreR
 */

#ifndef NRF24_H_
#define NRF24_H_

#ifdef STM32F10X_CL
#include "stm32f10x.h"
#include "stm32f10x_gpio.h"
#else
#include "stm32f4xx.h"
#include "stm32f4xx_gpio.h"
#endif

#include "util/spi.h"
#include "util/fifo_packet_synced.h"

#include "FreeRTOS.h"
#include "queue.h"
#include "semphr.h"

#define NRF24_MAX_LENGTH 32
#define NRF24_MAX_BOT_ID 14

#define NRF24_SPEED_250K	0x20
#define NRF24_SPEED_1M		0x00
#define NRF24_SPEED_2M		0x08

struct _NRF24;

typedef void(*NRF24RxCallback)(struct _NRF24*);

typedef struct _NRF24Stats
{
	uint32_t txBytesSent;	// = bytes loaded to nRF24 module
	uint32_t txBytesLost;
	uint32_t rxBytesRecv;
	uint32_t rxBytesLost;

	uint32_t txPacketsSent;
	uint32_t txPacketsLost;
	uint32_t rxPacketsRecv;
	uint32_t rxPacketsLost;

	uint32_t retransmissions;
} NRF24Stats;

typedef struct _NRF24
{
	GPIO_TypeDef* pCEPort; // one of GPIOx
	uint16_t cePin; // one of GPIO_Pin_x
	uint8_t slaveId;
	SPI* pSPI;
	uint8_t enabled;
	uint8_t rx;
	uint8_t channel;
	uint8_t maxRt;
	uint8_t speed;
	uint8_t rxCancelled;
	uint8_t id;
	uint8_t ackDelay;

	volatile uint8_t lastStatus;			// status response of any command

	FifoPacketSynced rxFifo;
	FifoPacketSynced txFifo;

	xQueueHandle hRegisterQueue;
	xSemaphoreHandle hTxMutex;
	xSemaphoreHandle hRxMutex;
	xSemaphoreHandle hIRQ;
	xSemaphoreHandle hWriteReg;

	uint8_t procBuf[NRF24_MAX_LENGTH+1];

	NRF24RxCallback rxCallback;

	NRF24Stats stats;
} NRF24;

typedef struct _NRF24Data
{
	// Fifos
	uint8_t* pDataTx;
	uint16_t txSize;
	uint8_t* pDataRx;
	uint16_t rxSize;

	SPI* pSPI;
	uint8_t rx;
	GPIO_TypeDef* pCEPort;
	uint16_t cePin;
	GPIO_TypeDef* pCSPort;
	uint16_t csPin;

	// RF
	uint8_t id;
	uint8_t channel;
	uint8_t maxRt;
	uint8_t speed;
	uint16_t ackDelayUs;
} NRF24Data;

/**
 * Channel and address are initialized to defaults.
 * Use ChangeAddress and ChangeChannel after init.
 */
int16_t NRF24Init(NRF24* pNRF24, NRF24Data* pData);
void NRF24SetRxCallback(NRF24* pNRF24, NRF24RxCallback cb);
void NRF24IRQ(NRF24* pNRF24);

/**
 * Process RX or TX once.
 * TX: Sends all data from tx queue and gathers all rx payloads.
 * RX: Reads all data from rx module queue and puts one tx payload to module if available
 */
int16_t NRF24Process(NRF24* pNRF24);

/**
 * Put data to tx queue.
 * Data is not transmitted immediately. Call NRF24Process to send data.
 */
int16_t NRF24Send(NRF24* pNRF24, uint16_t dataLength, uint8_t* pData);

int16_t NRF24SendNoAck(NRF24* pNRF24, uint16_t dataLength, uint8_t* pData);

/**
 * Helper task for rx modules. Calls callback if new data arrived.
 */
void NRF24RxTask(void* params);

int16_t NRF24ChangeId(NRF24* pNRF24, uint8_t id);
int16_t NRF24ChangeChannel(NRF24* pNRF24, uint8_t channel);
int16_t NRF24ChangeSpeed(NRF24* pNRF24, uint8_t speed);
int16_t NRF24ChangeMaxRt(NRF24* pNRF24, uint8_t maxRt);
int16_t NRF24ChangeAckDelay(NRF24* pNRF24, uint16_t delayUs);
int16_t NRF24CheckChannel(NRF24* pNRF24, uint8_t* pRPD);

int16_t NRF24ReadRegister(NRF24* pNRF24, uint8_t reg, uint8_t* pValue);
int16_t NRF24WriteRegister(NRF24* pNRF24, uint8_t reg, uint8_t value);

#endif /* NRF24_H_ */
