/*
 * network.c
 *
 *  Created on: 05.01.2013
 *      Author: AndreR
 */

#include "network.h"

#include "stm32_eth.h"
#include "stm32f10x.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_exti.h"
#include "misc.h"

#include "constants.h"
#include "util/console.h"
#include "config.h"

#include "FreeRTOS.h"
#include "task.h"

#include "commands.h"

#include "network/inet.h"
#include "network/ethernet.h"
#include "network/arp.h"
#include "network/ipv4.h"
#include "network/udp.h"
#include "network/icmp.h"
#include "network/igmpv3.h"

#include <string.h>

#define PHY_ADDRESS       0x01

Network network;

static uint8_t rxBuffer[NETWORK_NUM_DESCRIPTORS*NETWORK_BUFFER_SIZE];
static uint8_t txBuffer[NETWORK_NUM_DESCRIPTORS*NETWORK_BUFFER_SIZE];

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP);
int16_t NetworkSendICMPEchoReply(ICMPEcho* pEchoRequest, uint8_t* pData, uint16_t dataLength, MAC targetMAC, IPv4Address dstIp);

void ETH_IRQHandler(void)
{
	volatile uint32_t dmasr = ETH->DMASR;
	ETH->DMASR = dmasr;
//	ETH->DMASR = 0xFFFFFFFF; // clear all flags

	int32_t yield = pdFALSE;
	xQueueSendFromISR(network.eventQueue, (uint32_t*)&dmasr, &yield);
	if(yield)
		taskYIELD();
}

void EXTI9_5_IRQHandler()
{
	EXTI->PR = EXTI_PR_PR5;

	network.macIRQ = 1;
}

static void ethDMATxDescRingInit()
{
	uint32_t i;
	ETH_DMADESCTypeDef* pDMADesc;

	memset(txBuffer, 0, sizeof(txBuffer));

	/* Set the DMATxDescToSet pointer with the first one of the DMATxDescTab list */
	network.pTxDescToSet = network.txDescTab;

	/* Fill each DMATxDesc descriptor with the right values */
	for(pDMADesc = network.txDescTab, i = 0; i < NETWORK_NUM_DESCRIPTORS; i++, pDMADesc++)
	{
		/* Set Buffer1 address pointer */
		pDMADesc->Buffer1Addr = (uint32_t)(&txBuffer[i*NETWORK_BUFFER_SIZE]);

		pDMADesc->Status = ETH_DMATxDesc_CIC_TCPUDPICMP_Full;

		/* Set Transmit End of Ring bit for last descriptor: The DMA returns to the base
       address of the list, creating a Desciptor Ring */
		if(i == (NETWORK_NUM_DESCRIPTORS-1))
		{
			/* Set Transmit End of Ring bit */
			pDMADesc->Status |= ETH_DMATxDesc_TER;
		}
	}

	/* Set Transmit Desciptor List Address Register */
	ETH->DMATDLAR =  (uint32_t) network.txDescTab;
}

static void ethDMARxDescRingInit()
{
	uint32_t i;
	ETH_DMADESCTypeDef* pDMADesc;

	memset(rxBuffer, 0, sizeof(rxBuffer));

	/* Set the DMATxDescToSet pointer with the first one of the DMATxDescTab list */
	network.pRxDescToGet = network.rxDescTab;

	/* Fill each DMATxDesc descriptor with the right values */
	for(pDMADesc = network.rxDescTab, i = 0; i < NETWORK_NUM_DESCRIPTORS; i++, pDMADesc++)
	{
		/* Set Buffer1 address pointer */
		pDMADesc->Buffer1Addr = (uint32_t)(&rxBuffer[i*NETWORK_BUFFER_SIZE]);

		pDMADesc->Status = ETH_DMARxDesc_OWN;
		pDMADesc->ControlBufferSize = NETWORK_BUFFER_SIZE;

		/* Set Transmit End of Ring bit for last descriptor: The DMA returns to the base
       address of the list, creating a Desciptor Ring */
		if(i == (NETWORK_NUM_DESCRIPTORS-1))
		{
			/* Set Transmit End of Ring bit */
			pDMADesc->ControlBufferSize |= ETH_DMARxDesc_RER;
		}
	}

	/* Set Transmit Desciptor List Address Register */
	ETH->DMARDLAR =  (uint32_t) network.rxDescTab;
}


static void hwInit()
{
	/* Enable ETHERNET clock  */
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_ETH_MAC | RCC_AHBPeriph_ETH_MAC_Tx | RCC_AHBPeriph_ETH_MAC_Rx, ENABLE);

	NVIC_InitTypeDef nvicInit;

	/* Enable the Ethernet global Interrupt */
	nvicInit.NVIC_IRQChannel = ETH_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_ETH_DATA;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);

	// -- ETH GPIO
	GPIO_InitTypeDef gpioInit;

	/* ETHERNET pins configuration */
	/* AF Output Push Pull:
	 - ETH_MII_MDIO:	PA2
	 - ETH_MII_TXD3:	PB8
	 - ETH_MII_TX_EN:	PB11
	 - ETH_MII_TXD0:	PB12
	 - ETH_MII_TXD1:	PB13
	 - ETH_MII_MDC:		PC1
	 - ETH_MII_TXD2:	PC2
	 */

	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_Mode = GPIO_Mode_AF_PP;

	gpioInit.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_8 | GPIO_Pin_11 | GPIO_Pin_12 | GPIO_Pin_13;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_1 | GPIO_Pin_2;
	GPIO_Init(GPIOC, &gpioInit);

	// PHY Reset: PA6
	gpioInit.GPIO_Mode = GPIO_Mode_Out_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_6;
	GPIO_Init(GPIOA, &gpioInit);

	GPIOA->BSRR = GPIO_Pin_6;	// clear reset => device active


	/* Input (Reset Value):
	 - ETH_MII_CRS:		PA0
	 - ETH_MII_RX_CLK:	PA1
	 - ETH_MII_COL:		PA3
	 - ETH_MII_RX_DV:	PA7
	 - ETH_MII_RXD2:	PB0
	 - ETH_MII_RXD3:	PB1
	 - ETH_MII_RX_ER:	PB10
	 - ETH_MII_TX_CLK:	PC3
	 - ETH_MII_RXD0:	PC4
	 - ETH_MII_RXD1:	PC5
	*/

	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_Mode = GPIO_Mode_IN_FLOATING;

	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_3 | GPIO_Pin_7;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_10;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_3 | GPIO_Pin_4 | GPIO_Pin_5;
	GPIO_Init(GPIOC, &gpioInit);

	// MII INT: PA5
	gpioInit.GPIO_Pin = GPIO_Pin_5;
	gpioInit.GPIO_Mode = GPIO_Mode_IPU;
	GPIO_Init(GPIOA, &gpioInit);

	// EXTI config for MII IRQ pin
	GPIO_EXTILineConfig(GPIO_PortSourceGPIOA, GPIO_PinSource5);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line5;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;
	EXTI_Init(&extiInit);

	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_ETH_LINK;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI9_5_IRQn;
	NVIC_Init(&nvicInit);
}

static void ethInit()
{
	ETH_InitTypeDef ethInit;

	/* MII/RMII Media interface selection ------------------------------------------*/
	GPIO_ETH_MediaInterfaceConfig(GPIO_ETH_MediaInterface_MII);

	/* Reset ETHERNET on AHB Bus */
	ETH_DeInit();

	/* Software reset */
	ETH_SoftwareReset();

	/* Wait for software reset */
	while(ETH_GetSoftwareResetStatus() == SET)
	{
		vTaskDelay(2/portTICK_RATE_MS);
	}

	/* ETHERNET Configuration ------------------------------------------------------*/
	/* Call ETH_StructInit if you don't like to configure all ETH_InitStructure parameter */
	ETH_StructInit(&ethInit);

	/* Fill ETH_InitStructure parametrs */
	/*------------------------   MAC   -----------------------------------*/
	ethInit.ETH_AutoNegotiation = ETH_AutoNegotiation_Disable;
	ethInit.ETH_Speed = ETH_Speed_100M;
	ethInit.ETH_Mode = ETH_Mode_FullDuplex;
	ethInit.ETH_LoopbackMode = ETH_LoopbackMode_Disable;
	ethInit.ETH_RetryTransmission = ETH_RetryTransmission_Disable;
	ethInit.ETH_AutomaticPadCRCStrip = ETH_AutomaticPadCRCStrip_Enable;
	ethInit.ETH_ReceiveAll = ETH_ReceiveAll_Disable;
	ethInit.ETH_BroadcastFramesReception = ETH_BroadcastFramesReception_Enable;
	ethInit.ETH_PromiscuousMode = ETH_PromiscuousMode_Disable;
	ethInit.ETH_MulticastFramesFilter = ETH_MulticastFramesFilter_Perfect;
	ethInit.ETH_UnicastFramesFilter = ETH_UnicastFramesFilter_Perfect;
	ethInit.ETH_ChecksumOffload = ETH_ChecksumOffload_Enable;
	ethInit.ETH_PauseTime = 65535;
	ethInit.ETH_ZeroQuantaPause = ETH_ZeroQuantaPause_Enable;
	ethInit.ETH_PauseLowThreshold = ETH_PauseLowThreshold_Minus4;
	ethInit.ETH_UnicastPauseFrameDetect = ETH_UnicastPauseFrameDetect_Enable;
	ethInit.ETH_ReceiveFlowControl = ETH_ReceiveFlowControl_Enable;
	ethInit.ETH_TransmitFlowControl = ETH_TransmitFlowControl_Enable;


	/*------------------------   DMA   -----------------------------------*/

	/* When we use the Checksum offload feature, we need to enable the Store and Forward mode:
	 the store and forward guarantee that a whole frame is stored in the FIFO, so the MAC can insert/verify the checksum,
	 if the checksum is OK the DMA can handle the frame otherwise the frame is dropped */
	ethInit.ETH_DropTCPIPChecksumErrorFrame =	ETH_DropTCPIPChecksumErrorFrame_Enable;
	ethInit.ETH_ReceiveStoreForward = ETH_ReceiveStoreForward_Enable;
	ethInit.ETH_TransmitStoreForward = ETH_TransmitStoreForward_Enable;
	ethInit.ETH_ForwardErrorFrames = ETH_ForwardErrorFrames_Disable;
	ethInit.ETH_ForwardUndersizedGoodFrames =	ETH_ForwardUndersizedGoodFrames_Enable;
	ethInit.ETH_SecondFrameOperate = ETH_SecondFrameOperate_Disable;
	ethInit.ETH_AddressAlignedBeats = ETH_AddressAlignedBeats_Enable;
	ethInit.ETH_FixedBurst = ETH_FixedBurst_Enable;
	ethInit.ETH_RxDMABurstLength = ETH_RxDMABurstLength_32Beat;
	ethInit.ETH_TxDMABurstLength = ETH_TxDMABurstLength_32Beat;
	ethInit.ETH_DMAArbitration = ETH_DMAArbitration_RoundRobin_RxTx_2_1;
	ethInit.ETH_FlushReceivedFrame = ETH_FlushReceivedFrame_Enable;

	/* Configure Ethernet */
	ETH_Init(&ethInit, PHY_ADDRESS);

	// configure MII IRQ
	ETH_WritePHYRegister(PHY_ADDRESS, 0x11, 0x03);	// MICR: set interrupt enable and set IRQ to output
	ETH_WritePHYRegister(PHY_ADDRESS, 0x12, 0x20);	// MISR: enable IRQ on link change

	/* Enable the Ethernet Rx Interrupt */
//	ETH_DMAITConfig(ETH_DMA_IT_NIS | ETH_DMA_IT_R, ENABLE);
	ETH_DMAITConfig(ETH_DMA_IT_NIS | ETH_DMA_IT_AIS | ETH_DMA_IT_RO | ETH_DMA_IT_R | ETH_DMA_IT_RPS | ETH_DMA_IT_RBU, ENABLE);
//	ETH->DMAIER = 0xFFFF;

	ethDMARxDescRingInit();
	ethDMATxDescRingInit();

	ETH_MACAddressConfig(ETH_MAC_Address0, network.my.mac.u8);

	MAC igmpMAC = MACSet(0x01, 0x00, 0x5E, 0x00, 0x00, 0x01);	// 224.0.0.1

	ETH_MACAddressConfig(ETH_MAC_Address1, igmpMAC.u8);
	ETH_MACAddressFilterConfig(ETH_MAC_Address1, ETH_MAC_AddressFilter_DA);
	ETH_MACAddressPerfectFilterCmd(ETH_MAC_Address1, ENABLE);

	ETH_MACAddressConfig(ETH_MAC_Address2, network.vision.mac.u8);
	ETH_MACAddressFilterConfig(ETH_MAC_Address2, ETH_MAC_AddressFilter_DA);
	ETH_MACAddressPerfectFilterCmd(ETH_MAC_Address2, ENABLE);

	ETH_Start();

	uint16_t phyidr1 = ETH_ReadPHYRegister(PHY_ADDRESS, 0x02);
	uint16_t phyidr2 = ETH_ReadPHYRegister(PHY_ADDRESS, 0x03);

	ConsolePrint("PHY IDR: %hX %hX\r\n", phyidr1, phyidr2);
}

void NetworkInit()
{
	memset(&network, 0, sizeof(Network));

	network.ethernetIdent = 0;

	network.my.mac = MACSet(2, 84, 73, 71, 69, 82); // ASCII: 02-T-I-G-E-R
	network.my.ip = config.values.ip;
	network.my.port = config.values.port;

	network.ucast.mac = MACSet(2, 0, 0, 0, 0, 0);
	network.ucast.ip = IPv4AddressSet(0, 0, 0, 0);
	network.ucast.port = 0;

	network.vision.mac = MACSet(0x01, 0x00, 0x5E, 0x05, 0x17, 0x02);
	network.vision.ip = IPv4AddressSet(224, 5, 23, 2);
	network.vision.port = 10002;

	network.igmp.respCounter = 0;
	network.igmp.respTime = 500;
	network.igmp.lastSendTime = 0;
	network.igmp.robustness = 2;

	vSemaphoreCreateBinary(network.hSendPacket);

	network.eventQueue = xQueueCreate(10, sizeof(uint32_t));

	hwInit();

	ethInit();
}

static int16_t sendEthernetFrame(uint8_t* pData, uint16_t length)
{
	for(uint16_t bytesSent = 0; bytesSent < length; bytesSent += NETWORK_BUFFER_SIZE)
	{
		if(network.pTxDescToSet->Status & ETH_DMATxDesc_OWN)
		{
			vTaskDelay(2/portTICK_RATE_MS);
			bytesSent -= NETWORK_BUFFER_SIZE;
			continue;
		}

		size_t sendLen = length-bytesSent;
		if(sendLen > NETWORK_BUFFER_SIZE)
			sendLen = NETWORK_BUFFER_SIZE;

		memcpy((void*)network.pTxDescToSet->Buffer1Addr, pData, sendLen);

		pData += sendLen;

		uint32_t tdes0 = ETH_DMATxDesc_OWN;
		if(bytesSent == 0)
			tdes0 |= ETH_DMATxDesc_FS;
		if(bytesSent+sendLen >= length)
			tdes0 |= ETH_DMATxDesc_LS;

		network.pTxDescToSet->Status &= ~(ETH_DMATxDesc_FS | ETH_DMATxDesc_LS);
		network.pTxDescToSet->ControlBufferSize = sendLen;
		network.pTxDescToSet->Status |= tdes0;

		network.pTxDescToSet++;
		if(network.pTxDescToSet == network.txDescTab+NETWORK_NUM_DESCRIPTORS)
			network.pTxDescToSet = network.txDescTab;
	}

	ETH_ResumeDMATransmission();

	return 0;
}

static int16_t decodeFrame(uint8_t* pDecoded, uint16_t length)
{
	FrameHeader* pFrameHeader;

	int16_t result = EthernetDecode(pDecoded, length, &pDecoded, &length, &pFrameHeader);
	if(result)
	{
		return result;
	}

	if(pFrameHeader->dstMAC.cast.multicast == 0 && MACCompare(pFrameHeader->dstMAC, network.my.mac) == 0)
		return 0; // this should not happen, as we have a hardware MAC filter

//	ConsolePrint("A packet for us :)\r\n");

	if(pFrameHeader->type == ETHERNET_PROTOCOL_ARP)
	{
		ARPPacketIPv4* pARP;

		result = ARPDecode(pDecoded, length, &pARP);
		if(result)
			return result;

//		ConsolePrint("ARP\r\n");

		if(	pARP->operation == ARP_OPERATION_REQUEST && pARP->dstProtAddr.u32 == network.my.ip.u32)
		{
			// someone wants to know our MAC address
			ConsolePrint("Replying to ARP!\r\n");

			return NetworkSendARP(ARP_OPERATION_REPLY, pARP->srcHwAddr, pARP->srcProtAddr);
		}
	}

	if(pFrameHeader->type == ETHERNET_PROTOCOL_IPV4)
	{
		IPv4Header* pIPHeader;

		result = IPv4Decode(pDecoded, length, &pDecoded, &length, &pIPHeader);
		if(result)
			return result;

		if(	pIPHeader->dstIP.u32 != network.vision.ip.u32 && pIPHeader->dstIP.u32 != network.my.ip.u32 &&
			pIPHeader->dstIP.u32 != IPv4AddressSet(224, 0, 0, 1).u32)
			return 0;	// we are not addressed

		if(pIPHeader->protocol == IPV4_PROTOCOL_UDP)
		{
			UDPHeader* pUDPHeader;

			result = UDPDecode(pDecoded, length, &pDecoded, &length, &pUDPHeader);
			if(result)
				return result;

			if(pIPHeader->dstIP.u32 == network.vision.ip.u32)
			{
				if(pUDPHeader->dstPort != network.vision.port)
				{
					ConsolePrint("Detected non-official vision packet on port %hu\r\n", pUDPHeader->dstPort);
					ConsolePrint("Source: %hhu.%hhu.%hhu.%hhu\r\n", pIPHeader->srcIP.u8[0], pIPHeader->srcIP.u8[1], pIPHeader->srcIP.u8[2], pIPHeader->srcIP.u8[3]);
					return 0;
				}

//				ConsolePrint("Vision frame! Length: %hu\r\n", pUDPHeader->length);
				if(network.pVisionCb)
				{
					result = (*network.pVisionCb)(pDecoded, length);
				}
				return result;
			}

			if(pUDPHeader->dstPort != network.my.port)
				return 0;

//			ConsolePrint("Valid UDP packet for us\r\n");

			// check for auth packet
			PacketHeader* pHeader = (PacketHeader*)pDecoded;
			if(pHeader->section == SECTION_BASE_STATION && pHeader->cmd == CMD_BASE_STATION_AUTH)
			{
				pDecoded += PACKET_HEADER_SIZE;

				BaseStationAuth* pAuth = (BaseStationAuth*)pDecoded;

				if(pAuth->key == 0x42424242)
				{
					network.ucast.mac = pFrameHeader->srcMAC;
					network.ucast.ip = pIPHeader->srcIP;
					network.ucast.port = pUDPHeader->srcPort;

					ConsolePrint("Valid authentication from:\r\n");
					ConsolePrint("MAC: %02hhX-%02hhX-%02hhX-%02hhX-%02hhX-%02hhX\r\n", network.ucast.mac.u8[0], network.ucast.mac.u8[1],
							network.ucast.mac.u8[2], network.ucast.mac.u8[3], network.ucast.mac.u8[4], network.ucast.mac.u8[5]);
					ConsolePrint("IP: %hhd.%hhd.%hhd.%hhd\r\n", network.ucast.ip.u8[0], network.ucast.ip.u8[1], network.ucast.ip.u8[2], network.ucast.ip.u8[3]);
					ConsolePrint("Port: %hd\r\n", network.ucast.port);
				}
				else
				{
					ConsolePrint("Invalid base station authentication try\r\n");
				}

				return result;
			}

			if(network.pUDPCb)
			{
				result = (*network.pUDPCb)(pDecoded, length);
			}
			return result;
		}

		if(pIPHeader->protocol == IPV4_PROTOCOL_ICMP)
		{
			ICMPHeader* pICMPHeader;

			result = ICMPDecode(pDecoded, length, &pDecoded, &length, &pICMPHeader);
			if(result)
				return result;

			if(pICMPHeader->type == ICMP_TYPE_ECHO_REQUEST)
			{
				ICMPEcho* pICMPEcho;

				result = ICMPDecodeEcho(pDecoded, length, &pDecoded, &length, &pICMPEcho);
				if(result)
					return result;

				ConsolePrint("ICMP echo request, ident: %hu, seq: %hu, dataLen: %hu\r\n", pICMPEcho->identifier, pICMPEcho->sequenceNumber, length);

				result = NetworkSendICMPEchoReply(pICMPEcho, pDecoded, length, pFrameHeader->srcMAC, pIPHeader->srcIP);
				if(result)
					return result;
			}
		}

		if(pIPHeader->protocol == IPV4_PROTOCOL_IGMP)
		{
			uint8_t type = 0;
			IGMPv3GetType(pDecoded, &type);

			ConsolePrint("IGMP msg. Type: %hhX\r\n", type);

			if(type == IGMP_MEMBERSHIP_QUERY)
			{
				IGMPv3MembershipQuery* pQuery;
				IPv4Address* pSourceList;

				result = IGMPv3DecodeMembershipQuery(pDecoded, length, &pQuery, &pSourceList);
				if(result)
					return result;

				if(pQuery->groupAddress.u32 == 0 || pQuery->groupAddress.u32 == network.vision.ip.u32)
				{
					// general query
					network.igmp.lastSendTime = xTaskGetTickCount();
					network.igmp.respCounter = 0;
					network.igmp.robustness = pQuery->QRV ? pQuery->QRV : 2;
					uint32_t respTime = IGMPv3DecodeMaxRespCode(pQuery->maxRespCode);
					network.igmp.respTime = xTaskGetTickCount() % respTime;
				}
			}

			if(type == IGMPV3_MEMBERSHIP_REPORT)
			{
			}
		}
	}

	return 0;
}

static void gatherEthernetFrame()
{
	uint32_t bytesRead = 0;

	while(1)
	{
		uint32_t status = network.pRxDescToGet->Status;
		if(status & ETH_DMARxDesc_OWN)
			return;

		uint32_t dataLength = NETWORK_BUFFER_SIZE;
		uint32_t frameLength = (status & 0xDFFF0000) >> 16;
//		uint32_t firstLast = (status & 0x300) >> 8;
//		uint32_t pce = (status & 0x1);

//		ConsolePrint("frameLength: %u, firstLast: %X, PCE: %u\r\n", frameLength, firstLast, pce);

		if(!(status & ETH_DMARxDesc_ES))
		{
			if(status & ETH_DMARxDesc_LS)
			{
				dataLength = frameLength % NETWORK_BUFFER_SIZE;
			}

			memcpy(network.rxProcBuf+bytesRead, (void*)network.pRxDescToGet->Buffer1Addr, dataLength);

			bytesRead += dataLength;

			if(status & ETH_DMARxDesc_LS)
			{
	//			uint16_t ethLen = *((uint16_t*)(network.rxProcBuf+12));

	//			ConsolePrint("Frame complete, length: %u, %hu\r\n", bytesRead, ethLen);

				int16_t result = decodeFrame(network.rxProcBuf, bytesRead);
				if(result)
					ConsolePrint("Decode result: %hX\r\n", result);

				bytesRead = 0;
			}
		}
		else
		{
			ConsolePrint("Errornous frame (overflow)\r\n");
		}

		network.pRxDescToGet->Status |= ETH_DMARxDesc_OWN;

		network.pRxDescToGet++;
		if(network.pRxDescToGet == network.rxDescTab+NETWORK_NUM_DESCRIPTORS)
			network.pRxDescToGet = network.rxDescTab;
	}
}

void NetworkSetUDPCallback(UDPCallback cb)
{
	network.pUDPCb = cb;
}

void NetworkSetVisionCallback(VisionCallback cb)
{
	network.pVisionCb = cb;
}

void NetworkTask(void* params)
{
	(void)params;

	uint32_t event;

	int32_t lastStatsTime = xTaskGetTickCount();

	while (1)
	{
		if(xQueueReceive(network.eventQueue, &event, 10/portTICK_RATE_MS) == pdTRUE)
		{
//			ConsolePrint("DMASR: 0x%X\r\n", event);

			if(event & ETH_DMASR_RS) // receive status
			{
				gatherEthernetFrame();
			}
			else
			{
				ConsolePrint("None-RX event: %08X\r\n", event);
			}

			if(event & ETH_DMASR_RBUS)
			{
				ConsolePrint("RX buffer unavailable\r\n");
			}

			if(event & ETH_DMASR_RPSS)
			{
				ConsolePrint("RX stopped\r\n");
			}

			if(event & ETH_DMASR_ROS)
			{
				ConsolePrint("RX overflow!\r\n");
			}
		}
		else
		{
			if(!(network.pRxDescToGet->Status & ETH_DMARxDesc_OWN))
			{
				gatherEthernetFrame();
			}
		}

		// IGMPv3 work
		if(network.igmp.respTime != 0)
		{
			if(xTaskGetTickCount() - network.igmp.lastSendTime > network.igmp.respTime/portTICK_RATE_MS)
			{
				network.igmp.lastSendTime = xTaskGetTickCount();

				IGMPv3GroupRecord record;
				record.auxDataLen = 0;
				record.numberOfSources = 0;
				record.recordType = IGMPV3_MODE_IS_EXCLUDE;
				record.multicastAddress.u32 = network.vision.ip.u32;

				NetworkSendIGMPv3MembershipReport(1, record._data, IGMPV3_GROUP_RECORD_SIZE);

				network.igmp.respCounter++;
				if(network.igmp.respCounter >= network.igmp.robustness)
					network.igmp.respTime = 0;
			}
		}

		if(network.macIRQ)
		{
			network.macIRQ = 0;

			ETH_ReadPHYRegister(PHY_ADDRESS, 0x12);	// clear IRQ by reading MISR register

			uint16_t status = ETH_ReadPHYRegister(PHY_ADDRESS, PHY_BSR);	// for an unknown reason the first read does not reflect internal state
			status = ETH_ReadPHYRegister(PHY_ADDRESS, PHY_BSR);

			if(status & PHY_Linked_Status)
			{
				ConsolePrint("eth0 up\r\n");

				network.igmp.respTime = 500;	// send IGMPv3 membership
			}

			if(!(status & PHY_Linked_Status))
			{
				ConsolePrint("eth0 down\r\n");
			}
		}

		if(xTaskGetTickCount() - lastStatsTime > 500/portTICK_RATE_MS)
		{
			lastStatsTime = xTaskGetTickCount();

			// construct header
			PacketHeader header;
			header.section = SECTION_BASE_STATION;
			header.cmd = CMD_BASE_STATION_STATS;
			header.len = sizeof(BaseStationStats);

			BaseStationStats stats;
			memcpy(&stats.wifi, &wifi.nrf24.stats, sizeof(stats.wifi));
			// TODO: create and fill eth stats
			memset(&stats.eth, 0, sizeof(stats.eth));

			NetworkSendPacket(&header, (uint8_t*)&stats);
		}
	}
}

int16_t NetworkSendPacket(PacketHeader* pHeader, uint8_t* pData)
{
	if(pHeader->len + PACKET_HEADER_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	memcpy(network.txProcBuf, pHeader->_data, PACKET_HEADER_SIZE);
	memcpy(network.txProcBuf + PACKET_HEADER_SIZE, pData, pHeader->len);

	uint16_t encLength = pHeader->len + PACKET_HEADER_SIZE;

	uint16_t dstPort = network.ucast.port;

	IPv4Address dstIp = network.ucast.ip;

	MAC dstMAC = network.ucast.mac;

	int16_t result = UDPEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength, network.my.port, dstPort);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = IPv4Encode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIp, IPV4_PROTOCOL_UDP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = sendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

int16_t NetworkSendARP(uint8_t operation, MAC targetMAC, IPv4Address targetIP)
{
	if(ARP_PACKET_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	uint16_t encLength;

	int16_t result = ARPEncode(	network.txProcBuf, NETWORK_MAX_FRAME_SIZE, &encLength,
								operation, network.my.mac, network.my.ip, targetMAC, targetIP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, targetMAC, ETHERNET_PROTOCOL_ARP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = sendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

int16_t NetworkSendICMPEchoReply(ICMPEcho* pEchoRequest, uint8_t* pData, uint16_t dataLength, MAC targetMAC, IPv4Address dstIp)
{
	if(dataLength + ICMP_ECHO_SIZE + ICMP_HEADER_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	uint16_t encLength;

	memcpy(network.txProcBuf, pData, dataLength);

	int16_t result = ICMPEncodeEcho(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, dataLength, &encLength,
								pEchoRequest->identifier, pEchoRequest->sequenceNumber);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = ICMPEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength, ICMP_TYPE_ECHO_REPLY, 0);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = IPv4Encode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIp, IPV4_PROTOCOL_ICMP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, targetMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = sendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

int16_t NetworkSendIGMPv3MembershipReport(uint16_t numberOfGroupRecords, uint8_t* recordData, uint16_t recordDataLength)
{
	if(recordDataLength + IGMPV3_MEMBERSHIP_REPORT_SIZE > NETWORK_MAX_FRAME_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	ConsolePrint("Sending IGMPv3 membership report\r\n");

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	uint16_t encLength = 0;
	int16_t result = 0;

	result = IGMPv3EncodeMembershipReport(	network.txProcBuf, NETWORK_MAX_FRAME_SIZE,
											&encLength, numberOfGroupRecords, recordData, recordDataLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	IPv4Address dstIP = IPv4AddressSet(224, 0, 0, 22);
	MAC dstMAC = MACSet(0x01, 0x00, 0x5E, 0x00, 0x00, 0x16);	// 224.0.0.22

	result = IPv4EncodeWithRouterAlertOption(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
						network.ethernetIdent, network.my.ip, dstIP, IPV4_PROTOCOL_IGMP);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = EthernetEncode(network.txProcBuf, NETWORK_MAX_FRAME_SIZE, encLength, &encLength,
							network.my.mac, dstMAC, ETHERNET_PROTOCOL_IPV4);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	result = sendEthernetFrame(network.txProcBuf, encLength);
	if(result)
	{
		xSemaphoreGive(network.hSendPacket);
		return result;
	}

	network.ethernetIdent++;

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

void NetworkChangePort(uint16_t port)
{
	config.values.port = port;
	network.my.port = port;
}

void NetworkChangeIP(IPv4Address ip)
{
	config.values.ip = ip;
	network.my.ip = ip;
}
