/*
 * wifi.c
 *
 *  Created on: 02.12.2012
 *      Author: AndreR, FriederB
 */

#include "wifi.h"

#include "hal/nrf24.h"
#include "util/console.h"
#include "util/spi.h"
#include "util/bit_banding.h"
#include "util/critical.h"
#include "spi3.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_exti.h"
#include "misc.h"
#include "constants.h"
#include "config.h"
#include "commands.h"
#include "vision.h"

#include "FreeRTOS.h"
#include "task.h"

WifiGlobal wifi;

#define STATE_NO_LINK 0x01
#define STATE_DAMAGED 0x02

void EXTI2_IRQHandler()
{
	BBPeriphSet(EXTI_BASE + 0x14, 2);	// reset EXTI flag

	NRF24IRQ(&wifi.nrf24);
}

static void process(uint8_t id)
{
	int16_t result;
	uint16_t length;

	if(id > NRF24_MAX_BOT_ID)
		return;

	wifi.bots[id].lastExecTime = rtosHfTickCount;

	NRF24ChangeId(&wifi.nrf24, id);

	result = NRF24Process(&wifi.nrf24);
	if(result == ERROR_NRF24_IRQ_MALFUNCTION)
	{
//		wifi.state[i] = STATE_DAMAGED;
		ConsolePrint("IRQ malfunction => module disabled\r\n");
		return;
	}
	if(result == ERROR_NRF24_NO_LINK)
	{
		// remote module not reachable
		wifi.bots[id].noLinkCounter++;

		if(wifi.bots[id].noLinkCounter > config.values.maxNoLinkTries)
		{
			wifi.bots[id].state = STATE_NO_LINK;
			ConsolePrint("Remote module %hhu not responding => no link\r\n", id);
		}
	}
	if(!result)
	{
		wifi.bots[id].noLinkCounter = 0;
		wifi.bots[id].state = 0;
	}

	while(FifoPacketSyncedNumPackets(&wifi.nrf24.rxFifo) > 0)
	{
		FifoPacketSyncedGetNextPacketDataLength(&wifi.nrf24.rxFifo, &length);

		// construct header
		PacketHeader header;
		header.section = SECTION_BASE_STATION;
		header.cmd = CMD_BASE_STATION_ACOMMAND;
		header.len = length+1;

		// construct BaseStationACommand
		BaseStationACommand* pACmd = (BaseStationACommand*)wifi.sendBuf;
		pACmd->id = id;

		// now fill child packet
		result = FifoPacketSyncedPop(&wifi.nrf24.rxFifo, &length, wifi.sendBuf+1, WIFI_SEND_BUFFER_SIZE);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&wifi.nrf24.rxFifo);
			return;
		}

//		PacketHeader* pHeader = (PacketHeader*)(wifi.sendBuf+1);
//		ConsolePrint("Rx cmd: 0x%hhX 0x%hhX, len: %hu\r\n", pHeader->section, pHeader->cmd, header.len);

		result = NetworkSendPacket(&header, wifi.sendBuf);
		if(result)
			ConsolePrint("Network send failed: 0x%hX\r\n", result);
	}
}

static int16_t udpCallback(uint8_t* pData, uint16_t dataLength)
{
	(void)dataLength;

	// pData should point to a BaseStationACommand packet
	PacketHeader* pHeader = (PacketHeader*)pData;
	if(pHeader->section != SECTION_BASE_STATION)
	{
		ConsolePrint("Invalid section 0x%X\r\n", pHeader->section);
		return 0;
	}

	pData += PACKET_HEADER_SIZE;

	switch(pHeader->cmd)
	{
		case CMD_BASE_STATION_ACOMMAND:
		{
			BaseStationACommand* pCmd = (BaseStationACommand*)pData;

			uint8_t botId = pCmd->id;
			uint16_t length = pHeader->len-1;

			if(botId > NRF24_MAX_BOT_ID)
			{
				ConsolePrint("Invalid bot id %hhu\r\n", botId);
				return 0;
			}

//			ConsolePrint("ID: %hhu, length: %hu\r\n", botId, length);

			TASK_MUTEX(wifi.hTxMutex)
			{
				if(wifi.txFifo.fifo.EmptyBytes < length+1+FIFO_PACKET_SYNCED_HEADER_SIZE)
				{
					ConsolePrint("Command for bot %hhu lost, out of memory\r\n", botId);
					return ERROR_NOT_ENOUGH_MEMORY;
				}

				FifoPacketSyncedPush(&wifi.txFifo, 1, &botId);
				FifoPacketSyncedPush(&wifi.txFifo, length, pData+1);
				FifoPacketSyncedComplete(&wifi.txFifo);
			}

			/*
			TASK_MUTEX(wifi.nrfMutex)
			{
				NRF24Send(&wifi.nrf24, length, pData+1);

				process(botId);

				LinkedListRemove(&wifi.processQueue, &wifi.bots[botId]);
				LinkedListPushBack(&wifi.processQueue, &wifi.bots[botId]);
			}
			*/
		}
		break;
		case CMD_BASE_STATION_PING:
		{
			int16_t result = NetworkSendPacket(pHeader, pData);
			if(result)
			{
				ConsolePrint("NetworkSendPacket failed 0x%hX\r\n", result);
			}
		}
		break;
		case CMD_BASE_STATION_VISION_CONFIG:
		{
			BaseStationVisionConfig* pConfig = (BaseStationVisionConfig*)pData;

			WifiSetVisionRate(pConfig->visionRate);
			vision.invertPos = pConfig->invertPosition;
			vision.blueTeam = pConfig->blueTeam;
		}
		break;
		default:
		{
			ConsolePrint("Invalid command 0x%X\r\n", pHeader->cmd);
			return 0;
		}
		break;
	}

	return 0;
}

void WifiSetVisionRate(uint16_t rateHz)
{
	uint32_t rate = rateHz;
	wifi.visionPauseTime = 50000/rate;
}

void WifiSendPose(uint8_t botId, float x, float y, float orient)
{
	if(botId > NRF24_MAX_BOT_ID)
	{
		ConsolePrint("Invalid pose id %hhu\r\n", botId);
		return;
	}

	if(wifi.bots[botId].state != 0)
		return;

	if(rtosHfTickCount - wifi.bots[botId].lastVisionTime < wifi.visionPauseTime)
		return;

	wifi.bots[botId].lastVisionTime = rtosHfTickCount;

	static uint8_t data[PACKET_HEADER_SIZE+LEN_CTRL_VISION_POS];

	PacketHeader* pHeader = (PacketHeader*)data;
	pHeader->cmd = CMD_CTRL_VISION_POS;
	pHeader->section = SECTION_CTRL;
	pHeader->len = LEN_CTRL_VISION_POS;

	CtrlVisionPos* pPos = (CtrlVisionPos*)(data+PACKET_HEADER_SIZE);
	pPos->pos[0] = (int16_t)x;
	pPos->pos[1] = (int16_t)y;
	pPos->pos[2] = (int16_t)(orient*1000.0f);

	TASK_MUTEX(wifi.hTxMutex)
	{
		if(wifi.txFifo.fifo.EmptyBytes < PACKET_HEADER_SIZE+LEN_CTRL_VISION_POS+1+FIFO_PACKET_SYNCED_HEADER_SIZE)
		{
			ConsolePrint("Vision pose for bot %hhu lost, out of memory\r\n", botId);
			return;
		}

		FifoPacketSyncedPush(&wifi.txFifo, 1, &botId);
		FifoPacketSyncedPush(&wifi.txFifo, PACKET_HEADER_SIZE+LEN_CTRL_VISION_POS, data);
		FifoPacketSyncedComplete(&wifi.txFifo);
	}

	/*
	TASK_MUTEX(wifi.nrfMutex)
	{
		NRF24Send(&wifi.nrf24, PACKET_HEADER_SIZE+LEN_CTRL_VISION_POS, data);

		process(botId);

		LinkedListRemove(&wifi.processQueue, &wifi.bots[botId]);
		LinkedListPushBack(&wifi.processQueue, &wifi.bots[botId]);
	}
	*/
}

static uint8_t lockFifo(void* pUserData)
{
	(void)pUserData;
	return 0;
}

static void releaseFifo(uint8_t lock, void* pUserData)
{
	(void)lock;
	(void)pUserData;
}

void WifiInit()
{
	int16_t result;

	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IPU;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOD, &gpioInit);	// shared IRQ line

	GPIO_EXTILineConfig(GPIO_PortSourceGPIOD, GPIO_PinSource2);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line2;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;	// active low
	EXTI_Init(&extiInit);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_WIFI;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	static uint8_t txQueueData[WIFI_TX_QUEUE_SIZE];

	wifi.hTxMutex = xSemaphoreCreateMutex();
	FifoPacketSyncedInit(&wifi.txFifo, txQueueData, WIFI_TX_QUEUE_SIZE, &lockFifo, &releaseFifo, 0);

	vTaskDelay(200/portTICK_RATE_MS);	// power up delay for NRF24 module

	for(uint8_t i = 0; i < WIFI_MAX_ROBOTS; i++)
	{
		wifi.bots[i].state = STATE_NO_LINK;
		wifi.bots[i].lastExecTime = 0;
		wifi.bots[i].lastVisionTime = 0;
		wifi.bots[i].noLinkCounter = 0;
		wifi.bots[i].id = i;
	}

	NRF24Data data;
	static uint8_t rxData[WIFI_MODULE_BUF_SIZE];
	static uint8_t txData[WIFI_MODULE_BUF_SIZE];

	data.pDataRx = rxData;
	data.pDataTx = txData;
	data.rxSize = WIFI_MODULE_BUF_SIZE;
	data.txSize = WIFI_MODULE_BUF_SIZE;
	data.pSPI = &spi3;
	data.rx = 0;
	data.channel = config.values.channel;
	data.id = 1;
	data.speed = config.values.speed;
	data.maxRt = config.values.maxRt;
	data.ackDelayUs = config.values.ackDelayUs;

	data.pCEPort = GPIOC;
	data.cePin = GPIO_Pin_13;
	data.pCSPort = GPIOC;
	data.csPin = GPIO_Pin_14;

	result = NRF24Init(&wifi.nrf24, &data);
	if(result)
		ConsolePrint("NRF24Init failed: 0x%02hX\r\n", result);

	vTaskDelay(100/portTICK_RATE_MS);

	NetworkSetUDPCallback(&udpCallback);

	wifi.scan = 0xFF;
	wifi.visionPauseTime = 1500; // 30ms
}

void WifiChangeChannel(uint8_t channel)
{
	NRF24ChangeChannel(&wifi.nrf24, channel);

	config.values.channel = channel;
}

void WifiChangeSpeed(uint8_t speed)
{
	if(speed != NRF24_SPEED_250K && speed != NRF24_SPEED_1M && speed != NRF24_SPEED_2M)
		return;

	NRF24ChangeSpeed(&wifi.nrf24, speed);

	config.values.speed = speed;
}

void WifiChangeMaxRt(uint8_t maxRt)
{
	if(maxRt == 0 || maxRt > 15)
		return;

	NRF24ChangeMaxRt(&wifi.nrf24, maxRt);

	config.values.maxRt = maxRt;
}

int16_t WifiChangeAckDelay(uint16_t delayUs)
{
	config.values.ackDelayUs = delayUs;

	return NRF24ChangeAckDelay(&wifi.nrf24, delayUs);
}

static WifiRobot* findLRURobot()
{
	WifiRobot* pLRUBot = 0;
	uint32_t execDiff = 0;

	for(uint8_t i = 0; i < WIFI_MAX_ROBOTS; i++)
	{
		WifiRobot* pBot = &wifi.bots[i];

		if(pBot->state != 0)
			continue;

		if(rtosHfTickCount - pBot->lastExecTime > execDiff)
		{
			execDiff = rtosHfTickCount - pBot->lastExecTime;

			pLRUBot = pBot;
		}
	}

	return pLRUBot;
}

void WifiTask(void* params)
{
	(void) params;

	ConsolePrint("WifiTask started\r\n");

	int16_t result = 0;

	uint32_t startTime = rtosHfTickCount;
	uint32_t endTime;
	uint8_t worked = 0;

	while(1)
	{
		// scan something?
		if(wifi.scan != 0xFF)
		{
			uint8_t scanResult;

			ConsolePrint("Scanning wifi channels...\r\n");
			ConsolePrint("     00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15\r\n000: ");

			for(uint8_t i = 0; i < 128; i++)
			{
				result = NRF24ChangeChannel(&wifi.nrf24, i);
				if(result)
					ConsolePrint("Change channel failed: 0x%hX\r\n", result);

				result = NRF24CheckChannel(&wifi.nrf24, &scanResult);
				if(result)
					ConsolePrint("Check channel failed: 0x%hX\r\n", result);

				ConsolePrint("%2hhu ", scanResult);

				if((i+1)%16 == 0 && i != 0 && i != 127)
				{
					ConsolePrint("\r\n%03hhu: ", i+1);

					vTaskDelay(100/portTICK_RATE_MS);
				}
			}

			ConsolePrint("\r\n");

			NRF24ChangeChannel(&wifi.nrf24, config.values.channel);

			wifi.scan = 0xFF;
		}

		startTime = rtosHfTickCount;
		worked = 0;

		while(42)
		{
			uint16_t length;

			// first check tx fifo
			while(FifoPacketSyncedNumPackets(&wifi.txFifo) > 0)
			{
				TASK_MUTEX(wifi.hTxMutex)
				{
					result = FifoPacketSyncedPop(&wifi.txFifo, &length, wifi.sendBuf, WIFI_SEND_BUFFER_SIZE);
					if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
					{
						FifoPacketSyncedDrop(&wifi.txFifo);
						ConsolePrint("Dropped tx wifi packet. Buffer too small\r\n");
						continue;
					}
					if(result)
						continue;
				}

				uint8_t botId = wifi.sendBuf[0];

				NRF24Send(&wifi.nrf24, length-1, wifi.sendBuf+1);

				process(botId);
			}

			// Do usual LRU scheduling
			WifiRobot* pBot = findLRURobot();
			if(pBot == 0 || (rtosHfTickCount - pBot->lastExecTime) < 500)
			{
				break;
			}

			process(pBot->id);

			worked = 1;
			break;
		}

		if(worked == 0)
			vTaskDelay(2/portTICK_RATE_MS);

		endTime = rtosHfTickCount;

		wifi.runtime = (endTime-startTime)/5;
	}
}
