/*
 * bootloader.c
 *
 *  Created on: 03.06.2013
 *      Author: AndreR
 */

#include "bootloader.h"

#include "util/crc.h"
#include "util/fifo_packet_synced.h"
#include "network.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/boot.h"
#include "util/console.h"

#include "FreeRTOS.h"
#include "semphr.h"
#include "queue.h"

#define FIFO_CMD_SIZE 1200

// events
#define EVENT_ERASE 1
#define EVENT_DATA 2

struct _Bootloader
{
	uint8_t cmdBuf[FIFO_CMD_SIZE];
	xSemaphoreHandle hCmdMutex;
	FifoPacketSynced commands;

	xQueueHandle hEventQueue;

	BootloaderData procData;

	uint8_t targetMain;
} bootloader;

static void bootloaderCb(IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;
	(void)dataLength;

	switch(pHeader->command)
	{
		case BOOTLOADER_RESPONSE:
		{
			BootloaderResponse* pResponse = (BootloaderResponse*)pData;

			PacketHeader header;
			header.section = SECTION_BOOTLOADER;
			header.cmd = CMD_BOOTLOADER_RESPONSE;
			header.len = LEN_BOOTLOADER_RESPONSE;

			NetworkSendPacket(&header, (uint8_t*)pResponse);
		}
		break;
	}
}

static void sendACK(uint32_t offset)
{
	BootloaderResponse response;
	response.type = BOOTLOADER_RESPONSE_ACK;
	response.offset = offset;

	PacketHeader header;
	header.section = SECTION_BOOTLOADER;
	header.cmd = CMD_BOOTLOADER_RESPONSE;
	header.len = LEN_BOOTLOADER_RESPONSE;

	NetworkSendPacket(&header, (uint8_t*)&response);
}

static void sendNACK(uint32_t offset)
{
	BootloaderResponse response;
	response.type = BOOTLOADER_RESPONSE_NACK;
	response.offset = offset;

	PacketHeader header;
	header.section = SECTION_BOOTLOADER;
	header.cmd = CMD_BOOTLOADER_RESPONSE;
	header.len = LEN_BOOTLOADER_RESPONSE;

	NetworkSendPacket(&header, (uint8_t*)&response);
}

void BootloaderInit()
{
	IntercomClient client;
	client.pUserData = 0;
	client.section = INTERCOM_SECTION_BOOTLOADER;
	client.rxCallback = &bootloaderCb;

	IntercomAddClient(&intercom1, client);

	bootloader.targetMain = 1;

	bootloader.hEventQueue = xQueueCreate(10, sizeof(uint8_t));

	bootloader.hCmdMutex = xSemaphoreCreateRecursiveMutex();

	FifoPacketSyncedInit(&bootloader.commands, bootloader.cmdBuf, FIFO_CMD_SIZE, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, bootloader.hCmdMutex);
}

void BootloaderStartProgramming(uint8_t processor)
{
	if(processor == INTERCOM_ADDRESS_MAIN)
	{
		bootloader.targetMain = 1;
	}

	if(processor == INTERCOM_ADDRESS_MEDIA)
	{
		bootloader.targetMain = 0;
	}

	uint8_t event = EVENT_ERASE;
	xQueueSend(bootloader.hEventQueue, &event, portMAX_DELAY);
}


void BootloaderFlashData(BootloaderData* pData)
{
	int16_t result = FifoPacketSyncedPush(&bootloader.commands, sizeof(BootloaderData), (uint8_t*)pData);
	if(result)
	{
		ConsolePrint("Store data to fifo failed 0x%hX\r\n", result);
		sendNACK(pData->offset);
		return;
	}

	FifoPacketSyncedComplete(&bootloader.commands);

	uint8_t event = EVENT_DATA;
	xQueueSend(bootloader.hEventQueue, &event, portMAX_DELAY);
}

void BootloaderTask(void* params)
{
	(void)params;

	uint8_t event;

	while(1)
	{
		xQueueReceive(bootloader.hEventQueue, &event, portMAX_DELAY);

		switch(event)
		{
			case EVENT_ERASE:
			{
				if(bootloader.targetMain)
				{
					if(BootEraseProgramFlash())
						ConsolePrint("Erase failed\r\n");

					sendACK(0);
				}
				else
				{
					IntercomHeader header;
					header.address = INTERCOM_ADDRESS_MEDIA;
					header.section = INTERCOM_SECTION_BOOTLOADER;
					header.command = BOOTLOADER_ERASE;

					IntercomSend(&intercom1, &header, 0, 0);

					ConsolePrint("Erasing media\r\n");
				}
			}
			break;
			case EVENT_DATA:
			{
				uint16_t length;
				int16_t result = FifoPacketSyncedPop(&bootloader.commands, &length, (uint8_t*)&bootloader.procData, sizeof(BootloaderData));
				if(result)
					continue;

				uint32_t crc = CRC32Checksum(bootloader.procData.data, 64);

				if(crc != bootloader.procData.crc)
				{
					ConsolePrint("CRC mismatch\r\n");
					ConsolePrint("%08X/%08X\r\n", bootloader.procData.crc, crc);
					sendNACK(bootloader.procData.offset);

					continue;
				}

				if(bootloader.targetMain)
				{
					BootProgramFlash(bootloader.procData.data, 64, bootloader.procData.offset);

					sendACK(bootloader.procData.offset);
				}
				else
				{
					IntercomHeader header;
					header.address = INTERCOM_ADDRESS_MEDIA;
					header.section = INTERCOM_SECTION_BOOTLOADER;
					header.command = BOOTLOADER_DATA;

					IntercomSend(&intercom1, &header, sizeof(BootloaderData), (uint8_t*)&bootloader.procData);
				}
			}
			break;
		}
	}
}

