/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"

#include "stm32f4xx_rcc.h"

#include "FreeRTOS.h"
#include "task.h"

#include "hal/buzzer.h"
#include "hal/led.h"
#include "commands.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/critical.h"
#include "util/fifo8_synced.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "util/boot.h"
#include "config.h"

#include "network.h"

#include <stdio.h>
#include <string.h>

CLIGlobal cli;

static void processCommand();

#define COLOR_NONE	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_BLUE	3

#define COLOR_MEDIA COLOR_BLUE
#define COLOR_MAIN COLOR_NONE

static void echoColor(uint8_t color)
{
	TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
	{
		switch(color)
		{
		case COLOR_RED:
			sprintf((char*)cli.procBuf, "\e[0;31m");
			break;
		case COLOR_GREEN:
			sprintf((char*)cli.procBuf, "\e[0;32m");
			break;
		case COLOR_BLUE:
			sprintf((char*)cli.procBuf, "\e[0;36m");
			break;
		default:
			sprintf((char*)cli.procBuf, "\e[0m");
			break;
		}

		FifoSyncedPush(&usart1.txFifo, strlen((char*)cli.procBuf), cli.procBuf);
	}
}

static void echoTargetColor(uint8_t addr)
{
	switch(addr)
	{
	case INTERCOM_ADDRESS_MEDIA:
		echoColor(COLOR_MEDIA);
		break;
	default:
		echoColor(COLOR_MAIN);
		break;
	}
}

// IRQ Level: USART1 IRQ
static void usart1Callback()
{
	// called on new complete packet in usart1 RX fifo
}

// IRQ Level: Task
static void consoleCallback()
{
	// called after ConsolePrint
}

// IRQ Level: Task
static void intercomCallback(IntercomHeader* pHeader, uint16_t length, uint8_t* pData, void* pUserData)
{
	pUserData = pUserData;

	// called after new console intercom rx
	TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
	{
		echoTargetColor(pHeader->command);
		FifoSyncedPush(&usart1.txFifo, length, pData);
		echoTargetColor(cli.currentTarget);

		USART1StartTx();
	}
}


void CLIInit()
{
	cli.currentTarget = INTERCOM_ADDRESS_MAIN;
	cli.hProcBufMutex = xSemaphoreCreateRecursiveMutex();

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];
	static uint8_t printData[CONSOLE_PRINT_SIZE];
	static uint8_t scanData[CONSOLE_SCAN_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.pPrintBuf = printData;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.pScanBuf = scanData;
	conData.scanSize = CONSOLE_SCAN_SIZE;
	conData.txCallback = &consoleCallback;
	conData.pUserData = 0;

	ConsoleInit(conData);

	usart1.callback = &usart1Callback;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &intercomCallback;
	client.section = INTERCOM_SECTION_CONSOLE;

	IntercomAddClient(&intercom1, client);
}

void CLITask(void* params)
{
	params = params;

	portTickType xLastWakeTime;
	portTickType xLastLEDTime;

	int16_t result = 0;
	uint16_t length;

	xLastWakeTime = xTaskGetTickCount();
	xLastLEDTime = xLastWakeTime;

	LEDMustOff(LED_MAIN_GREEN);

	while(1)
	{
		TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
		{
			// check user input to terminal
			while(FifoPacketSyncedNumPackets(&usart1.rxFifo) > 0)
			{
				// get packet
				result = FifoPacketSyncedPop(&usart1.rxFifo, &length, cli.procBuf, CLI_PROC_SIZE-1);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&usart1.rxFifo);
					continue;
				}

				if(result)
					continue;

				ConsolePrint("\r\n");

				cli.procBuf[length] = '\0';

				// target selection logic
				if(strcmp((char*)cli.procBuf, "target main") == 0)
				{
					cli.currentTarget = INTERCOM_ADDRESS_MAIN;
					echoTargetColor(cli.currentTarget);
					ConsolePrint("New target: MAIN\r\n");
					continue;
				}

				if(strcmp((char*)cli.procBuf, "target media") == 0)
				{
					cli.currentTarget = INTERCOM_ADDRESS_MEDIA;
					echoTargetColor(cli.currentTarget);
					ConsolePrint("New target: MEDIA\r\n");
					continue;
				}

				// check current target
				if(cli.currentTarget == INTERCOM_ADDRESS_MAIN)
				{
					// process onboard
					result = FifoPacketSyncedBegin(&console.rxFifo);
					if(result)
						continue;

					result = FifoPacketSyncedPush(&console.rxFifo, length, cli.procBuf);
					if(result)
					{
						FifoPacketSyncedCancel(&console.rxFifo);
						continue;
					}

					FifoPacketSyncedComplete(&console.rxFifo);
				}
				else
				{
					// forward to intercom
					IntercomHeader header;
					header.address = cli.currentTarget;
					header.command = 0;
					header.section = INTERCOM_SECTION_CONSOLE;

					IntercomSend(&intercom1, &header, length, cli.procBuf);
				}
			}

			// check data from main console to terminal
			while(FifoPacketSyncedNumPackets(&console.txFifo) > 0)
			{
				echoTargetColor(INTERCOM_ADDRESS_MAIN);

				result = FifoPacketSyncedPop(&console.txFifo, &length, cli.procBuf, CLI_PROC_SIZE);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&console.txFifo);
					continue;
				}

				if(result)
					continue;

				FifoSyncedPush(&usart1.txFifo, length, cli.procBuf);

				echoTargetColor(cli.currentTarget);

				USART1StartTx();
			}
		}

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(xLastWakeTime - xLastLEDTime > 500/portTICK_RATE_MS)
		{
			xLastLEDTime = xLastWakeTime;

			LEDToggle(LED_MAIN_RED);
		}

		vTaskDelayUntil(&xLastWakeTime, 50/portTICK_RATE_MS);
	}
}

static void sensorsCommands()
{
}

static void miscCommands()
{
	if(ConsoleCmpCmd("reset"))
	{
		BootReset();
	}

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	float f = 21.0f;
	if(ConsoleScanCmd("float %f", &f) == 1)
	{
		ConsolePrint("Float: %f\r\n", f);
	}

	if(ConsoleCmpCmd("float"))
	{
		ConsolePrint("Float: %f\r\n", 42.0f);
	}

	if(ConsoleCmpCmd("timeout"))
	{
		ConsolePrint("Timeout state: %d\r\n", network.enableTimeout);
	}

	if(ConsoleCmpCmd("usart2 stats"))
	{
		ConsolePrint("--- USART2 ---\r\n");
		ConsolePrint("Rx invalid: %d\r\n", usart2.stats.rxInvalid);
	}

	if(ConsoleCmpCmd("led get"))
	{
		ConsolePrint("LED State: 0x%X\r\n", LEDGetState());
	}

	if(ConsoleCmpCmd("task list"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tState\tPrio\tStack\tID");

		ConsoleLockPrintBuf();
		vTaskList((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("task stats"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tTicks\t\tRatio");

		ConsoleLockPrintBuf();
		vTaskGetRunTimeStats((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	uint8_t reg;
	if(ConsoleScanCmd("nrf read %hhu", &reg) == 1)
	{
		uint8_t value;
		NRF24ReadRegister(&network.nrf24, reg, &value);
		ConsolePrint("Reg 0x%X: 0x%X\r\n", reg, value);
	}

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("-- TX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.txBytesSent, network.nrf24.stats.txBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.txPacketsSent, network.nrf24.stats.txPacketsLost);
		ConsolePrint("Retrans: %10u\r\n",network.nrf24.stats.retransmissions);
		ConsolePrint("-- RX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.rxBytesRecv, network.nrf24.stats.rxBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.rxPacketsRecv, network.nrf24.stats.rxPacketsLost);
	}

	if(ConsoleCmpCmd("float16"))
	{
		float x = 1.0f;
		float16_t f16 = floatToHalfFloat(x);

		float y = halfFloatToFloat(f16);

		ConsolePrint("In: %f, F16: 0x%04X, out: %f\r\n", x, f16, y);
	}

	if(ConsoleCmpCmd("net hello"))
	{
		ConsolePrint("Saying hello to network print\r\n");
		NetworkPrint("Hello World");
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("ID: %hhu\r\n", config.botId);
		ConsolePrint("Channel: %hhu\r\n", config.channel);
		if(config.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(config.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(config.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
		ConsolePrint("Gyro scale: %f\r\n", config.gyroScale);
	}
}

static void buzzerCommands()
{
	uint32_t u32;

	if(ConsoleCmpCmd("buzz on"))
	{
		ConsolePrint("Enabling buzzer\r\n");

		BuzzerEnable(1);
	}

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerEnable(0);
	}

	if(ConsoleScanCmd("buzz freq %d", &u32))
	{
		ConsolePrint("Setting buzzer frequency to: %dHz\r\n", u32);

		BuzzerSetFrequency(u32);
	}
}

static void processCommand()
{
	sensorsCommands();
	miscCommands();
	buzzerCommands();

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Version: 0.0.0 - dev\r\n");
	}
}
