/*
 * boot.c
 *
 *  Created on: 21.04.2013
 *      Author: AndreR
 */

#include "boot.h"

#include "stm32f4xx_iwdg.h"
#include "stm32f4xx_flash.h"

#include "intercom_constants.h"

#include "FreeRTOS.h"
#include "task.h"

#define BOOT_MAGIC 0xA5E3C6F1

volatile uint32_t bootCode __attribute__ ((section (".boot")));

FLASH_Status waitForLastOperation()
{
	__IO FLASH_Status status = FLASH_COMPLETE;

	/* Check for the FLASH Status */
	status = FLASH_GetStatus();

	/* Wait for the FLASH operation to complete by polling on BUSY flag to be reset.
     Even if the FLASH operation fails, the BUSY flag will be reset and an error
     flag will be set */
	while(status == FLASH_BUSY)
	{
		status = FLASH_GetStatus();
		vTaskDelay(2/portTICK_RATE_MS);
	}

	/* Return the operation status */
	return status;
}

#define SECTOR_MASK ((uint32_t)0xFFFFFF07)

static FLASH_Status eraseSector(uint32_t sector)
{
	uint32_t tmp_psize = 0x0;
	FLASH_Status status = FLASH_COMPLETE;

	tmp_psize = FLASH_PSIZE_WORD;
	/* Wait for last operation to be completed */
	status = waitForLastOperation();

	if(status == FLASH_COMPLETE)
	{
		/* if the previous operation is completed, proceed to erase the sector */
		FLASH->CR &= CR_PSIZE_MASK;
		FLASH->CR |= tmp_psize;
		FLASH->CR &= SECTOR_MASK;
		FLASH->CR |= FLASH_CR_SER | sector;
		FLASH->CR |= FLASH_CR_STRT;

		/* Wait for last operation to be completed */
		status = waitForLastOperation();

		/* if the erase operation is completed, disable the SER Bit */
		FLASH->CR &= (~FLASH_CR_SER);
		FLASH->CR &= SECTOR_MASK;
	}

	/* Return the Erase Status */
	return status;
}

int16_t BootEraseProgramFlash()
{
	FLASH_Unlock();

	// Clear pending flags (if any)
	FLASH_ClearFlag(FLASH_FLAG_EOP | FLASH_FLAG_OPERR | FLASH_FLAG_WRPERR |	FLASH_FLAG_PGAERR | FLASH_FLAG_PGPERR|FLASH_FLAG_PGSERR);

	if(eraseSector(FLASH_Sector_5) != FLASH_COMPLETE)
		return 1;
	if(eraseSector(FLASH_Sector_6) != FLASH_COMPLETE)
		return 1;
	if(eraseSector(FLASH_Sector_7) != FLASH_COMPLETE)
		return 1;

	FLASH_Lock();

	return 0;
}

void BootProgramFlash(uint32_t* pData, uint32_t length, uint32_t offset)
{
	FLASH_Unlock();

	// Clear pending flags (if any)
	FLASH_ClearFlag(FLASH_FLAG_EOP | FLASH_FLAG_OPERR | FLASH_FLAG_WRPERR |	FLASH_FLAG_PGAERR | FLASH_FLAG_PGPERR|FLASH_FLAG_PGSERR);

	uint32_t flashAddr = 0x08020000 + offset;

	for(uint32_t i = 0; i < length; i++)
	{
		FLASH_ProgramWord(flashAddr, pData[i]);
		flashAddr += 4;
	}

	FLASH_Lock();
}

uint32_t BootIsBootloaderSelected()
{
	if(bootCode == BOOT_MAGIC)
		return 1;

	return 0;
}

uint32_t BootGetBootCode()
{
	return bootCode;
}

void BootSetBootloaderSelected(uint8_t enable)
{
	if(enable)
		bootCode = BOOT_MAGIC;
	else
		bootCode = 0;
}

typedef void (*pFunction)(void);

void BootJumpToApplication()
{
	volatile uint32_t JumpAddress;
	pFunction Jump_To_Application;

	__set_MSP(*(volatile uint32_t*) (0x8020000));
	JumpAddress = *(volatile uint32_t*) (0x8020000 + 4);
	Jump_To_Application = (pFunction) JumpAddress;
	Jump_To_Application();
}

void BootReset()
{
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_4);
	IWDG_SetReload(2);
	IWDG_ReloadCounter();
	IWDG_Enable();
//	while(1);
}

void BootStartBootloader(Intercom* pIntercom)
{
	BootSetBootloaderSelected(1);

	IntercomHeader iHeader;
	iHeader.section = INTERCOM_SECTION_BOOTLOADER;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.command = 0;

	// Media processor will trigger reset after receiving this command
	IntercomSend(pIntercom, &iHeader, 0, 0);
}
