/*
 * fifo_packet_synced.c
 *
 *  Created on: 13.02.2011
 *      Author: AndreR
 */

#include "fifo_packet_synced.h"

#include "FreeRTOS.h"
#include "semphr.h"

typedef struct __AtomicLockInfo
{
	uint8_t restoreState;
	uint8_t run;
	FifoPacketSynced* pFifo;
} _AtomicLockInfo;

static __inline__ void __iRelease(_AtomicLockInfo* pLock)
{
	(*pLock->pFifo->releaseCallback)(pLock->restoreState, pLock->pFifo->syncData);

	__asm__ volatile("" ::: "memory");
}

static __inline__ _AtomicLockInfo __iLock(FifoPacketSynced* pFifo)
{
	_AtomicLockInfo lock;
	lock.run = 1;
	lock.pFifo = pFifo;
	lock.restoreState = (*pFifo->lockCallback)(pFifo->syncData);

	return lock;
}

#define FIFO_CRITICAL_SECTION(pFifo) \
for(_AtomicLockInfo __info __attribute((__cleanup__(__iRelease))) = __iLock(pFifo); __info.run; __info.run = 0)

void FifoPacketSyncedInit(FifoPacketSynced* pFifo, uint8_t* pBuf, uint16_t size, FifoPacketSyncedLockCallback lockCb, FifoPacketSyncedReleaseCallback releaseCb, void* pSyncData)
{
	pFifo->syncData = pSyncData;
	pFifo->lockCallback = lockCb;
	pFifo->releaseCallback = releaseCb;
	pFifo->numPackets = 0;
	pFifo->currentLength = 0;
	pFifo->pLengthPosition = 0;
	FifoInit(&pFifo->fifo, pBuf, size);
}

void FifoPacketSyncedClear(FifoPacketSynced* pFifo)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		FifoClear(&pFifo->fifo);
	}

	pFifo->numPackets = 0;
}

int16_t FifoPacketSyncedBegin(FifoPacketSynced* pFifo)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(FIFO_PACKET_SYNCED_HEADER_SIZE > pFifo->fifo.EmptyBytes)
			return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

		if(pFifo->pLengthPosition != 0)
			FifoPacketSyncedComplete(pFifo);

		pFifo->currentLength = 0;

		pFifo->pLengthPosition = pFifo->fifo.pWrite;

		FifoPush(&pFifo->fifo, 0);
		FifoPush(&pFifo->fifo, 0);
	}

	return 0;
}

int16_t FifoPacketSyncedPushByte(FifoPacketSynced* pFifo, uint8_t byte)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->fifo.EmptyBytes == 0)
			return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

		if(pFifo->pLengthPosition == 0)
			FifoPacketSyncedBegin(pFifo);

		FifoPush(&pFifo->fifo, byte);

		pFifo->currentLength++;
	}

	return 0;
}

int16_t FifoPacketSyncedComplete(FifoPacketSynced* pFifo)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->pLengthPosition == 0)
			return 0;

		*pFifo->pLengthPosition = (uint8_t)pFifo->currentLength;
		if(pFifo->pLengthPosition == pFifo->fifo.pDataEnd)
			pFifo->pLengthPosition = pFifo->fifo.pData;
		else
			pFifo->pLengthPosition++;
		*pFifo->pLengthPosition = (uint8_t)(pFifo->currentLength >> 8);

		pFifo->pLengthPosition = 0;

		pFifo->numPackets++;
	}

	return 0;
}

void FifoPacketSyncedCancel(FifoPacketSynced* pFifo)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->pLengthPosition == 0)
			return;

		pFifo->fifo.pWrite = pFifo->pLengthPosition;	// reset write pointer
		pFifo->fifo.EmptyBytes += pFifo->currentLength+2;	// release bytes

		pFifo->pLengthPosition = 0;
	}
}

int16_t FifoPacketSyncedPush(FifoPacketSynced* pFifo, uint16_t packetLength, uint8_t* pData)
{
	if(pFifo->fifo.EmptyBytes < packetLength+FIFO_PACKET_SYNCED_HEADER_SIZE)
		return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->pLengthPosition == 0)
			FifoPacketSyncedBegin(pFifo);

		for(uint16_t i = 0; i < packetLength; i++)
		{
			FifoPush(&pFifo->fifo, pData[i]);
		}

		pFifo->currentLength += packetLength;
	}

	return 0;
}

int16_t FifoPacketSyncedPop(FifoPacketSynced* pFifo, uint16_t* pPacketLength, uint8_t* pData, uint16_t suppliedLength)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		FifoPacketSyncedGetNextPacketDataLength(pFifo, pPacketLength);

		if(suppliedLength < *pPacketLength)
			return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

		pFifo->numPackets--;
		FifoPop(&pFifo->fifo, ((uint8_t*)pPacketLength)+0);
		FifoPop(&pFifo->fifo, ((uint8_t*)pPacketLength)+1);
	}

	for(uint16_t i = 0; i < *pPacketLength; i++)
	{
		FIFO_CRITICAL_SECTION(pFifo)
		{
			FifoPop(&pFifo->fifo, pData+i);
		}
	}

	return 0;
}

int16_t FifoPacketSyncedCopyPacket(FifoPacketSynced* pSrcFifo, FifoPacketSynced* pDstFifo)
{
	uint16_t packetLength;
	uint8_t temp;

	FIFO_CRITICAL_SECTION(pSrcFifo)
	{
		if(pSrcFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		FIFO_CRITICAL_SECTION(pDstFifo)
		{
			FifoPacketSyncedGetNextPacketDataLength(pSrcFifo, &packetLength);

			if(pDstFifo->fifo.EmptyBytes < packetLength+FIFO_PACKET_SYNCED_HEADER_SIZE)
				return ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY;

			pSrcFifo->numPackets--;
			FifoPop(&pSrcFifo->fifo, ((uint8_t*)&packetLength)+0);
			FifoPop(&pSrcFifo->fifo, ((uint8_t*)&packetLength)+1);

			if(pDstFifo->pLengthPosition == 0)
				FifoPacketSyncedBegin(pDstFifo);

			for(uint16_t i = 0; i < packetLength; i++)
			{
				FifoPop(&pSrcFifo->fifo, &temp);
				FifoPush(&pDstFifo->fifo, temp);
			}

			pDstFifo->currentLength += packetLength;
		}
	}

	FifoPacketSyncedComplete(pDstFifo);

	return 0;
}

int16_t FifoPacketSyncedCopyPacketData(FifoPacketSynced* pSrcFifo, FifoSynced* pDstFifo)
{
	uint16_t packetLength;
	uint8_t temp;

	FIFO_CRITICAL_SECTION(pSrcFifo)
	{
		if(pSrcFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		FifoPacketSyncedGetNextPacketDataLength(pSrcFifo, &packetLength);

		FIFO_CRITICAL_SECTION(pDstFifo)
		{
			if(pDstFifo->EmptyBytes < packetLength)
				return ERROR_FIFO_SYNCED_NOT_ENOUGH_MEMORY;
		}

		pSrcFifo->numPackets--;
		FifoPop(&pSrcFifo->fifo, ((uint8_t*)&packetLength)+0);
		FifoPop(&pSrcFifo->fifo, ((uint8_t*)&packetLength)+1);

		for(uint16_t i = 0; i < packetLength; i++)
		{
			FifoPop(&pSrcFifo->fifo, &temp);
			FifoSyncedPushByte(&pDstFifo, temp);
		}
	}

	return 0;
}

int16_t FifoPacketSyncedPeekData(FifoPacketSynced* pFifo, uint8_t* pData, uint16_t suppliedLength)
{
	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		uint16_t nextLen = 0;

		FifoPacketSyncedGetNextPacketDataLength(pFifo, &nextLen);

		if(suppliedLength > nextLen)
			suppliedLength = nextLen;

		//save read state
		volatile uint8_t* pRead = pFifo->fifo.pRead;

		// drop length info
		FifoPop(&pFifo->fifo, pData);
		FifoPop(&pFifo->fifo, pData);

		for(uint16_t i = 0; i < suppliedLength; i++)
		{
			FifoPop(&pFifo->fifo, pData+i);
		}

		// restore state
		pFifo->fifo.pRead = pRead;
		pFifo->fifo.EmptyBytes -= FIFO_PACKET_SYNCED_HEADER_SIZE+suppliedLength;
	}

	return 0;
}

int16_t FifoPacketSyncedDrop(FifoPacketSynced* pFifo)
{
	uint16_t len;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		pFifo->numPackets--;
		FifoPop(&pFifo->fifo, ((uint8_t*)&len)+0);
		FifoPop(&pFifo->fifo, ((uint8_t*)&len)+1);
	}

	uint8_t lost;

	for(uint16_t i = 0; i < len; i++)
	{
		FIFO_CRITICAL_SECTION(pFifo)
		{
			FifoPop(&pFifo->fifo, &lost);
		}
	}


	return 0;
}

int16_t FifoPacketSyncedPopHalfWordNoInc(FifoPacketSynced* pFifo, volatile uint16_t* pDest, uint16_t frontDrop)
{
	uint16_t len;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		if(pFifo->numPackets == 0)
			return ERROR_FIFO_PACKET_SYNCED_EMPTY;

		pFifo->numPackets--;
		FifoPop(&pFifo->fifo, ((uint8_t*)&len)+0);
		FifoPop(&pFifo->fifo, ((uint8_t*)&len)+1);
	}

	uint16_t buf;

	for(uint16_t i = 0; i < frontDrop; i++)
		FifoPop(&pFifo->fifo, (uint8_t*)&buf);

	len -= frontDrop;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		for(uint16_t i = 0; i < len; i += 2)
		{
			FifoPop(&pFifo->fifo, ((uint8_t*)&buf)+0);
			FifoPop(&pFifo->fifo, ((uint8_t*)&buf)+1);

			*pDest = buf;
		}
	}

	return 0;
}

int16_t FifoPacketSyncedGetNextPacketDataLength(FifoPacketSynced* pFifo, uint16_t* pLength)
{
	if(pFifo->numPackets == 0)
		return ERROR_FIFO_PACKET_SYNCED_EMPTY;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		//save read state
		volatile uint8_t* pRead = pFifo->fifo.pRead;

		FifoPop(&pFifo->fifo, ((uint8_t*)pLength)+0);
		FifoPop(&pFifo->fifo, ((uint8_t*)pLength)+1);

		pFifo->fifo.pRead = pRead;
		pFifo->fifo.EmptyBytes -= FIFO_PACKET_SYNCED_HEADER_SIZE;
	}

	return 0;
}

uint16_t FifoPacketSyncedNumPackets(FifoPacketSynced* pFifo)
{
	return pFifo->numPackets;
}

int16_t FifoPacketSyncedRevert(FifoPacketSynced* pFifo)
{
	if(pFifo->currentLength == 0)
		return ERROR_FIFO_PACKET_SYNCED_EMPTY;

	FIFO_CRITICAL_SECTION(pFifo)
	{
		pFifo->fifo.EmptyBytes--;
		pFifo->fifo.pWrite--;
		if(pFifo->fifo.pWrite < pFifo->fifo.pData)
			pFifo->fifo.pWrite = pFifo->fifo.pDataEnd-1;

		pFifo->currentLength--;
	}

	return 0;
}

// Note: use recursive mutices for this
uint8_t FifoPacketSyncedMutexLock(void* pData)
{
	xSemaphoreHandle hMutex = pData;

	xSemaphoreTakeRecursive(hMutex, portMAX_DELAY);

	return 0;
}

void FifoPacketSyncedMutexRelease(uint8_t state, void* pData)
{
	state = state;

	xSemaphoreHandle hMutex = pData;

	xSemaphoreGiveRecursive(hMutex);
}

