/*
 * usart.c
 *
 *  Created on: 04.11.2012
 *      Author: Andre
 */

#include "usart.h"

#include "critical.h"
#include "bit_banding.h"
#include "dma_gen.h"

#include "misc.h"

#include "task.h"

#include "util/console.h"

// uses DMA_StreamTypeDef
#define DMAClearEn(pDMA)	BBPeriphReset((uint32_t)&pDMA->CR, 0)
#define DMASetEn(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 0)
#define DMAGetEn(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 0)

#define DMAClearTCI(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCI(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCI(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 4)

// uses USART_TypeDef
#define USARTClearEIE(pRegister)	BBPeriphReset((uint32_t)&pRegister->CR3, 0)
#define USARTSetEIE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR3, 0)
#define USARTGetEIE(pRegister)		BBPeriphGet((uint32_t)&pRegister->CR3, 0)

#define USARTClearTCIE(pRegister)	BBPeriphReset((uint32_t)&pRegister->CR1, 6)
#define USARTSetTCIE(pRegister)		BBPeriphSet((uint32_t)&pRegister->CR1, 6)
#define USARTGetTCIE(pRegister)		BBPeriphGet((uint32_t)&pRegister->CR1, 6)

#define TX_STATE_IDLE		1
#define TX_STATE_ACTIVE		2
#define TX_STATE_WAIT_START	4

#define EVENT_TX_DONE			0x01
#define EVENT_TX_ERROR			0x04
#define EVENT_RX_ERROR_ORE		0x08
#define EVENT_RX_ERROR_NE		0x10
#define EVENT_RX_FE				0x20
#define EVENT_TX_START			0x40
#define EVENT_TX_SYNC_FALLING	0x80

#define RX_SYNC_HIGH(pUSART)	(pUSART->data.pRxSyncPort->BSRRL = pUSART->data.rxSyncPin)
#define RX_SYNC_LOW(pUSART)		(pUSART->data.pRxSyncPort->BSRRH = pUSART->data.rxSyncPin)
#define TX_SYNC_STATE(pUSART)	(pUSART->data.pTxSyncPort->IDR & pUSART->data.txSyncPin)

void USARTIRQ(USART* pUSART)
{
	volatile uint8_t sr = pUSART->pRegister->SR;

	uint8_t event = 0;
	int32_t yield = pdFALSE;

	if((sr & USART_SR_TC) && USARTGetTCIE(pUSART->pRegister))	// transfer complete
	{
		pUSART->pRegister->SR &= ~USART_SR_TC;

		USARTClearTCIE(pUSART->pRegister);

		// transfer complete, disable transmitter
		DMAClearEn(pUSART->pTxStream);

		event = EVENT_TX_DONE;
		xQueueSendFromISR(pUSART->hEventQueue, &event, &yield);
	}

	if((sr & (USART_SR_ORE | USART_SR_FE | USART_SR_NE)) && USARTGetEIE(pUSART->pRegister))
	{
		DMAClearEn(pUSART->pRxStream);

		RX_SYNC_HIGH(pUSART);

		// clear interrupt enable
		USARTClearEIE(pUSART->pRegister);

		// read DR clears all those flags
		pUSART->pRegister->SR &= ~(USART_SR_ORE | USART_SR_FE | USART_SR_NE);
		pUSART->pRegister->DR;

		*pUSART->pDMARxIFCR = DMAGenerateMaskTEIF(pUSART->data.rxStream);

		if(sr & USART_SR_ORE)
			event |= EVENT_RX_ERROR_ORE;
		if(sr & USART_SR_FE)
			event |= EVENT_RX_FE;
		if(sr & USART_SR_NE)
			event |= EVENT_RX_ERROR_NE;

		xQueueSendFromISR(pUSART->hEventQueue, &event, &yield);
	}

	if(yield)
		taskYIELD();
}

void USARTDMATx(USART* pUSART)
{
	volatile uint32_t sr = *pUSART->pDMATxISR;

	if(sr & DMAGenerateMaskTCIF(pUSART->data.txStream))
	{
		*pUSART->pDMATxIFCR = DMAGenerateMaskTCIF(pUSART->data.txStream);

		USARTSetTCIE(pUSART->pRegister); // now wait for TCI to complete transfer
	}
}

void USARTDMARx(USART* pUSART)
{
	volatile uint32_t sr = *pUSART->pDMARxISR;

	DMAClearEn(pUSART->pRxStream);

	if(sr & DMAGenerateMaskTCIF(pUSART->data.rxStream))
	{
		*pUSART->pDMARxIFCR = DMAGenerateMaskTCIF(pUSART->data.rxStream);
	}
}

void USARTTxSyncFalling(USART* pUSART)
{
	uint8_t event = EVENT_TX_SYNC_FALLING;
	int32_t yield = pdFALSE;

	xQueueSendFromISR(pUSART->hEventQueue, &event, &yield);

	if(yield)
		taskYIELD();
}

void USARTInit(USART* pUSART, USART_TypeDef* pRegister, USARTData data)
{
	pUSART->data = data;
	pUSART->pRegister = pRegister;
	pUSART->callback = 0;
	pUSART->stats.rxInvalid = 0;
	pUSART->txState = TX_STATE_IDLE;

	// Init sync pins
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_OD;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	gpioInit.GPIO_Pin = data.txSyncPin;
	GPIO_Init(data.pTxSyncPort, &gpioInit);

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Pin = data.rxSyncPin;
	GPIO_Init(data.pRxSyncPort, &gpioInit);

	RX_SYNC_HIGH(pUSART);

	// Init action queue
	pUSART->hEventQueue = xQueueCreate(10, 1);

	// Init mutex
	pUSART->hTxInUse = xSemaphoreCreateRecursiveMutex();
	pUSART->hRxInUse = xSemaphoreCreateRecursiveMutex();

	// Fifo init
	FifoPacketSyncedInit(&pUSART->rxFifo, pUSART->data.pDataRx, pUSART->data.size, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, pUSART->hRxInUse);
	FifoPacketSyncedInit(&pUSART->txFifo, pUSART->data.pDataTx, pUSART->data.size, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, pUSART->hTxInUse);

	// find interrupt flag clear register and generate mask
	pUSART->pRxStream = DMAGenerateStreamTypeDef(data.pDMA, data.rxStream);
	pUSART->pTxStream = DMAGenerateStreamTypeDef(data.pDMA, data.txStream);
	pUSART->pDMARxIFCR = DMAGenerateIFCR(data.pDMA, data.rxStream);
	pUSART->pDMATxIFCR = DMAGenerateIFCR(data.pDMA, data.txStream);
	pUSART->pDMARxISR = DMAGenerateISR(data.pDMA, data.rxStream);
	pUSART->pDMATxISR = DMAGenerateISR(data.pDMA, data.txStream);
	pUSART->IFCRRxMask = DMAGenerateMaskAll(data.rxStream);
	pUSART->IFCRTxMask = DMAGenerateMaskAll(data.txStream);

	// DMA setup
	// TX
	DMA_InitTypeDef dmaInit;
	dmaInit.DMA_Channel = data.dmaChannel;
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&pRegister->DR;
	dmaInit.DMA_Memory0BaseAddr = (uint32_t)data.pDmaDataTx;
	dmaInit.DMA_DIR = DMA_DIR_MemoryToPeripheral;
	dmaInit.DMA_BufferSize = data.dmaSize;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
	dmaInit.DMA_Mode = DMA_Mode_Normal;
	dmaInit.DMA_Priority = DMA_Priority_Medium;
	dmaInit.DMA_FIFOMode = DMA_FIFOMode_Enable;
	dmaInit.DMA_MemoryBurst = DMA_MemoryBurst_Single;
	dmaInit.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
	dmaInit.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
	DMA_Init(pUSART->pTxStream, &dmaInit);

	// DMA RX
	dmaInit.DMA_Memory0BaseAddr = (uint32_t)data.pDmaDataRx;
	dmaInit.DMA_DIR = DMA_DIR_PeripheralToMemory;
	DMA_Init(pUSART->pRxStream, &dmaInit);

	// NVIC setup
	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannel = data.txIRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = data.IRQL;
	nvicInit.NVIC_IRQChannelSubPriority = 0x00;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = data.rxIRQn;
	NVIC_Init(&nvicInit);

	nvicInit.NVIC_IRQChannel = data.usartIRQn;
	NVIC_Init(&nvicInit);

	// USART setup
	pRegister->CR1 |= USART_CR1_OVER8;

	USART_InitTypeDef usartInit;
	usartInit.USART_BaudRate = data.baudrate;
	usartInit.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	usartInit.USART_Mode = 0;
	usartInit.USART_Parity = USART_Parity_No;
	usartInit.USART_StopBits = USART_StopBits_1;
	usartInit.USART_WordLength = USART_WordLength_8b;
	USART_Init(pRegister, &usartInit);

	pRegister->CR3 |= USART_CR3_ONEBIT;

	pRegister->SR = 0; // clear all flags
	USARTSetEIE(pRegister);
	USARTClearTCIE(pRegister);

	// clear DMA flags
	*pUSART->pDMARxIFCR = pUSART->IFCRRxMask;
	*pUSART->pDMATxIFCR = pUSART->IFCRTxMask;

	// finally enable USART
	USART_Cmd(pRegister, ENABLE);

	// enable DMA for TX and RX
	USART_DMACmd(pRegister, USART_DMAReq_Rx | USART_DMAReq_Tx, ENABLE);

	DMASetTCI(pUSART->pRxStream);		// enable interrupt on transfer complete
	DMASetTCI(pUSART->pTxStream);

	pUSART->pRegister->CR1 |= USART_CR1_TE;

	// enable RX DMA
	pUSART->pRxStream->NDTR = data.dmaSize;

	// enable receiver
	pRegister->CR1 |= USART_CR1_RE;
	DMASetEn(pUSART->pRxStream);		// enable DMA

	RX_SYNC_LOW(pUSART);
}

int16_t USARTSend(USART* pUSART, uint16_t length, uint8_t* pData)
{
	// use semaphore to block concurrent calls from multiple threads
	TASK_MUTEX_RECURSIVE(pUSART->hTxInUse)
	{
		if(pUSART->txFifo.fifo.EmptyBytes < FIFO_PACKET_SYNCED_HEADER_SIZE + length)
		{
			// that data does not fit in our fifo :(
			return ERROR_FIFO_SYNCED_NOT_ENOUGH_MEMORY;
		}

		// enough memory :) push packet
		FifoPacketSyncedBegin(&pUSART->txFifo);
		FifoPacketSyncedPush(&pUSART->txFifo, length, pData);
		FifoPacketSyncedComplete(&pUSART->txFifo);

		uint8_t event = EVENT_TX_START;
		if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
			return ERROR_USART_QUEUE_FULL;
	}

	return 0;
}

int16_t USARTSendConcat(USART* pUSART, uint16_t length, uint8_t* pData, uint16_t length2, uint8_t* pData2)
{
	// use semaphore to block concurrent calls from multiple threads
	TASK_MUTEX_RECURSIVE(pUSART->hTxInUse)
	{
		if(pUSART->txFifo.fifo.EmptyBytes < FIFO_PACKET_SYNCED_HEADER_SIZE + length + length2)
		{
			// that data does not fit in our fifo :(
			return ERROR_FIFO_SYNCED_NOT_ENOUGH_MEMORY;
		}

		// enough memory :) push packet
		FifoPacketSyncedBegin(&pUSART->txFifo);
		FifoPacketSyncedPush(&pUSART->txFifo, length, pData);
		FifoPacketSyncedPush(&pUSART->txFifo, length2, pData2);
		FifoPacketSyncedComplete(&pUSART->txFifo);

		uint8_t event = EVENT_TX_START;
		if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
			return ERROR_USART_QUEUE_FULL;
	}

	return 0;
}

int16_t USARTStartTransfer(USART* pUSART)
{
	if(pUSART->txState != TX_STATE_IDLE) // already transmitting?
		return 0;

	uint8_t event = EVENT_TX_START;
	if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
		return ERROR_USART_QUEUE_FULL;

	return 0;
}

void USARTTask(void* params)
{
	USART* pUSART = (USART*)params;

	uint8_t event;
	int32_t osResult;

	DMAClearEn(pUSART->pRxStream);
	pUSART->pRegister->CR1 &= ~USART_CR1_RE;
	pUSART->pRegister->SR &= ~(USART_SR_ORE | USART_SR_FE | USART_SR_NE);
	pUSART->pRegister->DR;

	xQueueReset(pUSART->hEventQueue);

	RX_SYNC_LOW(pUSART);

	pUSART->pRxStream->NDTR = pUSART->data.dmaSize;
	DMASetEn(pUSART->pRxStream);
	pUSART->pRegister->CR1 |= USART_CR1_RE;
	USARTSetEIE(pUSART->pRegister);

	uint32_t txTime = xTaskGetTickCount();

	while(42)
	{
		event = 0;
		osResult = xQueueReceive(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS);

		if(osResult != pdTRUE && (pUSART->txState == TX_STATE_ACTIVE) && xTaskGetTickCount()-txTime > 500/portTICK_RATE_MS)
		{
			//disable transmitter
			DMAClearEn(pUSART->pTxStream);
			pUSART->pRegister->SR &= ~USART_SR_TC;

			pUSART->txState = TX_STATE_IDLE;

			ConsolePrint("TX timeout\r\n");

			event = EVENT_TX_START;
			if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
				ConsolePrint("4: queue full\r\n");

			continue;
		}

		if(osResult != pdTRUE)
			continue;

		if(event == EVENT_RX_FE && DMAGetEn(pUSART->pRxStream))
		{
			ConsolePrint("Warning: DMA RX EN set + RX done\r\n");
		}

		if(event == EVENT_RX_FE)
		{
			uint16_t dataLength = pUSART->data.dmaSize-pUSART->pRxStream->NDTR-1; // -1 because errornous break frame is saved too

			// check for enough space in fifo
			if(pUSART->rxFifo.fifo.EmptyBytes < FIFO_PACKET_SYNCED_HEADER_SIZE + dataLength)
			{
				// that does not fit :(
				++pUSART->stats.rxInvalid;

				ConsolePrint("Packet too big (RX data) %d\r\n", dataLength);
			}
			else
			{
				// enough space, save data
				FifoPacketSyncedBegin(&pUSART->rxFifo);
				FifoPacketSyncedPush(&pUSART->rxFifo, dataLength, pUSART->data.pDmaDataRx);
				FifoPacketSyncedComplete(&pUSART->rxFifo);

				if(pUSART->callback)
					(*pUSART->callback)(pUSART->pUserData);
			}

			// clear flags
			pUSART->pRegister->SR &= ~(USART_SR_ORE | USART_SR_FE | USART_SR_NE);
			pUSART->pRegister->DR;
			USARTSetEIE(pUSART->pRegister);

			// prepare new RX
			pUSART->pRxStream->NDTR = pUSART->data.dmaSize;
			DMASetEn(pUSART->pRxStream);

			RX_SYNC_LOW(pUSART);

			continue;
		}

		if((event & EVENT_RX_ERROR_NE) || (event & EVENT_RX_ERROR_ORE))
		{
			ConsolePrint("RXE - Tx: 0x%X, EV: 0x%X\r\n", pUSART->txState, event);

			// clear flags
			pUSART->pRegister->SR &= ~(USART_SR_ORE | USART_SR_FE | USART_SR_NE);
			pUSART->pRegister->DR;
			USARTSetEIE(pUSART->pRegister);

			pUSART->pRxStream->NDTR = pUSART->data.dmaSize;
			DMASetEn(pUSART->pRxStream);

			RX_SYNC_LOW(pUSART);

			continue;
		}

		if(event == EVENT_TX_ERROR)
		{
			ConsolePrint("TXE - Tx: 0x%X, EV: 0x%X\r\n", pUSART->txState, event);

			continue;
		}


		if(pUSART->txState == TX_STATE_IDLE && event == EVENT_TX_START)
		{
			uint16_t length;
			int16_t result;

			pUSART->txState = TX_STATE_ACTIVE;

			result = FifoPacketSyncedPop(&pUSART->txFifo, &length, pUSART->data.pDmaDataTx, pUSART->data.dmaSize);
			if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
			{
				// packet too big
				FifoPacketSyncedDrop(&pUSART->txFifo);

				pUSART->txState = TX_STATE_IDLE;

				if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
					ConsolePrint("1: queue full\r\n");

				ConsolePrint("Packet too big\r\n");
			}

			if(result)
			{
				pUSART->txState = TX_STATE_IDLE;

				continue;
			}

			txTime = xTaskGetTickCount();

			pUSART->pTxStream->NDTR = length;
			DMASetEn(pUSART->pTxStream);	// enable DMA

			continue;
		}

		if(pUSART->txState == TX_STATE_ACTIVE && event == EVENT_TX_DONE)
		{
			pUSART->pRegister->CR1 |= USART_CR1_SBK; // send break

			continue;
		}

		if(pUSART->txState == TX_STATE_WAIT_START && event == EVENT_TX_SYNC_FALLING)
		{
			event = EVENT_TX_START;
			if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
				ConsolePrint("0: queue full\r\n");

			continue;
		}

		if(pUSART->txState == TX_STATE_ACTIVE && event == EVENT_TX_SYNC_FALLING)
		{
			pUSART->txState = TX_STATE_IDLE;

			event = EVENT_TX_START;
			if(xQueueSend(pUSART->hEventQueue, &event, 500/portTICK_RATE_MS) != pdTRUE)
				ConsolePrint("3: queue full\r\n");

			continue;
		}
	}
}
