/*
 * main.c
 *
 *  Created on: 17.03.2011
 *      Author: AndreR
 *
 * Pin Configuration
 *
 * PB1: CHG / OC1A
 * PB2: CS
 * PB3: MOSI
 * PB4: MISO
 * PB5: SCK
 *
 * PC1: T IGBT 1 / ADC1
 * PC2: CHG-ADC / ADC2
 * PC3: CAP-ADC / ADC3
 * PC4: T IGBT 2 / ADC4
 *
 * PD0: LEDR
 * PD1: LEDG
 * PD3: FAN / OC2B
 * PD5: DIS / OC0B
 * PD6: EN-DIS1
 * PD7: EN-DIS2
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "kicker_data.h"
#include "spi.h"
#include "hal.h"

#define STATE_IDLE		0
#define STATE_FIRE		1
#define STATE_CHARGE	2

#define CHARGE_MODE_MANUAL		0
#define CHARGE_MODE_AUTOLOAD	1

#define NUM_INTERVALLS 5

typedef struct PowerIntervall
{
	uint16_t capVoltIntervall;

	uint16_t on;
	uint16_t off;
} PowerIntervall;

PowerIntervall powerIntervalls[NUM_INTERVALLS];

KickerData data;

typedef struct _Channel
{
	uint16_t offset;
	uint16_t avg;
	uint16_t bias;
} Channel;

Channel ch[4];

volatile uint16_t sysTick;
volatile uint16_t chargeTimer;
volatile uint16_t chargeTimerEnd;
volatile uint8_t fireDone;

uint8_t state;
uint8_t chargeMode;

uint16_t maxCapLevel;


ISR(TIMER2_COMPA_vect, ISR_NOBLOCK)	// systick
{
	// 100us, systick
	sysTick++;

	if(chargeTimer != 0xFFFF)
	{
		chargeTimer++;
		if(chargeTimer == chargeTimerEnd)
			DisableCharge();
	}
}

ISR(TIMER0_COMPB_vect, ISR_NOBLOCK)	// DIS1+DIS2 compare match
{
	DisableFire();

	fireDone = 1;
}

int main()
{
	// init ports
	PORTD = (1 << LEDG) | (1 << LEDR);
	DDRD = (1 << LEDG) | (1 << LEDR) | (1 << DIS1) | (1 << DIS2);
	DDRB = (1 << CHG);

	PORTD &= ~(1 << LEDR);

	// init variables
	uint8_t i;

	sysTick = 0;

	state = STATE_IDLE;
	chargeMode = CHARGE_MODE_MANUAL;
	fireDone = 1;

	uint8_t overChargeCounter = 0;

	chargeTimer = 0xFFFF;

	maxCapLevel = 0;

	uint8_t hitMaxCap = 0;

	// init autoload settings
	powerIntervalls[0].on	= 48;
	powerIntervalls[0].off	= 204;
	powerIntervalls[0].capVoltIntervall=0;
	powerIntervalls[1].on	= 48;
	powerIntervalls[1].off	= 204;
	powerIntervalls[1].capVoltIntervall=200;
	powerIntervalls[2].on	= 102;	// K:  6, S: 12
	powerIntervalls[2].off	= 204;	// K: 36, S: 66
	powerIntervalls[2].capVoltIntervall=500;
	powerIntervalls[3].on	= 168;	// K:  6, S: 12
	powerIntervalls[3].off	= 102;	// K: 12, S: 30
	powerIntervalls[3].capVoltIntervall=2500;
	powerIntervalls[4].on	= 48;	// K: 6, S: 12
	powerIntervalls[4].off	= 204;	// K: 6, S: 18
	powerIntervalls[4].capVoltIntervall=5000;	// safety interval

	powerIntervalls[0].on	= 48;
	powerIntervalls[0].off	= 204;
	powerIntervalls[0].capVoltIntervall=0;
	powerIntervalls[1].on	= 24;
	powerIntervalls[1].off	= 102;
	powerIntervalls[1].capVoltIntervall=200;
	powerIntervalls[2].on	= 51;	// K:  6, S: 12
	powerIntervalls[2].off	= 102;	// K: 36, S: 66
	powerIntervalls[2].capVoltIntervall=500;
	powerIntervalls[3].on	= 84;	// K:  6, S: 12
	powerIntervalls[3].off	= 84;	// K: 12, S: 30
	powerIntervalls[3].capVoltIntervall=2500;
	powerIntervalls[4].on	= 48;	// K: 6, S: 12
	powerIntervalls[4].off	= 204;	// K: 6, S: 18
	powerIntervalls[4].capVoltIntervall=5000;	// safety interval

	uint16_t maxChargeCurrent = 10000;	// 10A

	// init SPI
	SPIInit();

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1);	// Enable, prescaler 64 = 187.5kHz
	DIDR0 = 0x3F;

	// init Timer 0 (DIS1+DIS2)
	TCCR0A = 0;	// normal mode
	TCCR0B = 0;
	TCNT0 = 0;
	OCR0B = 100;
	TIMSK0 = (1 << OCIE0B);

	// init Timer 2 (systick)
	TCCR2A = (1 << WGM21);	// CTC mode, OCR2A is TOP
	TCNT2 = 0;	// clear counter
	OCR2A = 150;	// with prescaler 8 makes a 10kHz PWM
	TIMSK2 = (1 << OCIE2A);	// enable output compare interrupt (systick 100us)
	TCCR2B |= (1 << CS21);	// enable

	// init Timer 1 (CHG)
	TCCR1B = (1 << WGM13);
	TCNT1 = 0;
	OCR1A = 0;
	ICR1 = 12;	// determines period (T), in clock cycles (@12Mhz)
	TCCR1B |= (1 << CS10);	// enable
	TCCR1A = (1 << COM1A1); // non-inverting Phase and frequency correct PWM, TOP = ICR1

	// init data
	data.cmd.type = 0;
	data.uCap = 0;
	data.iChg = 0;

	for(i = 0; i < 4; i++)
	{
		ch[i].bias = 3;
		ch[i].offset = 0;
		ch[i].avg = 0;
	}

	uint16_t meas[4];

	PORTD |= (1 << LEDR);

	sei();

	while(1)
	{
		// --- Update shared data ---
		SPIExchange(data.exchange.miso, data.exchange.mosi, KICKER_DATA_SIZE);

		// --- Command processing ---
		if(data.cmd.type == KICKER_CMD_FIRE && state != STATE_FIRE && data.cmd.fire.duration.value != 0)
		{
			state = STATE_FIRE;
			DisableCharge();

			fireDone = 0;

			if(data.cmd.fire.device == KICKER_DEVICE_STRAIGHT)
			{
				EnableFire1(data.cmd.fire.duration);
			}

			if(data.cmd.fire.device == KICKER_DEVICE_CHIP)
			{
				EnableFire2(data.cmd.fire.duration);
			}

//			chargeMode = CHARGE_MODE_MANUAL; // testing safety
		}

		if(data.cmd.type == KICKER_CMD_LOAD_AUTO)
		{
			maxCapLevel = data.cmd.loadAuto.maxCapLevel;
			if(maxCapLevel > 2350)	// 235V
				maxCapLevel = 2350;

			if(maxCapLevel < 10)
			{
				maxCapLevel = 10;
				chargeMode = CHARGE_MODE_MANUAL;
			}
			else
			{
				chargeMode = CHARGE_MODE_AUTOLOAD;
			}
		}

		if(data.cmd.type == KICKER_CMD_LOAD_MANUAL && state != STATE_FIRE)
		{
			chargeMode = CHARGE_MODE_MANUAL;

			chargeTimerEnd = data.cmd.loadManual.duration;
			chargeTimer = 0;

			EnableCharge(data.cmd.loadManual.tOnTicks, data.cmd.loadManual.tOffTicks);
		}

		data.cmd.type = 0;

		// --- State Machine ---
		if(state == STATE_FIRE)
		{
			if(fireDone)
			{
				state = STATE_IDLE;
			}
		}

		if(data.uCap > 2350)
		{
			DisableCharge();
			chargeMode = CHARGE_MODE_MANUAL;
		}

		if(state == STATE_IDLE && chargeMode == CHARGE_MODE_AUTOLOAD)
		{
			// automatic loading
			if(hitMaxCap == 0)
			{
				if(data.uCap <= maxCapLevel)
				{
					for(i = 0; i < NUM_INTERVALLS; i++)
					{
						if(data.uCap <= powerIntervalls[i].capVoltIntervall)
							break;
					}

					EnableCharge(powerIntervalls[i].on, powerIntervalls[i].off);
				}
				else
				{
					hitMaxCap = 1;
					DisableCharge();
				}
			}
			else
			{
				// hit max cap, wait 5V
				if(data.uCap <= (maxCapLevel - 50))
				{
					hitMaxCap = 0;
				}
			}
		}

		// --- ADC Sampling ---
		for(i = 0; i < 4; i++)
		{
			ADMUX = i+1;
			ADCSRA |= (1 << ADSC);
			while(ADCSRA & (1 << ADSC));	// ~78us
			meas[i] = ADC;
		}

		// Exponential Moving Average on all channels
		for(i = 0; i < 4; i++)
		{
			ch[i].avg = ((ch[i].avg*ch[i].bias) + (meas[i]*(10-ch[i].bias)))/10;
		}

		uint32_t val;

		// capacitor voltage conversion
		val = ch[2].avg;	// CAP
		val *= 20; // => [Ve-4]
		val *= 123;	// [Ve-4] => [Ve-7], Resistor divider gain = 123
		val /= 1000;	// [Ve-7] => [Ve-4]
		data.uCap = val;

		// charge current conversion
		data.iChg = ch[1].avg*20; // => [Ve-4] => [mA]

		// charge current limiter
		if(data.iChg > maxChargeCurrent)
		{
			overChargeCounter++;
//			DisableFire();
//			fireDone = 1;
			DisableCharge();

			if(overChargeCounter > 100)
			{
				overChargeCounter = 0;
				chargeMode = CHARGE_MODE_MANUAL;
			}
		}
		else
		{
			overChargeCounter = 0;
		}

		// --- Sys LED (green) ---
		if(sysTick >= 10000)
		{
			sysTick = 0;
			PORTD ^= (1 << LEDG);
		}
	}

	return 0;
}

