/*	Pin Configuration
 *
 *	PB0: M4-FLT
 *	PB1: M1-FLT
 *	PB2: CS
 *	PB3: MOSI
 *	PB4: MISO
 *	PB5: SCK
 *
 *	PC0: M4-AOUT
 *	PC1: M1-AOUT
 *	PC2: M2-AOUT
 *	PC3: M5-AOUT
 *	PC4: M3-AOUT
 *
 *	PD0: M5-FLT
 *	PD1: M2-FLT
 *	PD2: M3-FLT
 *	PD5: LED red
 *	PD6: LED green
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "motor_data.h"
#include "spi.h"

#define M1F PB1
#define M2F PD1
#define M3F PD2
#define M4F PB0
#define M5F PD0

#define M1A PC1
#define M2A PC3
#define M3A PC4
#define M4A PC0
#define M5A PC2
#define BDA	PC5

#define LEDR PD5
#define LEDG PD6

#define LEDIR PD3
#define BDIRQ PD4

#define IR_ON()		(PORTD |= (1 << LEDIR))
#define IR_OFF()	(PORTD &= ~(1 << LEDIR))
#define IRQ_ON()	(PORTD |= (1 << BDIRQ))
#define IRQ_OFF()	(PORTD &= ~(1 << BDIRQ))
#define LEDR_ON()	(PORTD &= ~(1 << LEDR))
#define LEDR_OFF()	(PORTD |= (1 << LEDR))


MotorData data;
register uint8_t sysTick asm("r5");

typedef struct _Channel
{
	uint16_t offset;
	uint16_t avg;
	uint16_t bias;
} Channel;

struct _MotorInternal
{
	uint16_t irLow;
	uint16_t irHigh;
	Channel ch[7];
} internal;

ISR(TIMER2_COMPA_vect, ISR_NAKED)
{
	// 20ms, systick, assembler style for fast ISR
	// sysTick++;
	asm volatile(
		"sei				\n\t"
		"push r0			\n\t"
		"in r0, __SREG__	\n\t"
		"inc r5				\n\t"
		"out __SREG__, r0	\n\t"
		"pop r0				\n\t"
		"reti				\n\t"
			::);
}

static void motorCalibration();

static uint16_t measure(uint8_t channel)
{
	ADMUX = channel;
	ADCSRA |= (1 << ADSC);
	while(ADCSRA & (1 << ADSC));

	return ADC;
}

int main()
{
	uint8_t i;

	sysTick = 0;

	// init ports
	PORTB = (1 << M1F) | (1 << M4F);
	DDRD = (1 << LEDR) | (1 << LEDG) | (1 << LEDIR) | (1 << BDIRQ);
	PORTD = (1 << LEDR) | (1 << LEDG) | (1 << M2F) | (1 << M3F) | (1 << M5F);

	LEDR_ON();

	// init SPI
	SPIInit();

	// init systick (Timer 2)
	TCCR2A = (1 << WGM21);	// CTC mode, OCR2A is TOP
	TCNT2 = 0;	// clear counter
	OCR2A = 234;	// with prescaler 1024 makes approx. 50Hz
	TIMSK2 = (1 << OCIE2A);	// enable output compare interrupt (systick 20ms)
	TCCR2B |= (1 << CS21) | (1 << CS22) | (1 << CS20);	// enable

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1);	// Enable, prescaler 64 = 187.5kHz
	DIDR0 = 0x3F;

	// init data
	for(i = 0; i < 5; i++)
		data.current[i] = 0;
	data.fault = 0;

	for(i = 0; i < 7; i++)
	{
		internal.ch[i].bias = 3;
		internal.ch[i].offset = 0;
		internal.ch[i].avg = 0;
	}

	uint16_t ch[7];

	motorCalibration();

	data.irThreshold = 200;

	sei();

	LEDR_OFF();
	IR_ON();

	uint8_t irMeasOn = 1;

	while(1)
	{
		SPIExchange(data.exchange.miso, data.exchange.mosi, MOTOR_DATA_SIZE);

		if(data.cmd == MOTOR_CMD_SET_THRESHOLD)
			data.irThreshold = data.setThreshold;

		data.cmd = 0;

		// Gather new samples
		if(irMeasOn)
		{
			irMeasOn = 0;
			ch[5] = measure(i);
			IR_OFF();
		}
		else
		{
			irMeasOn = 1;
			ch[6] = measure(i);
			IR_ON();
		}

		for(i = 0; i < 5; i++)
		{
			ch[i] = measure(i);

			if(ch[i] < internal.ch[i].offset)
				ch[i] = 0;
			else
				ch[i] -= internal.ch[i].offset;
		}

		// Exponential Moving Average on all channels
		for(i = 0; i < 7; i++)
		{
			Channel* pCh = &internal.ch[i];

			pCh->avg = ((pCh->avg*pCh->bias) + (ch[i]*(10-pCh->bias)))/10;
		}

		// Motor processing
		for(i = 0; i < 5; i++)
		{
			// val * 2 = [mV],  * 4 = [mA] over shunt
			ch[i] = internal.ch[i].avg*8;
		}

		// IR processing
		uint16_t irOn = ch[5]*20;	// => [Ve-4]
		uint16_t irOff = ch[6]*20;	// => [Ve-4]

		if(irOff > irOn)
			data.irLevel = 0;
		else
			data.irLevel = irOn - irOff;

		// testing
//		data.irLevel = irOn;
//		data.irThreshold = irOff;

//		data.irLevel = ch[5]*20;	// => [Ve-4]

		// Motor channel reorder
		data.current[0] = ch[M1A];
		data.current[1] = ch[M2A];
		data.current[2] = ch[M3A];
		data.current[3] = ch[M4A];
		data.current[4] = ch[M5A];

		data.fault = 0;

		if(PINB & (1 << M1F))
			data.fault |= 0x01;

		if(PIND & (1 << M2F))
			data.fault |= 0x02;

		if(PIND & (1 << M3F))
			data.fault |= 0x04;

		if(PINB & (1 << M4F))
			data.fault |= 0x08;

		if(PIND & (1 << M5F))
			data.fault |= 0x10;

		// --- Check IR level ---
		if(data.irLevel < data.irThreshold)
		{
			LEDR_ON();
			IRQ_ON();
		}
		else
		{
			LEDR_OFF();
			IRQ_OFF();
		}

		// --- Sys LED (green) ---
		if(sysTick >= 25)
		{
			sysTick = 0;
			PORTD ^= (1 << LEDG);
		}
	}

	return 0;
}

static void motorCalibration()
{
	uint8_t i;

	// throw away first sample
	for(i = 0; i < 5; i++)
	{
		measure(i);
		internal.ch[i].offset = 0;
	}

	uint8_t iterations;
	for(iterations = 0; iterations < 10; iterations++)
	{
		for(i = 0; i < 5; i++)
		{
			internal.ch[i].offset += measure(i);
		}
	}

	for(i = 0; i < 5; i++)
		internal.ch[i].offset /= iterations;
}
