/*
 * main.c
 *
 *  Created on: 17.03.2011
 *      Author: AndreR
 *
 * Pin Configuration
 *
 * PB0: LED9G
 * PB1: LED10G
 * PB2: CS
 * PB3: MOSI
 * PB4: MISO
 * PB5: SCK
 *
 * PC0: 3V3-A / ADC0
 * PC1: 5V-A / ADC1
 * PC2: C1-A / ADC2
 * PC3: C2-A / ADC3
 * PC4: C3-A / ADC4
 * PC5: C4-A / ADC5
 * ADC7: VCC-A / ADC7
 *
 * PD0: LED1R
 * PD1: LED2R
 * PD2: LED3O
 * PD3: LED4O
 * PD4: LED5Y
 * PD5: LED6Y
 * PD6: LED7G
 * PD7: LED8G
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "power_data.h"
#include "spi.h"

#define LEDG PD3

PowerData data;

ISR(TIMER2_COMPA_vect, ISR_NAKED)
{
	// 20ms, systick, assembler style for fast ISR
	// sysTick++;
	asm volatile(
		"sei				\n\t"
		"push r0			\n\t"
		"in r0, __SREG__	\n\t"
		"inc r5				\n\t"
		"out __SREG__, r0	\n\t"
		"pop r0				\n\t"
		"reti				\n\t"
			::);
}

int main()
{
	uint8_t i;

	sysTick = 0;

	// init ports
	DDRD = (1 << LEDG);

	PORTD = (1 << LEDG);

	// init spi
	SPIInit();

	// init systick (Timer 2)
	TCCR2A = (1 << WGM21);	// CTC mode, OCR2A is TOP
	TCNT2 = 0;	// clear counter
	OCR2A = 234;	// with prescaler 1024 makes approx. 50Hz
	TIMSK2 = (1 << OCIE2A);	// enable output compare interrupt (systick 20ms)
	TCCR2B |= (1 << CS21) | (1 << CS22) | (1 << CS20);	// enable

	// init ADC
	ADCSRA = (1 << ADEN) | (1 << ADIF) | (1 << ADPS2) | (1 << ADPS1);	// Enable, prescaler 64 = 187.5kHz
	DIDR0 = 0x3F;

	// init data
	for(i = 0; i < 4; i++)
		data.uCell[i] = 4005;
	data.i3V3 = 0;
	data.i5V = 0;
	data.iVCC = 0;

	uint16_t ch[7];

	sei();

	while(1)
	{
		SPIExchange(data.exchange.miso, data.exchange.mosi, POWER_DATA_SIZE);

		for(i = 0; i < 7; i++)
			ch[i] = 0;

		uint8_t iterations;
		for(iterations = 0; iterations < 10; iterations++)
		{
			for(i = 0; i < 6; i++)
			{
				ADMUX = i;
				ADCSRA |= (1 << ADSC);
				while(ADCSRA & (1 << ADSC));
				ch[i] += ADC;
			}

			ADMUX = 7;
			ADCSRA |= (1 << ADSC);
			while(ADCSRA & (1 << ADSC));
			ch[6] += ADC;
		}

		for(i = 0; i < 7; i++)
		{
			ch[i] /= iterations;
			ch[i] *= 20;	// convert to [Ve-4]
		}

		uint32_t val = 0;

		val = ch[5];   		// Cell 1
		val *= 21;      	// [Ve-4] => [Ve-5]
		val /= 100;     	// [Ve-5] => [Ve-3] = [mV]
		data.uCell[0] = val;

		val = ch[4];   		// Cell 2
		val *= 42;
		val /= 100;
		data.uCell[1] = val;

		val = ch[3];    	// Cell 3
		val *= 63;
		val /= 100;
		data.uCell[2] = val;

		val = ch[2];    	// Cell 4
		val *= 84;
		val /= 100;
		data.uCell[3] = val;

		val = ch[0];		// 3V3
		val /= 20;          // Gain
		val *= 100;
		val /= 100;         // [uV]/[mOhm] = [mA]
		data.i3V3 = val;

		val = ch[1];    	// 5V
		val /= 20;          // Gain
		val *= 100;
		val /= 100;         // [uV]/[mOhm] = [mA]
		data.i5V = val;

		val = ch[6];		// VCC
		val *= 100;
		val /= 27;          // Gain
		val /= 2;           // [uV]/[mOhm] = [mA]
		data.iVCC = val;

		data.uCell[3] -= data.uCell[2];
		data.uCell[2] -= data.uCell[1];
		data.uCell[1] -= data.uCell[0];

		// --- Sys LED (green) ---
		if(sysTick >= 25)
		{
			sysTick = 0;
			PORTD ^= (1 << LEDG);
		}
	}

	return 0;
}
