/*
 * ctrl.h
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#ifndef CTRL_H_
#define CTRL_H_

#include "stm32f4xx.h"
#include "arm_math.h"
#include "util/vector.h"

#include "FreeRTOS.h"
#include "semphr.h"
#include "queue.h"
#include "task.h"

#define CTRL_MOTOR_TO_WHEEL_RATIO (15.0f/50.0f)
#define CTRL_WHEEL_TO_MOTOR_RATIO (50.0f/15.0f)
#define CTRL_MOTOR_KM (0.0255f) // motor torque constant [Nm/A]
#define CTRL_MOTOR_KN (374.0f*2.0f*PI/60.0f) // motor speed constant [rad/(s*V)]

#define CTRL_EVENT_RUN			1
#define CTRL_EVENT_UPDATE_MOTOR	2
#define CTRL_EVENT_UPDATE_ENC	3
#define CTRL_EVENT_UPDATE_GYR	4
#define CTRL_EVENT_UPDATE_ACC	5
#define CTRL_EVENT_UPDATE_POS	6
#define CTRL_EVENT_UPDATE_POWER	7
#define CTRL_EVENT_SET_CTRL		8

typedef struct _CtrlInput
{
	struct _motor
	{
		float current[4];	// FR, FL, RL, RR [A]
		uint8_t updated;
	} motor;

	struct _enc
	{
		float motorVel[4];	// FR, FL, RL, RR angular motor velocity [rad/s]
		uint8_t updated;
	} enc;

	struct _gyr
	{
		float rotVel[3];	// around X, Y, Z [rad/s]
		uint8_t updated;
	} gyr;

	struct _acc
	{
		float linAcc[3];	// X, Y, Z [m/s]
		uint8_t updated;
	} acc;

	struct _pos
	{
		float pos[3];		// X, Y, Z [m]
		uint8_t updated;
	} pos;

	struct _power
	{
		float voltage;
		uint8_t updated;	// battery voltage [V]
	} power;
} CtrlInput;

typedef struct _CtrlOutput
{
	float motorPWM[4];		// PWM duty cycle = avg. [V], Range -1.0f (CCW) to 1.0f (CW)
} CtrlOutput;

typedef void(*CtrlInitFunction)();
typedef void(*CtrlFunction)(CtrlInput*, CtrlOutput*);
typedef void(*CtrlExitFunction)();

typedef struct _CtrlInstance
{
	CtrlInitFunction pInit;
	CtrlFunction pControllerFunction;
	CtrlExitFunction pExit;

	uint32_t deltaT; // in 0.5ms unit
} CtrlInstance;

typedef struct _Ctrl
{
	CtrlInput input;
	CtrlInstance* pInstance;
	CtrlOutput output;

	CtrlInstance* pNewInstance;

	xQueueHandle hEventQueue;

	float dT;	// execution period	[s]

	// structural bot params
	struct
	{
		float wheelRadius;
		float frontAngle;
		float backAngle;
		float botRadius;
		float mass; // [kg]
		float momentOfInertia;
	} botParams;

	arm_matrix_instance_f32 matXYW2Motor; // 4x3
	arm_matrix_instance_f32 matMotor2XYW; // 3x4

	arm_matrix_instance_f32 matForceXYW2Motor; // 4x3
	arm_matrix_instance_f32 matForceMotor2XYW; // 3x4

	struct
	{
		uint8_t updated[3];
		float pos[3];
		float vel[3];
		float acc[3];

		arm_matrix_instance_f32 matPos;
		arm_matrix_instance_f32 matVel;
		arm_matrix_instance_f32 matAcc;
	} state;

	uint32_t lastHfStatusTime;

	uint8_t sendLog;
} Ctrl;

extern Ctrl ctrl;

void CtrlInit();
void CtrlTask(void* params);
void CtrlSetController(CtrlInstance* pInstance);
void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass);
void CtrlSetLog(uint8_t enable);

#endif /* CTRL_H_ */
