/*
 * ctrl_fusion.c
 *
 *  Created on: 18.03.2013
 *      Author: AndreR
 */

#include "ctrl_fusion.h"

#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/angle_math.h"
#include "network.h"

#include <string.h>

#define REACTION_THRESHOLD 0.01f
#define STALL_TIMEOUT 5000
#define DELTA_T 0.001f

static void init();
static void controller(CtrlInput* pInput, CtrlOutput* pOutput);

CtrlFusion ctrlFusion;
CtrlInstance ctrlFusionInstance = {.pInit = &init, .pExit = 0, .pControllerFunction = &controller, .deltaT = DELTA_T*2000.0f}; // deltaT unit: [0.5ms]

static float kfXData[KF_DATA_SIZE(3, 1, 4, 4)];
static float kfYData[KF_DATA_SIZE(3, 1, 4, 4)];
static float kfWData[KF_DATA_SIZE(3, 1, 4, 4)];

static const float A[9] = {
	1, DELTA_T, 0.5f*DELTA_T*DELTA_T,
	0, 1, DELTA_T,
	0, 0, 1,
};

static const float B[3] = {0, 0, 0};

static const float Ex[9] = {
	0.1f, 0, 0,
	0, 0.1f, 0,
	0, 0, 0.5f,
};

static const float Ez[16] = {
	10.0f, 0, 0, 0,
	0, 2.0f, 0, 0,
	0, 0, 1.0f, 0,
	0, 0, 0, 0.1f,
};

static HermiteSplineListItem listData[3][10];

static void init()
{
	// construct state transition matrix

	// init kalman filters
	KFInit(&ctrlFusion.kfX, 3, 1, 4, kfXData);
	KFInit(&ctrlFusion.kfY, 3, 1, 4, kfYData);
	KFInit(&ctrlFusion.kfW, 3, 1, 4, kfWData);

	KFSetOrientationComponent(&ctrlFusion.kfW, 0);

	memcpy(ctrlFusion.kfX.A.pData, A, sizeof(A));
	memcpy(ctrlFusion.kfY.A.pData, A, sizeof(A));
	memcpy(ctrlFusion.kfW.A.pData, A, sizeof(A));

	memcpy(ctrlFusion.kfX.B.pData, B, sizeof(B));
	memcpy(ctrlFusion.kfY.B.pData, B, sizeof(B));
	memcpy(ctrlFusion.kfW.B.pData, B, sizeof(B));

	memcpy(ctrlFusion.kfX.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusion.kfY.Ex.pData, Ex, sizeof(Ex));
	memcpy(ctrlFusion.kfW.Ex.pData, Ex, sizeof(Ex));

	memcpy(ctrlFusion.kfX.Ez.pData, Ez, sizeof(Ez));
	memcpy(ctrlFusion.kfY.Ez.pData, Ez, sizeof(Ez));
	memcpy(ctrlFusion.kfW.Ez.pData, Ez, sizeof(Ez));

	ctrlFusion.lastWakeTime = xTaskGetTickCount();
	ctrlFusion.lastReaction = ctrlFusion.lastWakeTime;
	ctrlFusion.lastPositionUpdateTime = 0;
	ctrlFusion.visionTimeout = 2.0f;
	ctrlFusion.visionLost = 1;
	ctrlFusion.visionBack = 0;
	ctrlFusion.visionBackTime = 0;

	// init matrices
	ctrlFusion.matCurMotor.pData = ctrlFusion.curMotor;
	ctrlFusion.matCurMotor.numCols = 1;
	ctrlFusion.matCurMotor.numRows = 4;

	ctrlFusion.matCurForce.pData = ctrlFusion.curForce;
	ctrlFusion.matCurForce.numCols = 1;
	ctrlFusion.matCurForce.numRows = 4;

	ctrlFusion.matCurAcc.pData = ctrlFusion.curAcc;
	ctrlFusion.matCurAcc.numCols = 1;
	ctrlFusion.matCurAcc.numRows = 3;

	ctrlFusion.matCurXYW.pData = ctrlFusion.curXYW;
	ctrlFusion.matCurXYW.numCols = 1;
	ctrlFusion.matCurXYW.numRows = 3;

	ctrlFusion.matGlobalOutXYW.pData = ctrlFusion.globalOutXYW;
	ctrlFusion.matGlobalOutXYW.numCols = 1;
	ctrlFusion.matGlobalOutXYW.numRows = 3;

	ctrlFusion.matOutXYW.pData = ctrlFusion.outXYW;
	ctrlFusion.matOutXYW.numCols = 1;
	ctrlFusion.matOutXYW.numRows = 3;

	ctrlFusion.matOutMotor.pData = ctrlFusion.outMotor;
	ctrlFusion.matOutMotor.numCols = 1;
	ctrlFusion.matOutMotor.numRows = 4;

	// init control helpers
	ctrlFusion.posCounter = 0;
	ctrlFusion.velCounter = 0;
	memset(ctrlFusion.posOutput, 0, sizeof(float)*3);
	memset(ctrlFusion.velOutput, 0, sizeof(float)*3);

	// init spline lists
	for(uint8_t i = 0; i < 3; i++)
	{
		HermiteSplineListInit(&ctrlFusion.splineList[i], listData[i], 10);
	}

	// init PIDs (pos and vel)
	for(uint8_t i = 0; i < 3; i++)
	{
		ctrlFusion.pidVel[i].Kp = 1.0f;
		ctrlFusion.pidVel[i].Ki = 0.5f;
		ctrlFusion.pidVel[i].Kd = 0.0f;
		arm_pid_init_f32(&ctrlFusion.pidVel[i], 1);

		ctrlFusion.pidPos[i].Kp = 0.5f;
		ctrlFusion.pidPos[i].Ki = 0.0f;
		ctrlFusion.pidPos[i].Kd = 0.0f;
		arm_pid_init_f32(&ctrlFusion.pidPos[i], 1);
	}

	// init motor acceleration PIDs (inverted system model)
	float Jm = 1.452e-5f; // Moment of inertia at motor
	for(uint8_t i = 0; i < 4; i++)
	{
		ctrlFusion.pidAccMotor[i].Kp = 1.2f*1/CTRL_MOTOR_KM*Jm;
		ctrlFusion.pidAccMotor[i].Ki = 1.0f/(CTRL_MOTOR_KN);
		ctrlFusion.pidAccMotor[i].Kd = 0.56e-3*1/CTRL_MOTOR_KM*Jm;
		arm_pid_init_f32(&ctrlFusion.pidAccMotor[i], 1);
	}
}

static void turnLocal2Global(float orientation, float localX, float localY, float* pGlobalX, float* pGlobalY)
{
	float localToGlobalAngle = -PI/2.0f+orientation;
	float sinA = arm_sin_f32(localToGlobalAngle);
	float cosA = arm_cos_f32(localToGlobalAngle);

	// turn to global system
	*pGlobalX = localX*cosA - localY*sinA;
	*pGlobalY = localY*cosA + localX*sinA;
}

static void turnGlobal2Local(float orientation, float globalX, float globalY, float* pLocalX, float* pLocalY)
{
	float globalToLocalAngle = PI/2.0f - orientation;
	float sinA = arm_sin_f32(globalToLocalAngle);
	float cosA = arm_cos_f32(globalToLocalAngle);

	// turn to local system
	*pLocalX = globalX*cosA - globalY*sinA;
	*pLocalY = globalY*cosA + globalX*sinA;
}

static void controller(CtrlInput* pInput, CtrlOutput* pOutput)
{
	uint8_t xyUpdateNeeded = 0;
	uint8_t wUpdateNeeded = 0;

	// clear C matrices
	arm_mat_zero_f32(&ctrlFusion.kfX.C);
	arm_mat_zero_f32(&ctrlFusion.kfY.C);
	arm_mat_zero_f32(&ctrlFusion.kfW.C);

	if(pInput->pos.updated)
	{
		ctrlFusion.lastPositionUpdateTime = rtosHfTickCount;

		if(ctrlFusion.visionLost)
		{
			ctrlFusion.visionBack = 1;
			ctrlFusion.visionBackTime = xTaskGetTickCount();
			ctrlFusion.visionLost = 0;
		}
	}

	// check vision connection
	if((rtosHfTickCount - ctrlFusion.lastPositionUpdateTime)*20e-6 > ctrlFusion.visionTimeout)
	{
		// vision lost! stop bot. No more KF updates.
		for(uint8_t i = 0; i < 4; i++)
			pOutput->motorPWM[i] = 0;

		ctrlFusion.visionLost = 1;

		return;
	}

	if(xTaskGetTickCount() - ctrlFusion.visionBackTime > 1000/portTICK_RATE_MS)
	{
		ctrlFusion.visionBack = 0;
	}

	// construct C depending on updates
	if(pInput->pos.updated)
	{
		pInput->pos.updated = 0;

		if(!ctrlFusion.visionBack && (fabsf(pInput->pos.pos[0] - ctrl.state.pos[0]) > 1.0f || fabsf(pInput->pos.pos[1] - ctrl.state.pos[1]) > 1.0f))
		{
//			NetworkPrint("Input X or Y change > 1m! (%f/%f)", pInput->pos.pos[0], ctrl.state.pos[0]);
//			NetworkPrint("Input X or Y change > 1m! (%f/%f)", pInput->pos.pos[1], ctrl.state.pos[1]);
		}
		else
		{
			xyUpdateNeeded = 1;

			MAT_ELEMENT(ctrlFusion.kfX.C, 0, 0) = 1.0f;
			MAT_ELEMENT(ctrlFusion.kfY.C, 0, 0) = 1.0f;

			MAT_ELEMENT(ctrlFusion.kfX.z, 0, 0) = pInput->pos.pos[0];
			MAT_ELEMENT(ctrlFusion.kfY.z, 0, 0) = pInput->pos.pos[1];
		}

		if(!ctrlFusion.visionBack && fabsf(pInput->pos.pos[2] - ctrl.state.pos[2]) > 8.0f)
		{
			NetworkPrint("Input W change > 8rad! (%f/%f)", pInput->pos.pos[2], ctrl.state.pos[2]);
		}
		else
		{
			wUpdateNeeded = 1;

			MAT_ELEMENT(ctrlFusion.kfW.C, 0, 0) = 1.0f;

			MAT_ELEMENT(ctrlFusion.kfW.z, 0, 0) = pInput->pos.pos[2];
		}
	}

	if(pInput->enc.updated)
	{
		pInput->enc.updated = 0;
		xyUpdateNeeded = 1;
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfX.C, 1, 1) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfY.C, 1, 1) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfW.C, 1, 1) = 1.0f;

		// convert motor speed to local XYW system
		memcpy(ctrlFusion.curMotor, pInput->enc.motorVel, sizeof(float)*4);

		ctrlFusion.curMotor[0] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[1] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[2] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO;
		ctrlFusion.curMotor[3] *= ctrl.botParams.wheelRadius*CTRL_MOTOR_TO_WHEEL_RATIO; // value is now speed over ground [m/s]

		arm_mat_mult_f32(&ctrl.matMotor2XYW, &ctrlFusion.matCurMotor, &ctrlFusion.matCurXYW);
		// matCurXYW now contains X and Y in [m/s] and W in [rad/s] from encoder data

		MAT_ELEMENT(ctrlFusion.kfW.z, 1, 0) = ctrlFusion.matCurXYW.pData[2];	// store new encoder rotation speed around Z

		// transfer and rotation of velocity vector is done below, if needed
	}

	if(pInput->acc.updated)
	{
		pInput->acc.updated = 0;
		xyUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfX.C, 2, 2) = 1.0f;
		MAT_ELEMENT(ctrlFusion.kfY.C, 2, 2) = 1.0f;

		// transfer and rotation of acceleration vector is done below, if needed
	}

	if(pInput->gyr.updated)
	{
		pInput->gyr.updated = 0;
		wUpdateNeeded = 1;

		MAT_ELEMENT(ctrlFusion.kfW.C, 2, 1) = 1.0f;

		MAT_ELEMENT(ctrlFusion.kfW.z, 2, 0) = pInput->gyr.rotVel[2];	// update gyro rotation around Z axis
	}

	if(pInput->motor.updated)
	{
		pInput->motor.updated = 0;
//		xyUpdateNeeded = 1;
//		wUpdateNeeded = 1;

//		MAT_ELEMENT(ctrlFusion.kfX.C, 3, 2) = 1.0f;
//		MAT_ELEMENT(ctrlFusion.kfY.C, 3, 2) = 1.0f;
//		MAT_ELEMENT(ctrlFusion.kfW.C, 3, 2) = 1.0f;

		for(uint8_t i = 0; i < 4; i++)
		{
			// force = current * motor_constant * gear_ratio * 1/wheel_radius
			// F = I * k_m * G * 1/r_rad
			// F = I * 0.0255 * 50/15 * 1/r_rad
			// F = I * 0.085 * 1/r_rad
			ctrlFusion.curForce[i] = pInput->motor.current[i] * 0.085f * 1.0f/ctrl.botParams.wheelRadius;

			if(ctrlFusion.curMotor[i] < 0)
				ctrlFusion.curForce[i] *= -1.0f;
		}

		// convert wheel forces over ground to global force and torque
		arm_mat_mult_f32(&ctrl.matForceMotor2XYW, &ctrlFusion.matCurForce, &ctrlFusion.matCurAcc);

		// convert forces and torque to (angular) acceleration
		// forces still have to be rotated to global system, done below
		ctrlFusion.curAcc[0] /= ctrl.botParams.mass;
		ctrlFusion.curAcc[1] /= ctrl.botParams.mass;
		ctrlFusion.curAcc[2] /= ctrl.botParams.momentOfInertia;

		MAT_ELEMENT(ctrlFusion.kfW.z, 3, 0) = ctrlFusion.curAcc[2];	// update angular acceleration around Z axis
	}

	// predict W first, as it is needed to rotate X and Y to global system
	KFPredict(&ctrlFusion.kfW);
	AngleNormalizePtr(&ctrlFusion.kfW.x.pData[0]);	// normalize orientation
	if(wUpdateNeeded)
	{
		KFUpdate(&ctrlFusion.kfW);
		AngleNormalizePtr(&ctrlFusion.kfW.x.pData[0]);	// normalize orientation

		arm_mat_clip_f32(&ctrlFusion.kfW.Sigma, -100.0f, 100.0f);
	}

	// predict X,Y
	KFPredict(&ctrlFusion.kfX);
	KFPredict(&ctrlFusion.kfY);

	if(xyUpdateNeeded)
	{
		// X and Y can be updated
		turnLocal2Global(ctrlFusion.kfW.x.pData[0], ctrlFusion.matCurXYW.pData[0], ctrlFusion.matCurXYW.pData[1],
				&MAT_ELEMENT(ctrlFusion.kfX.z, 1, 0), &MAT_ELEMENT(ctrlFusion.kfY.z, 1, 0));

		turnLocal2Global(ctrlFusion.kfW.x.pData[0], pInput->acc.linAcc[0], pInput->acc.linAcc[1],
				&MAT_ELEMENT(ctrlFusion.kfX.z, 2, 0), &MAT_ELEMENT(ctrlFusion.kfY.z, 2, 0));

		turnLocal2Global(ctrlFusion.kfW.x.pData[0], ctrlFusion.curAcc[0], ctrlFusion.curAcc[1],
				&MAT_ELEMENT(ctrlFusion.kfX.z, 3, 0), &MAT_ELEMENT(ctrlFusion.kfY.z, 3, 0));


		// now update
		KFUpdate(&ctrlFusion.kfX);
		KFUpdate(&ctrlFusion.kfY);

		arm_mat_clip_f32(&ctrlFusion.kfX.Sigma, -100.0f, 100.0f);
		arm_mat_clip_f32(&ctrlFusion.kfY.Sigma, -100.0f, 100.0f);
	}

	// test plausibility
//	if(fabsf(ctrl.state.pos[0] - MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update X error > 1m (%f/%f)", ctrl.state.pos[0], MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0);
//	}
//
//	if(fabsf(ctrl.state.pos[1] - MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update Y error > 1m (%f/%f)", ctrl.state.pos[1], MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0);
//	}
//
//	if(fabsf(ctrl.state.pos[2] - MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0)) > 1.0f)
//	{
//		NetworkPrint("KF Update W error > 1rad (%f/%f)", ctrl.state.pos[2], MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0));
//	}
//	else
//	{
//		ctrl.state.pos[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0);
//	}

	// update feedback state
	ctrl.state.pos[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 0, 0);
	ctrl.state.pos[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 0, 0);
	ctrl.state.pos[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 0, 0);

	ctrl.state.vel[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 1, 0);
	ctrl.state.vel[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 1, 0);
	ctrl.state.vel[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 1, 0);

//	ctrl.state.acc[0] = pInput->acc.linAcc[0];
//	ctrl.state.acc[1] = pInput->acc.linAcc[1];
	ctrl.state.acc[0] = MAT_ELEMENT(ctrlFusion.kfX.x, 2, 0);
	ctrl.state.acc[1] = MAT_ELEMENT(ctrlFusion.kfY.x, 2, 0);
	ctrl.state.acc[2] = MAT_ELEMENT(ctrlFusion.kfW.x, 2, 0);

//	for(uint8_t i = 0; i < 2; i++)
//	{
//		if(fabsf(ctrl.state.pos[i]) > 3.0f)
//		{
//			NetworkPrint("state pos %hhu @ %.3fm/s", i, ctrl.state.pos[i]);
//		}
//	}
//
//	for(uint8_t i = 0; i < 3; i++)
//	{
//		if(fabsf(ctrl.state.vel[i]) > 3.0f)
//		{
//			NetworkPrint("state vel %hhu @ %.3fm/s", i, ctrl.state.vel[i]);
//		}
//	}
//
//	for(uint8_t i = 0; i < 3; i++)
//	{
//		if(fabsf(ctrl.state.acc[i]) > 3.0f)
//		{
//			NetworkPrint("state acc %hhu @ %.3fm/s", i, ctrl.state.acc[i]);
//		}
//	}

	ctrl.state.updated[0] = 1;
	ctrl.state.updated[1] = 1;
	ctrl.state.updated[2] = 1;

	// do control...

	ctrlFusion.posCounter++;
	ctrlFusion.posCounter %= 20;
	ctrlFusion.velCounter++;
	ctrlFusion.velCounter %= 4;

	for(uint8_t i = 0; i < 3; i++)
	{
		float trajTime = (rtosHfTickCount - ctrlFusion.trajectoryStartTimeXY)*20e-6f;

		if(i == 2)
			trajTime = (rtosHfTickCount - ctrlFusion.trajectoryStartTimeW)*20e-6f;

		float partTime = 0.0f;
		HermiteSplineListItem* pItem = ctrlFusion.splineList[i].pItems;
		if(ctrlFusion.splineList[i].numItems > 0)
		{
			pItem = HermiteSplineListGetItem(&ctrlFusion.splineList[i], trajTime);
			partTime = trajTime - pItem->tStart;
		}

		if(trajTime > ctrlFusion.splineList[i].totalTime + 3.0f && i != 2)
		{
			HermiteSplineListClear(&ctrlFusion.splineList[i]);

//			IntercomHeader iHeader;
//			iHeader.section = INTERCOM_SECTION_SENSORS;
//			iHeader.section = INTERCOM_ADDRESS_MEDIA;
//			iHeader.command = 2;

//			IntercomSend(&intercom1, &iHeader, 0, 0);
		}

		// position control at 50Hz
//		if(ctrlFusion.posCounter == 0)
		{
			if(ctrlFusion.splineList[i].numItems > 0)
			{
				float setPos = HermiteSplineValue(&pItem->spline, partTime);
				float errorPos = setPos - ctrl.state.pos[i];
				if(i == 2)
					AngleNormalizePtr(&errorPos);	// shortest rotation distance

				ctrlFusion.posOutput[i] = arm_pid_f32(&ctrlFusion.pidPos[i], errorPos);
			}
			else
			{
				ctrlFusion.posOutput[i] = 0.0f;
			}
		}

		// velocity control at 1000Hz
		float setVel = 0.0f;
//		if(ctrlFusion.velCounter == 0)
		{
			if(ctrlFusion.splineList[i].numItems > 0)
				setVel = HermiteSplineFirstDerivative(&pItem->spline, partTime);

			float errorVel = ctrlFusion.posOutput[i] + setVel - ctrl.state.vel[i];
			ctrlFusion.velOutput[i] = arm_pid_f32(&ctrlFusion.pidVel[i], errorVel);
		}

		// acceleration pre-control
		float setAcc = 0.0f;
		if(ctrlFusion.splineList[i].numItems > 0)
			setAcc = HermiteSplineSecondDerivative(&pItem->spline, partTime);

//		float errorAcc = ctrlFusion.velOutput[i] + setAcc - ctrl.state.acc[i];
//		ctrlFusion.globalOutXYW[i] = arm_pid_f32(&ctrlFusion.pidAcc[i], errorAcc);
//		ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i];
		ctrlFusion.globalOutXYW[i] = ctrlFusion.velOutput[i] + ctrlFusion.pidAccMotor[0].Kp*setAcc;// - ctrl.state.acc[i];
//		ctrlFusion.globalOutXYW[i] = ctrlFusion.posOutput[i] + ctrlFusion.velOutput[i] + setAcc;// - ctrl.state.acc[i];
	}

	if(ctrlFusion.posCounter == 0)
	{
//		ConsolePrint("1: %.3f, 2: %.3f, 3: %.3f, 4: %.3f\r\n", pInput->motor.current[0], pInput->motor.current[1], pInput->motor.current[2], pInput->motor.current[3]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrl.state.acc[0], ctrl.state.acc[1], ctrl.state.acc[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusion.velOutput[0], ctrlFusion.velOutput[1], ctrlFusion.velOutput[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusion.curAcc[0], ctrlFusion.curAcc[1], ctrlFusion.curAcc[2]);
//		ConsolePrint("X: %.3f, Y: %.3f, W: %.3f\r\n", ctrlFusion.globalOutXYW[0], ctrlFusion.globalOutXYW[1], ctrlFusion.globalOutXYW[2]);
	}

	// turn back to local system
	turnGlobal2Local(ctrl.state.pos[2], ctrlFusion.globalOutXYW[0], ctrlFusion.globalOutXYW[1], &ctrlFusion.outXYW[0], &ctrlFusion.outXYW[1]);
	// angular acceleration is copied as-it-is
	ctrlFusion.outXYW[2] = ctrlFusion.globalOutXYW[2];

	// TODO: stall protection

	// grab current battery voltage (or guess)
	float UBattery = 15.0f;
	if(pInput->power.updated)
	{
		UBattery = pInput->power.voltage;
	}

	// convert to speed over ground of motors
	arm_mat_mult_f32(&ctrl.matXYW2Motor, &ctrlFusion.matOutXYW, &ctrlFusion.matOutMotor);

	for(uint8_t i = 0; i < 4; i++)
	{
		// convert speed over ground to motor speed
		// 50/15 * speed_over_ground * 1/r_rad
		float motorSpeed = CTRL_WHEEL_TO_MOTOR_RATIO * 1/ctrl.botParams.wheelRadius * ctrlFusion.outMotor[i];
		// and convert to voltage
		float outVoltage = motorSpeed/CTRL_MOTOR_KN;

		if(fabsf(pInput->motor.current[i]) > 7.0f)
		{
			NetworkPrint("Motor %hhu overcurrent %f", i, pInput->motor.current[i]);
			pOutput->motorPWM[i] = 0;
		}
		else
		{
			pOutput->motorPWM[i] = outVoltage/UBattery;
		}
	}
/*
	// convert (angular) acceleration to forces and torques
	ctrlFusion.outXYW[0] *= ctrl.botParams.mass;
	ctrlFusion.outXYW[1] *= ctrl.botParams.mass;
	ctrlFusion.outXYW[2] *= ctrl.botParams.momentOfInertia;

	// convert to wheel forces over ground
	arm_mat_mult_f32(&ctrl.matForceXYW2Motor, &ctrlFusion.matOutXYW, &ctrlFusion.matOutMotor);

	for(uint8_t i = 0; i < 4; i++)
	{
		// force = current * motor_constant * gear_ratio * 1/wheel_radius
		// F = I * k_m * G * 1/r_rad
		// I = F * r_rad/(k_m * G)
		// I = F * r_rad * 1/(0.0255*50/15)
		// I = F * r_rad * 11.76470588235294f;
		ctrlFusion.outMotor[i] *= ctrl.botParams.wheelRadius * 11.76470588235294f;
		// .matOutMotor is now required current in motor

		ctrlFusion.outMotor[i] *= 1.2f; // motor resistance. converts to required voltage

		if(fabsf(pInput->motor.current[i]) > 7.0f)
		{
			NetworkPrint("Motor %hhu overcurrent %f", i, pInput->motor.current[i]);
			pOutput->motorPWM[i] = 0;
		}
		else
		{
			pOutput->motorPWM[i] = ctrlFusion.outMotor[i]/UBattery;
		}
	}
*/
	ctrlFusion.lastWakeTime = xTaskGetTickCount();
}

void CtrlFusionSetParams(float kp, float ki, float kd)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	for(uint8_t i = 0; i < 3; i++)
	{
		ctrlFusion.pidVel[i].Kp = kp;
		ctrlFusion.pidVel[i].Ki = ki;
		ctrlFusion.pidVel[i].Kd = kd;
		arm_pid_init_f32(&ctrlFusion.pidVel[i], 1);
	}
}

void CtrlFusionSetStateUncertainty(float* pEx)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	for(uint16_t i = 0; i < ctrlFusion.kfX.f; i++)
	{
		MAT_ELEMENT(ctrlFusion.kfX.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusion.kfY.Ex, i, i) = pEx[i];
		MAT_ELEMENT(ctrlFusion.kfW.Ex, i, i) = pEx[i];
	}
}

void CtrlFusionSetSensorUncertainty(float* pEz)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	for(uint16_t i = 0; i < ctrlFusion.kfX.h; i++)
	{
		MAT_ELEMENT(ctrlFusion.kfX.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusion.kfY.Ez, i, i) = pEz[i];
		MAT_ELEMENT(ctrlFusion.kfW.Ez, i, i) = pEz[i];
	}
}

void CtrlFusionAddSplineCmdXY(CtrlSpline2D* pSpline)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	if(pSpline->options & CTRL_SPLINE_OPTION_CLEAR)
	{
		HermiteSplineListClear(&ctrlFusion.splineList[0]);
		HermiteSplineListClear(&ctrlFusion.splineList[1]);

		arm_pid_reset_f32(&ctrlFusion.pidPos[0]);
		arm_pid_reset_f32(&ctrlFusion.pidPos[1]);

		ctrlFusion.trajectoryStartTimeXY = rtosHfTickCount;

/*		IntercomHeader iHeader;
		iHeader.section = INTERCOM_SECTION_SENSORS;
		iHeader.address = INTERCOM_ADDRESS_MEDIA;
		iHeader.command = 1;

		IntercomSend(&intercom1, &iHeader, 0, 0);
*/
		ConsolePrint("Splines XY cleared\r\n");
	}

	if(pSpline->options & CTRL_SPLINE_OPTION_APPEND)
	{
		if(ctrlFusion.splineList[0].numItems == 0)
		{
			HermiteSplineListItem* pItemX = HermiteSplineListAllocate(&ctrlFusion.splineList[0]);
			if(!pItemX)
				return;

			HermiteSplineListItem* pItemY = HermiteSplineListAllocate(&ctrlFusion.splineList[1]);
			if(!pItemY)
				return;

			HermiteSpline* pSplineX = &pItemX->spline;
			HermiteSpline* pSplineY = &pItemY->spline;

			pSplineX->a = halfFloatToFloat(pSpline->splines[0].a);
			pSplineX->b = halfFloatToFloat(pSpline->splines[0].b);
			pSplineX->c = halfFloatToFloat(pSpline->splines[0].c);
			pSplineX->d = halfFloatToFloat(pSpline->splines[0].d);
			pSplineX->tEnd = halfFloatToFloat(pSpline->tEnd);
			pSplineY->a = halfFloatToFloat(pSpline->splines[1].a);
			pSplineY->b = halfFloatToFloat(pSpline->splines[1].b);
			pSplineY->c = halfFloatToFloat(pSpline->splines[1].c);
			pSplineY->d = halfFloatToFloat(pSpline->splines[1].d);
			pSplineY->tEnd = halfFloatToFloat(pSpline->tEnd);

			float maxVel = HermiteSpline2DMaxFirstDerivative(pSplineX, pSplineY);
			float maxAcc = HermiteSpline2DMaxSecondDerivative(pSplineX, pSplineY);

			Vector2f finalPos = Vector2fCreate(HermiteSplineValue(pSplineX, pSplineX->tEnd),
					HermiteSplineValue(pSplineY, pSplineY->tEnd));
			Vector2f finalVel = Vector2fCreate(HermiteSplineFirstDerivative(pSplineX, pSplineX->tEnd),
					HermiteSplineFirstDerivative(pSplineY, pSplineY->tEnd));

			Vector2f initialPos = Vector2fCreate(ctrl.state.pos[0], ctrl.state.pos[1]);
			Vector2f initialVel = Vector2fCreate(ctrl.state.vel[0], ctrl.state.vel[1]);

			HermiteSpline2DGenerate(initialPos, finalPos, initialVel, finalVel, maxVel, maxAcc, pSplineX, pSplineY);

			pItemX->tStart = 0;
			pItemX->tEnd = pSplineX->tEnd;

			pItemY->tStart = 0;
			pItemY->tEnd = pSplineY->tEnd;
		}
		else
		{
			for(uint8_t i = 0; i < 2; i++)
			{
				HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[i]);
				if(!pItem)
					return;

				pItem->spline.a = halfFloatToFloat(pSpline->splines[i].a);
				pItem->spline.b = halfFloatToFloat(pSpline->splines[i].b);
				pItem->spline.c = halfFloatToFloat(pSpline->splines[i].c);
				pItem->spline.d = halfFloatToFloat(pSpline->splines[i].d);
				pItem->spline.tEnd = halfFloatToFloat(pSpline->tEnd);

			}
		}

		HermiteSplineListUpdate(&ctrlFusion.splineList[0]);
		HermiteSplineListUpdate(&ctrlFusion.splineList[1]);

		ConsolePrint("Spline XY appended\r\n");
	}
}

void CtrlFusionAddSplineCmdW(CtrlSpline1D* pSpline1D)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	if(pSpline1D->options & CTRL_SPLINE_OPTION_CLEAR)
	{
		HermiteSplineListClear(&ctrlFusion.splineList[2]);

		arm_pid_reset_f32(&ctrlFusion.pidPos[2]);

		ctrlFusion.trajectoryStartTimeW = rtosHfTickCount;

		ConsolePrint("Splines W cleared\r\n");
	}

	if(pSpline1D->options & CTRL_SPLINE_OPTION_APPEND)
	{
		if(ctrlFusion.splineList[2].numItems == 0)
		{
			HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[2]);
			if(!pItem)
				return;

			HermiteSpline* pSpline = &pItem->spline;

			pSpline->a = halfFloatToFloat(pSpline1D->a);
			pSpline->b = halfFloatToFloat(pSpline1D->b);
			pSpline->c = halfFloatToFloat(pSpline1D->c);
			pSpline->d = halfFloatToFloat(pSpline1D->d);
			pSpline->tEnd = halfFloatToFloat(pSpline1D->tEnd);

			float maxVel = HermiteSplineMaxFirstDerivative(pSpline);
			float maxAcc = HermiteSplineMaxSecondDerivative(pSpline);

			float finalPos = HermiteSplineValue(pSpline, pSpline->tEnd);
			float finalVel = HermiteSplineFirstDerivative(pSpline, pSpline->tEnd);

			float initialPos = ctrl.state.pos[2];
			float initialVel = ctrl.state.vel[2];

			HermiteSplineGenerate(initialPos, finalPos, initialVel, finalVel, pSpline->tEnd, maxVel, maxAcc, 1, pSpline);

			pItem->tStart = 0;
			pItem->tEnd = pSpline->tEnd;
		}
		else
		{
			HermiteSplineListItem* pItem = HermiteSplineListAllocate(&ctrlFusion.splineList[2]);
			if(!pItem)
				return;

			pItem->spline.a = halfFloatToFloat(pSpline1D->a);
			pItem->spline.b = halfFloatToFloat(pSpline1D->b);
			pItem->spline.c = halfFloatToFloat(pSpline1D->c);
			pItem->spline.d = halfFloatToFloat(pSpline1D->d);
			pItem->spline.tEnd = halfFloatToFloat(pSpline1D->tEnd);
		}

		HermiteSplineListUpdate(&ctrlFusion.splineList[2]);

		ConsolePrint("Spline appended\r\n");
	}
}

void CtrlFusionSetPIDParams(CtrlSetPIDParams* pParams)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	switch(pParams->type)
	{
		case CTRL_PID_PARAM_POS_X:
		{
			ctrlFusion.pidPos[0].Kp = pParams->Kp;
			ctrlFusion.pidPos[0].Ki = pParams->Ki;
			ctrlFusion.pidPos[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[0], 1);
		}
		break;
		case CTRL_PID_PARAM_POS_Y:
		{
			ctrlFusion.pidPos[1].Kp = pParams->Kp;
			ctrlFusion.pidPos[1].Ki = pParams->Ki;
			ctrlFusion.pidPos[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[1], 1);
		}
		break;
		case CTRL_PID_PARAM_POS_W:
		{
			ctrlFusion.pidPos[2].Kp = pParams->Kp;
			ctrlFusion.pidPos[2].Ki = pParams->Ki;
			ctrlFusion.pidPos[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidPos[2], 1);
		}
		break;
		case CTRL_PID_PARAM_VEL_X:
		{
			ctrlFusion.pidVel[0].Kp = pParams->Kp;
			ctrlFusion.pidVel[0].Ki = pParams->Ki;
			ctrlFusion.pidVel[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[0], 1);
		}
		break;
		case CTRL_PID_PARAM_VEL_Y:
		{
			ctrlFusion.pidVel[1].Kp = pParams->Kp;
			ctrlFusion.pidVel[1].Ki = pParams->Ki;
			ctrlFusion.pidVel[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[1], 1);
		}
		break;
		case CTRL_PID_PARAM_VEL_W:
		{
			ctrlFusion.pidVel[2].Kp = pParams->Kp;
			ctrlFusion.pidVel[2].Ki = pParams->Ki;
			ctrlFusion.pidVel[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidVel[2], 1);
		}
		break;
		case CTRL_PID_PARAM_ACC_X:
		{
			ctrlFusion.pidAccMotor[0].Kp = pParams->Kp;
			ctrlFusion.pidAccMotor[0].Ki = pParams->Ki;
			ctrlFusion.pidAccMotor[0].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidAccMotor[0], 1);
		}
		break;
		case CTRL_PID_PARAM_ACC_Y:
		{
			ctrlFusion.pidAccMotor[1].Kp = pParams->Kp;
			ctrlFusion.pidAccMotor[1].Ki = pParams->Ki;
			ctrlFusion.pidAccMotor[1].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidAccMotor[1], 1);
		}
		break;
		case CTRL_PID_PARAM_ACC_W:
		{
			ctrlFusion.pidAccMotor[2].Kp = pParams->Kp;
			ctrlFusion.pidAccMotor[2].Ki = pParams->Ki;
			ctrlFusion.pidAccMotor[2].Kd = pParams->Kd;
			arm_pid_init_f32(&ctrlFusion.pidAccMotor[2], 1);
		}
		break;
	}
}

void CtrlFusionSetFilterParams(CtrlSetFilterParams* pParams)
{
	if(ctrl.pInstance != &ctrlFusionInstance)
		return;

	switch(pParams->type)
	{
		case CTRL_FILTER_PARAM_EX_POS:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 0, 0) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EX_VEL:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 1, 1) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EX_ACC:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ex, 2, 2) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ex, 2, 2) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ex, 2, 2) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_VISION:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 0, 0) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 0, 0) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 0, 0) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_ENCODER:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 1, 1) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 1, 1) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 1, 1) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_ACC_GYRO:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 2, 2) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 2, 2) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 2, 2) = pParams->params[2];
		}
		break;
		case CTRL_FILTER_PARAM_EZ_MOTOR:
		{
			MAT_ELEMENT(ctrlFusion.kfX.Ez, 3, 3) = pParams->params[0];
			MAT_ELEMENT(ctrlFusion.kfY.Ez, 3, 3) = pParams->params[1];
			MAT_ELEMENT(ctrlFusion.kfW.Ez, 3, 3) = pParams->params[2];
		}
		break;
	}
}
