/*
 * dribbler.c
 *
 *  Created on: 10.03.2013
 *      Author: AndreR
 */

#include "dribbler.h"

#include "hal/motor.h"
#include "network.h"
#include "commands.h"
#include "motor_ir.h"

#include "FreeRTOS.h"
#include "task.h"

#include <string.h>

#define COM_TO_BAR_FACTOR 0.45614f
#define TASK_DELAY 50 // [ms]
#define DRIBBLER_REACTION_THRESHOLD 100.0f // [RPM]
#define DRIBBLER_STALL_TIMEOUT 2000

Dribbler dribbler;

void DribblerInit()
{
	memset(&dribbler, 0, sizeof(Dribbler));

	dribbler.enabled = 1;
	dribbler.reachedToleranceIn = 100.0f;
	dribbler.reachedTolercanceOut = 500.0f;
	DribblerSetParams(0.1f, 0.1f, 0.0f);
}

void DribblerSetParams(float kp, float ki, float kd)
{
	dribbler.pid.Kp = kp;
	dribbler.pid.Ki = ki;
	dribbler.pid.Kd = kd;
	arm_pid_init_f32(&dribbler.pid, 1);
}

void DribblerSetTolerances(float in, float out)
{
	dribbler.reachedToleranceIn = in;
	dribbler.reachedTolercanceOut = out;
}

void DribblerSetTarget(float rpm)
{
	if(rpm > 18000.0f)
		rpm = 18000.0f;

	if(rpm < -18000.0f)
		rpm = -18000.0f;

	dribbler.setpoint = -rpm; // invert for back spin
}

void DribblerSetLog(uint8_t enable)
{
	dribbler.logEnabled = enable;
}

void DribblerSetSpeedCallback(SpeedReachedCallback cb)
{
	dribbler.speedCb = cb;
}

void DribblerTask(void* params)
{
	(void) params;

	uint32_t lastWakeTime = xTaskGetTickCount();

	while(1)
	{
		if(!dribbler.enabled)
		{
			vTaskDelay(10/portTICK_RATE_MS);
			continue;
		}

		float dt_real = ((float)(rtosHfTickCount - dribbler.lastHfTickCount))*20e-6f;	// [s]
		dribbler.lastHfTickCount = rtosHfTickCount;

		// get dribbler speed
		float commutations = motor.M[4].ccw == 1 ? -motor.M[4].coms : motor.M[4].coms;
		motor.M[4].coms = 0;

		float cps = commutations*1/dt_real;
		// rps_motor = cps/6.0f; // 6 commutations per round
		// rps_shaft = rps_motor*13/57; // planetary gear
		// rps_bar = rps_shaft*2; spur gear
		// rpm_bar = rps_bar*60;
		// => rpm_bar = cps*2.2807f;
		float rpm = cps*2.2807f;

		// do control
		float error = dribbler.setpoint - rpm;
		float rpm_out = arm_pid_f32(&dribbler.pid, error);

		// check if the desired speed has been reached, using hysteresis
		if(fabsf(error) < dribbler.reachedToleranceIn && fabsf(rpm) > 500.0f && dribbler.speedReached == 0)
		{
			dribbler.speedReached = 1;

			if(dribbler.speedCb)
				(*dribbler.speedCb)();
		}
		else if(fabsf(error) > dribbler.reachedTolercanceOut || fabsf(rpm) <= 500.0f)
		{
			dribbler.speedReached = 0;
		}

		// logging
		if(dribbler.logEnabled)
		{
			MotorPidLog log;
			log.latest = rpm;
			log.setpoint = dribbler.setpoint;
			log.pError = dribbler.pid.state[0];
			log.iError = dribbler.pid.state[1];
			log.dError = dribbler.pid.state[2];
			log.id = 4;
			log.output = rpm_out;
			log.overload = dribbler.overload;
			log.eCurrent = motorIr.atData.current[4];

			PacketHeader header;
			header.section = SECTION_MOTOR;
			header.cmd = CMD_MOTOR_PID_LOG;
			header.len = LEN_MOTOR_PID_LOG;

			NetworkSendPacket(&header, (uint8_t*)&log);
		}

		// scale output [RPM] to pwm output range
		// rpm_shaft = rpm_out/2; // spur gear
		// rpm_motor = rpm_shaft*57/13; // planetary gear
		// pwm = rpm_motor*50/39400; // scale to pwm range with rated speed
		// => pwm = rpm_out*0.00278f;
		float output = rpm_out*0.00278f;

//		ConsolePrint("RPM: %f, Out: %f, pwm: %f\r\n", rpm, rpm_out, output);

		if(output > 50.0f || output < -50.0f)
			dribbler.overload = 1;
		else
			dribbler.overload = 0;

		// Output
		MotorSetPower(4, (int16_t)output);

		// stall protection
		if(rpm < -DRIBBLER_REACTION_THRESHOLD || rpm > DRIBBLER_REACTION_THRESHOLD)
		{
			dribbler.lastReaction = lastWakeTime;
		}

		if(lastWakeTime - dribbler.lastReaction > DRIBBLER_STALL_TIMEOUT/portTICK_RATE_MS)
		{
			// last reaction was more than 5s ago
			arm_pid_reset_f32(&dribbler.pid);
			MotorSetPower(4, 0);
			dribbler.lastReaction = lastWakeTime;
		}


		vTaskDelayUntil(&lastWakeTime, TASK_DELAY/portTICK_RATE_MS);
	}
}
