/*
 * enc.c
 *
 *  Created on: 27.07.2010
 *      Author: AndreR
 */

#include "enc.h"

#include "stm32f4xx.h"
#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_tim.h"
#include "stm32f4xx_exti.h"

#include "arm_math.h"

#include "util/bit_banding.h"
#include "constants.h"

#include <string.h>

#define UPDATE_RATE 200 // [Hz]
#define UPDATE_TIME (1000/UPDATE_RATE) // [ms]
#define ENCODER_PPR 2048 // Encoder pulses per round [ppr]

#define PPR_TO_RADS ((float)(UPDATE_RATE))/((float)ENCODER_PPR)*PI*2.0f

Encoder encoder;

volatile int16_t* idToCntLUT[4] = { (int16_t*)&TIM3->CNT, (int16_t*)&TIM2->CNT, (int16_t*)&TIM1->CNT, (int16_t*)&TIM4->CNT };

void TIM6_DAC_IRQHandler()
{
	TIM6->SR = 0;

	int16_t cur[4];

	cur[0] = -*idToCntLUT[0];
	cur[1] = -*idToCntLUT[1];
	cur[2] = -*idToCntLUT[2];
	cur[3] = -*idToCntLUT[3];

	TIM1->CNT = 0;
	TIM2->CNT = 0;
	TIM3->CNT = 0;
	TIM4->CNT = 0;

	encoder.angularVelocity[0] = ((float)cur[0])*PPR_TO_RADS;
	encoder.angularVelocity[1] = ((float)cur[1])*PPR_TO_RADS;
	encoder.angularVelocity[2] = ((float)cur[2])*PPR_TO_RADS;
	encoder.angularVelocity[3] = ((float)cur[3])*PPR_TO_RADS;

	if(encoder.pUpdatedCallback)
		(*encoder.pUpdatedCallback)();
}

void EncEnableAutoUpdate(uint8_t enable)
{
	if(enable)
	{
		TIM1->CNT = 0;
		TIM2->CNT = 0;
		TIM3->CNT = 0;
		TIM4->CNT = 0;

		TIM_Cmd(TIM6, ENABLE);
	}
	else
	{
		TIM_Cmd(TIM6, DISABLE);
	}
}

float EncGetCurrentAngle(uint8_t id)
{
	if(id > 3)
		return 0;

	return ((float)*idToCntLUT[id])*PPR_TO_RADS/UPDATE_RATE;
}

void EncZeroAll()
{
	*idToCntLUT[0] = 0;
	*idToCntLUT[1] = 0;
	*idToCntLUT[2] = 0;
	*idToCntLUT[3] = 0;
}

void EncInit()
{
	memset(&encoder, 0, sizeof(Encoder));

	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM1, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM2, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM3, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6, ENABLE);

	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_9 | GPIO_Pin_11;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOE, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1;
	GPIO_Init(GPIOA, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7;
	GPIO_Init(GPIOC, &gpioInit);

	GPIO_Init(GPIOB, &gpioInit);

	GPIO_PinAFConfig(GPIOA, GPIO_PinSource0, GPIO_AF_TIM2);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource1, GPIO_AF_TIM2);

	GPIO_PinAFConfig(GPIOB, GPIO_PinSource6, GPIO_AF_TIM4);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource7, GPIO_AF_TIM4);

	GPIO_PinAFConfig(GPIOC, GPIO_PinSource6, GPIO_AF_TIM3);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource7, GPIO_AF_TIM3);

	GPIO_PinAFConfig(GPIOE, GPIO_PinSource9, GPIO_AF_TIM1);
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource11, GPIO_AF_TIM1);

	//Timer Init - Encoder inputs
	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = 0;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 0xFFFF;	// 65535
	timeBaseInit.TIM_Prescaler = 0;
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM1, &timeBaseInit);
	TIM_TimeBaseInit(TIM2, &timeBaseInit);
	TIM_TimeBaseInit(TIM3, &timeBaseInit);
	TIM_TimeBaseInit(TIM4, &timeBaseInit);

	TIM_EncoderInterfaceConfig(TIM1, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);
	TIM_EncoderInterfaceConfig(TIM2, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);
	TIM_EncoderInterfaceConfig(TIM3, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);
	TIM_EncoderInterfaceConfig(TIM4, TIM_EncoderMode_TI12, TIM_ICPolarity_Rising, TIM_ICPolarity_Rising);

	TIM_Cmd(TIM1, ENABLE);
	TIM_Cmd(TIM2, ENABLE);
	TIM_Cmd(TIM3, ENABLE);
	TIM_Cmd(TIM4, ENABLE);

	// 10ms precision timer for capturing encoder inputs
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = (UPDATE_TIME*2);	// generates 10ms clock
	timeBaseInit.TIM_Prescaler = 41999; // clock is now 2kHz
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM6, &timeBaseInit);

	TIM_ARRPreloadConfig(TIM6, DISABLE);
	TIM_CCPreloadControl(TIM6, DISABLE);
	TIM_InternalClockConfig(TIM6);
	TIM_ITConfig(TIM6, TIM_IT_Update, ENABLE);

	TIM_Cmd(TIM6, ENABLE);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = TIM6_DAC_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_ENCODER_TIMER;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);
}

void EncSetUpdateCallback(NotifyUpdated pCb)
{
	encoder.pUpdatedCallback = pCb;
}
