/*
 * kicker.c
 *
 *  Created on: 19.03.2011
 *      Author: AndreR
 */

#include "kicker.h"

#include "FreeRTOS.h"
#include "task.h"

#include "stm32f4xx_exti.h"
#include "stm32f4xx_syscfg.h"
#include "misc.h"

#include "hal/exti.h"
#include "network.h"
#include "commands.h"
#include "spi1_2_3.h"
#include "util/critical.h"
#include "util/console.h"
#include "util/bit_banding.h"
#include "motor_ir.h"
#include "dribbler.h"
#include "intercom1.h"
#include "intercom_constants.h"

#include "atmega/kicker_prog.h"

#define KICKER_ATMEGA_RST	GPIO_Pin_5	// GPIOB
#define KICKER_ATMEGA_CS	GPIO_Pin_15	// GPIOG

Kicker kicker;

#define IR_IRQ (1 << 15)

#define KICKER_EVENT_NONE		0
#define KICKER_EVENT_CMD		1
#define KICKER_EVENT_BARRIER	2
#define KICKER_EVENT_DRIBBLER	3
#define KICKER_EVENT_DISCHARGE	4

#define KICKER_STATE_STANDBY		0
#define KICKER_STATE_IR_ARMED		1
#define KICKER_STATE_DRIBBLER_ARMED	2
#define KICKER_STATE_DISCHARGE		3

#define BarrierInterrupted() BBPeriphGet((uint32_t)&GPIOF->IDR, 15)

typedef struct _KickerEvent
{
	uint8_t type;

	KickerCmd cmd;
} KickerEvent;


// called from ISR!
static void IRIRQ()
{
	int32_t yield = pdFALSE;

	KickerEvent event;
	event.type = KICKER_EVENT_BARRIER;

	xQueueSendFromISR(kicker.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void speedReached()
{
	KickerEvent event;
	event.type = KICKER_EVENT_DRIBBLER;

	xQueueSend(kicker.hEventQueue, &event, 0);
}

void KickerInit()
{
	GPIO_InitTypeDef gpioInit;

	// configure kicker ATmega
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_Pin = KICKER_ATMEGA_RST;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_DOWN;
	GPIO_Init(GPIOB, &gpioInit);

	vTaskDelay(10/portTICK_RATE_MS);

	if(GPIOB->IDR & KICKER_ATMEGA_RST)
		kicker.attached = 1;
	else
		kicker.attached = 0;

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Pin = KICKER_ATMEGA_RST;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = KICKER_ATMEGA_CS;
	GPIO_Init(GPIOG, &gpioInit);

	GPIOB->BSRRL = KICKER_ATMEGA_RST;
	GPIOG->BSRRL = KICKER_ATMEGA_CS;

	ATMegaInit(&kicker.atmega, &spi3, GPIOG, KICKER_ATMEGA_CS, GPIOB, KICKER_ATMEGA_RST);

	// configure IR barrier IRQ
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_DOWN;
	gpioInit.GPIO_Pin = IR_IRQ;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	GPIO_Init(GPIOF, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOF, EXTI_PinSource15);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line15;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising;
	EXTI_Init(&extiInit);

	ExtiSetCallback(&IRIRQ, 15);

	DribblerSetSpeedCallback(&speedReached);

	// init data structure
	kicker.data.cmd.type = KICKER_CMD_NONE;
	kicker.state = KICKER_STATE_STANDBY;

	kicker.logEnabled = 1;

	kicker.latestStatus.cap = 0;
	kicker.latestStatus.chg = 0;
	kicker.latestStatus.ir = 0;
	kicker.lastKick = 0;
	kicker.doDischarge = 0;

	kicker.hEventQueue = xQueueCreate(20, sizeof(KickerEvent));
}

void KickerSetLogging(uint8_t enable)
{
	kicker.logEnabled = enable;
}

void KickerLoadManual(uint16_t tOn, uint16_t tOff, uint16_t duration)
{
	duration *= 10; // convert to [se-4]

	KickerEvent event;
	event.type = KICKER_EVENT_CMD;
	event.cmd.loadManual.tOnTicks = tOn;
	event.cmd.loadManual.tOffTicks = tOff;
	event.cmd.loadManual.duration = duration;
	event.cmd.type = KICKER_CMD_LOAD_MANUAL;

	xQueueSend(kicker.hEventQueue, &event, 0);
}

void KickerLoadAuto(uint16_t maxCap)
{
	kicker.doDischarge = 0;

	KickerEvent event;
	event.type = KICKER_EVENT_CMD;
	event.cmd.loadAuto.maxCapLevel = maxCap;
	event.cmd.type = KICKER_CMD_LOAD_AUTO;

	xQueueSend(kicker.hEventQueue, &event, 0);
}

void KickerFire(float duration, uint8_t device)
{
	kicker.state = KICKER_STATE_STANDBY;

	// convert float [us] to FixedPoint142 [se-6]
	uint16_t exp = 0;
	while(duration > 16383.0f)
	{
		duration /= 10.0f;
		exp++;
	}

	if(exp > 3)
	{
		ConsolePrint("Kicker duration too big (>16.384s)\r\n");
		return;
	}

	KickerEvent event;
	event.type = KICKER_EVENT_CMD;
	event.cmd.fire.duration.value = (uint16_t)duration;
	event.cmd.fire.duration.exp = exp;
	event.cmd.fire.device = device;
	event.cmd.type = KICKER_CMD_FIRE;

	if(xQueueSend(kicker.hEventQueue, &event, 0) != pdTRUE)
		ConsolePrint("Send fire cmd to queue failed\r\n");
}

void KickerArmIr(float fireDuration, uint8_t device)
{
	kicker.armed.fireDuration = fireDuration;
	kicker.armed.device = device;
	kicker.state = KICKER_STATE_IR_ARMED;
}

void KickerArmDribbler(float fireDuration, uint8_t device)
{
	kicker.armed.fireDuration = fireDuration;
	kicker.armed.device = device;
	kicker.state = KICKER_STATE_DRIBBLER_ARMED;
}

void KickerDisarm()
{
	kicker.armed.fireDuration = 0;
	kicker.state = KICKER_STATE_STANDBY;
}

void KickerDischarge()
{
	KickerLoadAuto(0);

	kicker.doDischarge = 1;
}

void KickerTask(void* params)
{
	params = params;

	// -- Kicker
	uint8_t programUpdated = 0;
	int16_t result = ATMegaUpdateProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE, &programUpdated);

	ConsolePrint("--- Kicker AVR ---\r\n");
	ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
	ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	if(programUpdated)
		ConsolePrint("\e[0;36mUpdated kicker program\e[0m\r\n");
	if(result)
		ConsolePrint("\e[0;31mKicker program update failed (0x%X)\r\n", result);

	portTickType xLastWakeTime = xTaskGetTickCount();
	portTickType xLastStatusTime = xLastWakeTime;
	portTickType xLastIcStatusTime = xLastWakeTime;

	KickerEvent event;

	while(1)
	{
		xLastWakeTime = xTaskGetTickCount();

		if(xQueueReceive(kicker.hEventQueue, &event, 20/portTICK_RATE_MS) == pdTRUE)
		{
			if(event.type == KICKER_EVENT_CMD)
			{
				kicker.data.cmd = event.cmd;

				ATMegaExchange(&kicker.atmega, kicker.data.exchange.mosi, kicker.data.exchange.miso, KICKER_DATA_SIZE);

				if(event.cmd.type == KICKER_CMD_FIRE)
				{
					DribblerSetTarget(0);

					kicker.lastKick = xTaskGetTickCount();
				}

				kicker.data.cmd.type = KICKER_CMD_NONE;
			}

			if(event.type == KICKER_EVENT_BARRIER)
			{
				if(kicker.state == KICKER_STATE_IR_ARMED)
				{
					KickerFire(kicker.armed.fireDuration, kicker.armed.device);
				}

				if(kicker.state == KICKER_STATE_DRIBBLER_ARMED && dribbler.speedReached)
				{
					vTaskDelay(250/portTICK_RATE_MS); // ball needs some time to accelerate

					ConsolePrint("Dribbler delay to fire\r\n");
				}
			}
		}
		else
		{
			// queue was empty and wait timed out
			ATMegaExchange(&kicker.atmega, kicker.data.exchange.mosi, kicker.data.exchange.miso, KICKER_DATA_SIZE);

			kicker.data.cmd.type = KICKER_CMD_NONE;
		}

		if(kicker.state == KICKER_STATE_DRIBBLER_ARMED)
		{
			if(dribbler.speedReached && BarrierInterrupted())
			{
				KickerFire(kicker.armed.fireDuration, kicker.armed.device);

				ConsolePrint("Dribbler armed fire\r\n");
			}
		}

		if(kicker.state == KICKER_STATE_IR_ARMED && BarrierInterrupted())
		{
			KickerFire(kicker.armed.fireDuration, kicker.armed.device);
		}

		if(kicker.doDischarge)
		{
			if(kicker.data.uCap > 200)
			{
				if(xTaskGetTickCount() - kicker.lastKick > 200/portTICK_RATE_MS)
					KickerFire(1000, KICKER_DEVICE_STRAIGHT);
			}
			else
			{
				kicker.doDischarge = 0;
			}
		}

		kicker.latestStatus.chg = kicker.data.iChg;
		kicker.latestStatus.cap = kicker.data.uCap;
		kicker.latestStatus.ir = motorIr.atData.irLevel;

		if(xLastWakeTime - xLastStatusTime > 50/portTICK_RATE_MS && kicker.logEnabled)
		{
			xLastStatusTime = xLastWakeTime;

			PacketHeader header;
			header.section = SECTION_KICKER;
			header.cmd = CMD_KICKER_STATUSV3;
			header.len = LEN_KICKER_STATUSV3;

			NetworkSendPacket(&header, (uint8_t*)&kicker.latestStatus);
		}

		if(xLastWakeTime - xLastIcStatusTime > 200/portTICK_RATE_MS)
		{
			// send config to media
			IntercomHeader icHeader;
			icHeader.address = INTERCOM_ADDRESS_MEDIA;
			icHeader.section = INTERCOM_SECTION_PRESENTER;
			icHeader.command = IC_PRESENTER_KICKER_STATUS;

			IntercomSend(&intercom1, &icHeader, sizeof(KickerStatusV3), (uint8_t*)&kicker.latestStatus);
		}
	}
}
