/*
 * dect.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "network.h"

#include "stm32f4xx.h"
#include "misc.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_usart.h"
#include "stm32f4xx_exti.h"
#include "stm32f4xx_syscfg.h"

#include "task.h"

#include "util/bit_banding.h"

#include "util/console.h"
#include "errors.h"
#include "util/mem.h"
#include "commands.h"
#include "kicker.h"
#include "motor_ir.h"
#include "spi1_2_3.h"
#include "hal/exti.h"
#include "dribbler.h"
#include "ctrl_motor.h"
#include "ctrl_fusion.h"
#include "ctrl_fusion_vel.h"
#include "util/float16_t.h"
#include "intercom_constants.h"
#include "util/critical.h"
#include "intercom1.h"
#include "config.h"
#include "util/boot.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

Network network;

static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf);

static void nrfIrq()
{
	NRF24IRQ(&network.nrf24);
}

static void nrf24Callback(NRF24* pNRF24)
{
	uint16_t len;
	int16_t result = 0;

	while(FifoPacketSyncedNumPackets(&pNRF24->rxFifo) > 0)
	{
		result = FifoPacketSyncedPop(&pNRF24->rxFifo, &len, network.rxProcessingBuf, NETWORK_RX_PROCESSING_SIZE);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
			FifoPacketSyncedDrop(&pNRF24->rxFifo);
		if(result)
			continue;
		if(len < PACKET_HEADER_SIZE)
			continue;

		PacketHeader* pHeader = (PacketHeader*)network.rxProcessingBuf;
		uint8_t* pData = network.rxProcessingBuf+PACKET_HEADER_SIZE;

		processPacket(pHeader, pData);
	}
}

void NetworkInit()
{
	// NRF24 IRQ line setup
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = GPIO_Pin_10;
	GPIO_Init(GPIOE, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE, EXTI_PinSource10);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line10;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Falling;
	EXTI_Init(&extiInit);

	ExtiSetCallback(&nrfIrq, 10);

	// NRF24 setup
	NRF24Data data;

	static uint8_t rxData[NRF24_RX_SIZE];
	static uint8_t txData[NRF24_TX_SIZE];

	data.pDataRx = rxData;
	data.pDataTx = txData;
	data.rxSize = NRF24_RX_SIZE;
	data.txSize = NRF24_TX_SIZE;
	data.pSPI = &spi1;
	data.rx = 1;
	data.channel = config.channel;
	data.id = config.botId;
	data.speed = config.speed;
	data.maxRt = 1;
	data.pCEPort = GPIOA;
	data.cePin = GPIO_Pin_3;
	data.pCSPort = GPIOB;
	data.csPin = GPIO_Pin_0;

	NRF24Init(&network.nrf24, &data);

	NRF24SetRxCallback(&network.nrf24, &nrf24Callback);

	// other networking stuff setup
	vSemaphoreCreateBinary(network.hNewPacket);
	vSemaphoreCreateBinary(network.hSendPacket);
	network.hPrintMutex = xSemaphoreCreateMutex();

	network.posCb = 0;
}

void stopAll()
{
	// TODO: replace with new controller
/*	MovementSetAutomatic(0, 0, 0, 0);

	for(uint8_t i = 0; i < 4; i++)
	{
		MovementSetPidSetpoint(i, 0);
		MovementSetManual(i, 0);
	}

	MovementSetDribble(0);
*/

	KickerDisarm();

	KickerLoadManual(6, 78, 1);
	vTaskDelay(200/portTICK_RATE_MS);
	KickerFire(5e6f, KICKER_DEVICE_STRAIGHT);
}

void NetworkSetPosCallback(PosCallback cb)
{
	network.posCb = cb;
}

int16_t NetworkSendPacket(PacketHeader* pHeader, uint8_t* pData)
{
	if(pHeader->len + PACKET_HEADER_SIZE > NETWORK_TX_PROCESSING_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	xSemaphoreTake(network.hSendPacket, portMAX_DELAY);	// prevent re-entering

	MemCpy(pHeader->_data, network.txProcessingBuf, PACKET_HEADER_SIZE);
	MemCpy(pData, network.txProcessingBuf + PACKET_HEADER_SIZE, pHeader->len);

	// data to send is now in network.txProcessingBuf
	NRF24Send(&network.nrf24, pHeader->len+PACKET_HEADER_SIZE, network.txProcessingBuf);

	xSemaphoreGive(network.hSendPacket);

	return 0;
}

int32_t NetworkPrint(const char* fmt, ...)
{
	va_list args;
	int32_t bytesWritten;

	TASK_MUTEX(network.hPrintMutex)
	{
		va_start(args, fmt);
		bytesWritten = vsnprintf((char*)network.netPrint.text, sizeof(network.netPrint.text), fmt, args);
		va_end(args);

		if(bytesWritten > sizeof(network.netPrint.text))
			bytesWritten = sizeof(network.netPrint.text);

		network.netPrint.length = bytesWritten;
		network.netPrint.source = INTERCOM_ADDRESS_MAIN;

		PacketHeader header;
		header.cmd = CMD_SYSTEM_CONSOLE_PRINT;
		header.section = SECTION_SYSTEM;
		header.len = LEN_SYSTEM_CONSOLE_PRINT;

		NetworkSendPacket(&header, (uint8_t*)&network.netPrint);
	}

	return bytesWritten;
}

// this is the main network processing function
static uint8_t processPacket(PacketHeader* pHeader, uint8_t* pBuf)
{
	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_PING:
				{
					SystemPing* pPing = (SystemPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;
					header.len = LEN_SYSTEM_PONG;

					NetworkSendPacket(&header, (uint8_t*)&pong);

					return 1;
				}
				break;
				case CMD_SYSTEM_BIG_PING:
				{
					SystemBigPing* pPing = (SystemBigPing*)pBuf;

					SystemPong pong;
					pong.id = pPing->id;

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_PONG;
					header.len = LEN_SYSTEM_PONG;

					NetworkSendPacket(&header, (uint8_t*)&pong);

					return 1;
				}
				break;
				case CMD_SYSTEM_SET_LOGS:
				{
					SystemSetLogs* pLogs = (SystemSetLogs*)pBuf;

					for(uint8_t i = 0; i < 4; i++)
						CtrlMotorSetMotorLog(i, pLogs->motor[i]);

					DribblerSetLog(pLogs->motor[4]);
					KickerSetLogging(pLogs->kicker);
					CtrlSetLog(pLogs->movement);

					// old controller
					CtrlMotorSetLog(pLogs->movement);

					return 1;
				}
				break;
				case CMD_SYSTEM_CONSOLE_COMMAND:
				{
					SystemConsoleCommand* pCmd = (SystemConsoleCommand*)pBuf;

					if(pCmd->target == INTERCOM_ADDRESS_MAIN)
					{
						// process onboard
						int16_t result = FifoPacketSyncedBegin(&console.rxFifo);
						if(result)
							return 0;

						result = FifoPacketSyncedPush(&console.rxFifo, pCmd->length, pCmd->text);
						if(result)
						{
							FifoPacketSyncedCancel(&console.rxFifo);
							return 0;
						}

						FifoPacketSyncedComplete(&console.rxFifo);
					}

					if(pCmd->target == INTERCOM_ADDRESS_MEDIA)
					{
						// forward to intercom
						IntercomHeader header;
						header.address = INTERCOM_ADDRESS_MEDIA;
						header.command = 0;
						header.section = INTERCOM_SECTION_CONSOLE;

						IntercomSend(&intercom1, &header, pCmd->length, pCmd->text);
					}
				}
				break;
			}
		}
		break;
		case SECTION_DECT:
		{
		}
		break;
		case SECTION_KICKER:
		{
			switch(pHeader->cmd)
			{
				case CMD_KICKER_KICKV2:
				{
					KickerKickV2* pKick = (KickerKickV2*)pBuf;

					FixedPoint142* pDuration = (FixedPoint142*)&pKick->duration;
					float time = pDuration->value;
					if(pDuration->exp == 1)
						time *= 10;
					if(pDuration->exp == 2)
						time *= 100;
					if(pDuration->exp == 3)
						time *= 1000;
					time /= 10.0f;

					if(pKick->mode == KICKER_MODE_FORCE)
					{
						KickerFire(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_ARM)
					{
						KickerArmIr(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DRIBBLER)
					{
						KickerArmDribbler(time, pKick->device);
					}

					if(pKick->mode == KICKER_MODE_DISARM)
					{
						KickerDisarm();
					}

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_MANUAL:
				{
					KickerChargeManual* pCharge = (KickerChargeManual*)pBuf;

					KickerLoadManual(pCharge->tOnTicks, pCharge->tOffTicks, pCharge->duration);

					return 1;
				}
				break;
				case CMD_KICKER_CHARGE_AUTO:
				{
					KickerChargeAuto* pCharge = (KickerChargeAuto*)pBuf;

					KickerLoadAuto(pCharge->maxCapLevel);

					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_MOTOR:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOTOR_MOVE:
				{
					MotorMove* pMove = (MotorMove*)pBuf;

					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
				case CMD_MOTOR_MOVE_V2:
				{
					MotorMoveV2* pMove = (MotorMoveV2*)pBuf;

					CtrlMotorSetTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);
					CtrlFusionVelSetVelTarget(pMove->vx*0.001f, pMove->vy*0.001f, pMove->vw*0.001f);

					return 1;
				}
				break;
				case CMD_MOTOR_DRIBBLE:
				{
					MotorDribble* pDribble = (MotorDribble*)pBuf;

					DribblerSetTarget(pDribble->rpm);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_PARAMS:
				{
					MotorSetPidParams* pParams = (MotorSetPidParams*)pBuf;

					if(pParams->id == 4)
					{
						DribblerSetParams(pParams->Kp*0.0001f, pParams->Ki*0.0001f, pParams->Kd*0.0001f);
					}
					else
					{
						CtrlMotorSetParams(pParams->id, pParams->Kp*0.0001f, pParams->Ki*0.0001f, pParams->Kd*0.0001f);
					}

					return 1;
				}
				break;
				case CMD_MOTOR_SET_MODE:
				{
					MotorSetMode* pMode = (MotorSetMode*)pBuf;

					ConsolePrint("MotorSetMode\r\n");

					// TODO: supported any more?
//					MovementSetMode(pMode->mode);
//					MovementSetLog(pMode->id, pMode->log);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_MANUAL:
				{
					MotorSetManual* pManual = (MotorSetManual*)pBuf;

					// TODO: support?
					// MovementSetManual(pManual->id, pManual->power);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PID_SP:
				{
					MotorSetPidSp* pSp = (MotorSetPidSp*)pBuf;

					if(pSp->id == 4)
						DribblerSetTarget(pSp->setpoint);

					// TODO: support
					// MovementSetPidSetpoint(pSp->id, pSp->setpoint);

					return 1;
				}
				break;
				case CMD_MOTOR_SET_PARAMS:
				{
					MotorSetParams* pParams = (MotorSetParams*)pBuf;

					// TODO: mode support?
					// MovementSetMode(pParams->mode);

					DribblerSetParams(pParams->Kp[4]*0.0001f, pParams->Ki[4]*0.0001f, pParams->Kd[4]*0.0001f);

					for(uint8_t i = 0; i < 4; i++)
					{
						CtrlMotorSetParams(i, pParams->Kp[i]*0.0001f, pParams->Ki[i]*0.0001f, pParams->Kd[i]*0.0001f);
					}

					return 1;
				}
				break;
			}
		}
		break;
		case SECTION_MULTICAST:
		{
		}
		break;
		case SECTION_MOVEMENT:
		{
			switch(pHeader->cmd)
			{
				case CMD_MOVEMENT_SET_PARAMS:
				{
				}
				break;
			}
		}
		break;
		case SECTION_CTRL:
		{
			switch(pHeader->cmd)
			{
				case CMD_CTRL_VISION_POS:
				{
					CtrlVisionPos* pPos = (CtrlVisionPos*)pBuf;

					memcpy(network.lastVisionPos, pPos->pos, sizeof(int16_t)*3);

					if(network.posCb)
						(*network.posCb)();
				}
				break;
				case CMD_CTRL_SPLINE_2D:
				{
					CtrlSpline2D* pSpline = (CtrlSpline2D*)pBuf;

					CtrlFusionAddSplineCmdXY(pSpline);
				}
				break;
				case CMD_CTRL_SPLINE_1D:
				{
					CtrlSpline1D* pSpline = (CtrlSpline1D*)pBuf;

					CtrlFusionAddSplineCmdW(pSpline);
				}
				break;
				case CMD_CTRL_SET_FILTER_PARAMS:
				{
					CtrlSetFilterParams* pParams = (CtrlSetFilterParams*)pBuf;

					CtrlFusionSetFilterParams(pParams);
					CtrlFusionVelSetFilterParams(pParams);
				}
				break;
				case CMD_CTRL_SET_PID_PARAMS:
				{
					CtrlSetPIDParams* pParams = (CtrlSetPIDParams*)pBuf;

					CtrlFusionSetPIDParams(pParams);
					CtrlFusionVelSetPIDParams(pParams);
				}
				break;
				case CMD_CTRL_SET_CONTROLLER_TYPE:
				{
					CtrlSetControllerType* pCtrl = (CtrlSetControllerType*)pBuf;

					switch(pCtrl->controllerId)
					{
						case CTRL_TYPE_FUSION_VEL:
						{
							CtrlSetController(&ctrlFusionVelInstance);
						}
						break;
						case CTRL_TYPE_FUSION:
						{
							CtrlSetController(&ctrlFusionInstance);
						}
						break;
						default:
						{
							CtrlSetController(0);
						}
						break;
					}
				}
				break;
			}
		}
		break;
		case SECTION_BOOTLOADER:
		{
			switch(pHeader->cmd)
			{
				case CMD_BOOTLOADER_COMMAND:
				{
					BootloaderCommand* pCmd = (BootloaderCommand*)pBuf;

					switch(pCmd->cmd)
					{
						case BOOTLOADER_CMD_MODE_QUERY:
						{
							BootloaderResponse response;
							response.type = BOOTLOADER_RESPONSE_MODE_NORMAL;
							response.offset = 0;

							PacketHeader header;
							header.section = SECTION_BOOTLOADER;
							header.cmd = CMD_BOOTLOADER_RESPONSE;
							header.len = LEN_BOOTLOADER_RESPONSE;

							NetworkSendPacket(&header, (uint8_t*)&response);
						}
						break;
						case BOOTLOADER_CMD_ENTER:
						{
							BootStartBootloader(&intercom1);
						}
						break;
					}
				}
				break;
			}
		}
		break;
	}

	return 0;
}
