#include "fatfs/diskio.h"
#include "hal/rtc.h"
#include "usb.h"
#include "usbh_msc.h"

#include "FreeRTOS.h"
#include "task.h"
/*--------------------------------------------------------------------------

 Module Private Functions and Variables

 ---------------------------------------------------------------------------*/

static volatile DSTATUS Stat = STA_NOINIT; /* Disk status */

/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize(BYTE drv)
{
	(void)drv;

	if (HCD_IsDeviceConnected(&USB_OTG_Core))
	{
		Stat &= ~STA_NOINIT;
	}

	return Stat;

}

/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status(BYTE drv /* Physical drive number (0) */
)
{
	if (drv)
		return STA_NOINIT; /* Supports only single drive */

	return Stat;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read(BYTE drv, /* Physical drive number (0) */
BYTE *buff, /* Pointer to the data buffer to store read data */
DWORD sector, /* Start sector number (LBA) */
BYTE count /* Sector count (1..255) */
)
{
	if (drv || !count)
		return RES_PARERR;
	if (Stat & STA_NOINIT)
		return RES_NOTRDY;

	if (!HCD_IsDeviceConnected(&USB_OTG_Core))
		return RES_NOTRDY;

	if(USBH_MSC_Read10(&USB_OTG_Core, &USB_Host, buff, sector, 512 * count) == USBH_MSC_OK)
		return RES_OK;
	else
		return RES_ERROR;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _READONLY == 0
DRESULT disk_write(BYTE drv, /* Physical drive number (0) */
const BYTE *buff, /* Pointer to the data to be written */
DWORD sector, /* Start sector number (LBA) */
BYTE count /* Sector count (1..255) */
)
{
	if (drv || !count)
		return RES_PARERR;
	if (Stat & STA_NOINIT)
		return RES_NOTRDY;
	if (Stat & STA_PROTECT)
		return RES_WRPRT;

	if (!HCD_IsDeviceConnected(&USB_OTG_Core))
		return RES_NOTRDY;

	if(USBH_MSC_Write10(&USB_OTG_Core, &USB_Host, (uint8_t*)buff, sector, 512 * count) == USBH_MSC_OK)
		return RES_OK;
	else
		return RES_ERROR;
}
#endif /* _READONLY == 0 */

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL != 0
DRESULT disk_ioctl(BYTE drv, /* Physical drive number (0) */
BYTE ctrl, /* Control code */
void *buff /* Buffer to send/receive control data */
)
{
	DRESULT res = RES_OK;

	if (drv)
		return RES_PARERR;

	res = RES_ERROR;

	if (Stat & STA_NOINIT)
		return RES_NOTRDY;

	switch (ctrl)
	{
	case CTRL_SYNC: /* Make sure that no pending write process */

		res = RES_OK;
		break;

	case GET_SECTOR_COUNT: /* Get number of sectors on the disk (DWORD) */

		*(DWORD*) buff = (DWORD) usbhMsc.storageParams.capacity;
		res = RES_OK;
		break;

	case GET_SECTOR_SIZE: /* Get R/W sector size (WORD) */
		*(WORD*) buff = 512;
		res = RES_OK;
		break;

	case GET_BLOCK_SIZE: /* Get erase block size in unit of sector (DWORD) */

		*(DWORD*) buff = 512;

		break;

	default:
		res = RES_PARERR;
		break;
	}

	return res;
}

// implementation for FatFs support
DWORD get_fattime(void)
{
//  bit 31:25
//    Year from 1980 (0..127)
//  bit 24:21
//    Month (1..12)
//  bit 20:16
//    Day in month(1..31)
//  bit 15:11
//    Hour (0..23)
//  bit 10:5
//    Minute (0..59)
//  bit 4:0
//    Second / 2 (0..29)

	uint32_t packed = 0x00000000;

	DateTime now = RTCGetDateTime();

	packed |= ((now.year - 1980) & 0x7F) << 25;
	packed |= ((now.month) & 0x0F) << 21;
	packed |= ((now.day) & 0x1F) << 16;
	packed |= ((now.hour) & 0x1F) << 11;
	packed |= ((now.min) & 0x3F) << 5;
	packed |= ((now.sec / 2) & 0x1F) << 0;

	return packed;
}

#endif /* _USE_IOCTL != 0 */
