/**
 ******************************************************************************
 * @file    usb_hcd_int.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   Host driver interrupt subroutines
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "usb_core.h"
#include "usb_defines.h"
#include "usb_hcd_int.h"

#include "FreeRTOS.h"
#include "task.h"

#if defined (__GNUC__) /*!< GNU Compiler */
#pragma GCC optimize ("O0")
#endif

#define CLEAR_HC_INT(HC_REGS, intr) \
  {\
  USB_OTG_HCINTn_TypeDef  hcint_clear; \
  hcint_clear.d32 = 0; \
  hcint_clear.b.intr = 1; \
  USB_OTG_WRITE_REG32(&((HC_REGS)->HCINT), hcint_clear.d32);\
  }\

#define MASK_HOST_INT_CHH(hc_num) { USB_OTG_HCINTMSK_TypeDef  INTMSK; \
    INTMSK.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK); \
    INTMSK.b.chhltd = 0; \
    USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK, INTMSK.d32);}

#define UNMASK_HOST_INT_CHH(hc_num) { USB_OTG_HCINTMSK_TypeDef  INTMSK; \
    INTMSK.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK); \
    INTMSK.b.chhltd = 1; \
    USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK, INTMSK.d32);}

#define MASK_HOST_INT_ACK(hc_num) { USB_OTG_HCINTMSK_TypeDef  INTMSK; \
    INTMSK.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK); \
    INTMSK.b.ack = 0; \
    USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK, GINTMSK.d32);}

#define UNMASK_HOST_INT_ACK(hc_num) { USB_OTG_HCGINTMSK_TypeDef  INTMSK; \
    INTMSK.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK); \
    INTMSK.b.ack = 1; \
    USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCINTMSK, INTMSK.d32);}


static uint32_t USB_OTG_USBH_handle_sof_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_port_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_hc_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_hc_n_In_ISR(USB_OTG_CORE_HANDLE *pdev, uint32_t num);
static uint32_t USB_OTG_USBH_handle_hc_n_Out_ISR(USB_OTG_CORE_HANDLE *pdev, uint32_t num);
static uint32_t USB_OTG_USBH_handle_rx_qlvl_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_nptxfempty_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_ptxfempty_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_Disconnect_ISR(USB_OTG_CORE_HANDLE *pdev);
static uint32_t USB_OTG_USBH_handle_IncompletePeriodicXfer_ISR(USB_OTG_CORE_HANDLE *pdev);

/**
 * @brief  USB_OTG_ReadPacket : Reads a packet from the Rx FIFO
 * @param  pdev : Selected device
 * @param  dest : Destination Pointer
 * @param  bytes : No. of bytes
 * @retval None
 */
static void* USB_OTG_ReadPacket(USB_OTG_CORE_HANDLE *pdev, uint8_t *dest, uint16_t len)
{
	uint32_t i = 0;
	uint32_t count32b = (len + 3) / 4;

	__IO uint32_t *fifo = pdev->regs.DFIFO[0];

	for(i = 0; i < count32b; i++, dest += 4)
	{
		*(__attribute__((packed))  uint32_t *)dest = USB_OTG_READ_REG32(fifo);

	}
	return ((void *)dest);
}

/**
 * @brief  USB_OTG_ReadDropPacket : Reads a packet from the Rx FIFO and drops it to a temp buffer
 * @param  pdev : Selected device
 * @param  bytes : No. of bytes
 * @retval None
 */
static void USB_OTG_ReadDropPacket(USB_OTG_CORE_HANDLE *pdev, uint16_t len)
{
	uint32_t i = 0;
	uint32_t count32b = (len + 3) / 4;
	volatile uint32_t temp;

	__IO uint32_t *fifo = pdev->regs.DFIFO[0];

	for(i = 0; i < count32b; i++)
	{
		temp = USB_OTG_READ_REG32(fifo);
		temp = temp;
	}
}

/**
 * @brief  USB_OTG_ReadCoreItr : returns the Core Interrupt register
 * @param  pdev : Selected device
 * @retval Status
 */
static uint32_t USB_OTG_ReadCoreItr(USB_OTG_CORE_HANDLE *pdev)
{
	uint32_t v = 0;
	v = USB_OTG_READ_REG32(&pdev->regs.GREGS->GINTSTS);
	v &= USB_OTG_READ_REG32(&pdev->regs.GREGS->GINTMSK);
	return v;
}

/**
 * @brief  USB_OTG_GetMode : Get current mode
 * @param  pdev : Selected device
 * @retval current mode
 */
static uint32_t USB_OTG_GetMode(USB_OTG_CORE_HANDLE *pdev)
{
	return (USB_OTG_READ_REG32(&pdev->regs.GREGS->GINTSTS ) & 0x1);
}

/**
 * @brief  USB_OTG_IsHostMode : Check if it is host mode
 * @param  pdev : Selected device
 * @retval num_in_ep
 */
static uint8_t USB_OTG_IsHostMode(USB_OTG_CORE_HANDLE *pdev)
{
	return (USB_OTG_GetMode(pdev) == HOST_MODE);
}

/**
 * @brief  Issue a ping token
 * @param  None
 * @retval : None
 */
static USB_OTG_STS USB_OTG_HC_DoPing(USB_OTG_CORE_HANDLE *pdev, uint8_t hc_num)
{
	USB_OTG_STS status = USB_OTG_OK;
	USB_OTG_HCCHAR_TypeDef hcchar;
	USB_OTG_HCTSIZn_TypeDef hctsiz;

	hctsiz.d32 = 0;
	hctsiz.b.dopng = 1;
	hctsiz.b.pktcnt = 1;
	USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCTSIZ, hctsiz.d32);

	hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[hc_num]->HCCHAR);
	hcchar.b.chen = 1;
	hcchar.b.chdis = 0;
	USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[hc_num]->HCCHAR, hcchar.d32);
	return status;
}

/**
 * @brief  USB_OTG_ReadHostAllChannels_intr : Register PCD Callbacks
 * @param  pdev : Selected device
 * @retval Status
 */
static uint32_t USB_OTG_ReadHostAllChannels_intr(USB_OTG_CORE_HANDLE *pdev)
{
	return (USB_OTG_READ_REG32 (&pdev->regs.HREGS->HAINT) );
}

/**
 * @brief  HOST_Handle_ISR 
 *         This function handles all USB Host Interrupts
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
uint32_t USBH_OTG_ISR_Handler(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GINTSTS_TypeDef gintsts;
	uint32_t retval = 0;

	gintsts.d32 = 0;

	/* Check if HOST Mode */
	if(USB_OTG_IsHostMode(pdev))
	{
		gintsts.d32 = USB_OTG_ReadCoreItr(pdev);
		if(!gintsts.d32)
		{
			return 0;
		}

		if(gintsts.b.sofintr)
		{
			retval |= USB_OTG_USBH_handle_sof_ISR(pdev);
		}

		if(gintsts.b.rxstsqlvl)
		{
			retval |= USB_OTG_USBH_handle_rx_qlvl_ISR(pdev);
		}

		if(gintsts.b.nptxfempty)
		{
			retval |= USB_OTG_USBH_handle_nptxfempty_ISR(pdev);
		}

		if(gintsts.b.ptxfempty)
		{
			retval |= USB_OTG_USBH_handle_ptxfempty_ISR(pdev);
		}

		if(gintsts.b.hcintr)
		{
			retval |= USB_OTG_USBH_handle_hc_ISR(pdev);
		}

		if(gintsts.b.portintr)
		{
			retval |= USB_OTG_USBH_handle_port_ISR(pdev);
		}

		if(gintsts.b.disconnect)
		{
			retval |= USB_OTG_USBH_handle_Disconnect_ISR(pdev);

		}

		if(gintsts.b.incomplisoout)
		{
			retval |= USB_OTG_USBH_handle_IncompletePeriodicXfer_ISR(pdev);
		}

	}
	return retval;
}

/**
 * @brief  USB_OTG_USBH_handle_hc_ISR 
 *         This function indicates that one or more host channels has a pending
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_hc_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_HAINT_TypeDef haint;
	USB_OTG_HCCHAR_TypeDef hcchar;
	uint32_t i = 0;
	uint32_t retval = 0;

	/* Clear appropriate bits in HCINTn to clear the interrupt bit in
	 * GINTSTS */

	haint.d32 = USB_OTG_ReadHostAllChannels_intr(pdev);

	for(i = 0; i < pdev->cfg.host_channels; i++)
	{
		if(haint.b.chint & (1 << i))
		{
			hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[i]->HCCHAR);

			if(hcchar.b.epdir)
			{
				retval |= USB_OTG_USBH_handle_hc_n_In_ISR(pdev, i);
			}
			else
			{
				retval |= USB_OTG_USBH_handle_hc_n_Out_ISR(pdev, i);
			}
		}
	}

	return retval;
}

/**
 * @brief  USB_OTG_otg_hcd_handle_sof_intr 
 *         Handles the start-of-frame interrupt in host mode.
 * @param  pdev: Selected device
 * @retval status
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_sof_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GINTSTS_TypeDef gintsts;
	gintsts.d32 = 0;

	if(pdev->host.hcdCb && pdev->host.hcdCb->SOF)
		pdev->host.hcdCb->SOF(pdev);

	/* Clear interrupt */
	gintsts.b.sofintr = 1;
	USB_OTG_WRITE_REG32(&pdev->regs.GREGS->GINTSTS, gintsts.d32);

	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_Disconnect_ISR 
 *         Handles disconnect event.
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_Disconnect_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GINTSTS_TypeDef gintsts;

	gintsts.d32 = 0;

	if(pdev->host.hcdCb && pdev->host.hcdCb->DevDisconnected)
		pdev->host.hcdCb->DevDisconnected(pdev);

	/* Clear interrupt */
	gintsts.b.disconnect = 1;
	USB_OTG_WRITE_REG32(&pdev->regs.GREGS->GINTSTS, gintsts.d32);

	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_nptxfempty_ISR 
 *         Handles non periodic tx fifo empty.
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_nptxfempty_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GINTMSK_TypeDef intmsk;
	USB_OTG_HNPTXSTS_TypeDef hnptxsts;
	uint16_t len_words, len;

	hnptxsts.d32 = USB_OTG_READ_REG32(&pdev->regs.GREGS->HNPTXSTS);

	uint32_t ch = hnptxsts.b.chnum;	// TODO: This assumes that the endpoint the MAC currently processes does not change while writing to nptx fifo
	URB* pURB = pdev->host.hc[ch].pURB;

	if(pURB == 0)
	{
		intmsk.d32 = 0;
		intmsk.b.nptxfempty = 1;
		USB_OTG_MODIFY_REG32( &pdev->regs.GREGS->GINTMSK, intmsk.d32, 0);	// TODO: what about other nptx endpoints? this disables all nptx IRQs

		return 1;
	}

	uint32_t bytesRemaining = pURB->length - pURB->transferred;
	uint8_t* pDataProc = pURB->pData + pURB->transferred;

	len_words = (bytesRemaining + 3) / 4;

	while((hnptxsts.b.nptxfspcavail > len_words) && (bytesRemaining != 0))
	{
		len = hnptxsts.b.nptxfspcavail * 4;

		if(len > bytesRemaining)
		{
			/* Last packet */
			len = bytesRemaining;

			intmsk.d32 = 0;
			intmsk.b.nptxfempty = 1;
			USB_OTG_MODIFY_REG32( &pdev->regs.GREGS->GINTMSK, intmsk.d32, 0);	// TODO: what about other nptx endpoints? this disables all nptx IRQs
		}

		USB_OTG_WritePacket(pdev, pDataProc, ch, len);

		pDataProc += len;
		bytesRemaining -= len;
		pURB->transferred += len;

		len_words = (bytesRemaining + 3) / 4;

		hnptxsts.d32 = USB_OTG_READ_REG32(&pdev->regs.GREGS->HNPTXSTS);
	}

	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_ptxfempty_ISR 
 *         Handles periodic tx fifo empty
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_ptxfempty_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GINTMSK_TypeDef intmsk;
	USB_OTG_HPTXSTS_TypeDef hptxsts;
	uint16_t len_words, len;

	hptxsts.d32 = USB_OTG_READ_REG32(&pdev->regs.HREGS->HPTXSTS);

	uint32_t ch = hptxsts.b.chnum;	// TODO: This assumes that the endpoint the MAC currently processes does not change while writing to ptx fifo
	URB* pURB = pdev->host.hc[ch].pURB;

	if(pURB == 0)
		return 1;

	uint32_t bytesRemaining = pURB->length - pURB->transferred;
	uint8_t* pDataProc = pURB->pData + pURB->transferred;

	len_words = (bytesRemaining + 3) / 4;

	while((hptxsts.b.ptxfspcavail > len_words) && (bytesRemaining != 0))
	{
		len = hptxsts.b.ptxfspcavail * 4;

		if(len > bytesRemaining)
		{
			/* Last packet */
			len = bytesRemaining;

			intmsk.d32 = 0;
			intmsk.b.ptxfempty = 1;
			USB_OTG_MODIFY_REG32( &pdev->regs.GREGS->GINTMSK, intmsk.d32, 0);	// TODO: what about other ptx endpoints? this disables all ptx IRQs
		}

		USB_OTG_WritePacket(pdev, pDataProc, ch, len);

		pDataProc += len;
		bytesRemaining -= len;
		pURB->transferred += len;

		len_words = (bytesRemaining + 3) / 4;

		hptxsts.d32 = USB_OTG_READ_REG32(&pdev->regs.HREGS->HPTXSTS);
	}

	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_rx_qlvl_ISR
 *         Handles the Rx Status Queue Level Interrupt
 * @param  pdev: Selected device
 * @retval status
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_rx_qlvl_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_GRXFSTS_TypeDef grxsts;
	USB_OTG_GINTMSK_TypeDef intmsk;
	USB_OTG_HCTSIZn_TypeDef hctsiz;
	USB_OTG_HCCHAR_TypeDef hcchar;
	__IO uint8_t channelnum = 0;
//	uint32_t count;

	/* Disable the Rx Status Queue Level interrupt */
	intmsk.d32 = 0;
	intmsk.b.rxstsqlvl = 1;
	USB_OTG_MODIFY_REG32( &pdev->regs.GREGS->GINTMSK, intmsk.d32, 0);

	grxsts.d32 = USB_OTG_READ_REG32(&pdev->regs.GREGS->GRXSTSP);
	channelnum = grxsts.b.chnum;
	hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[channelnum]->HCCHAR);

	URB* pURB = pdev->host.hc[channelnum].pURB;

	if(pURB == 0)
	{
		/* Enable the Rx Status Queue Level interrupt */
		intmsk.b.rxstsqlvl = 1;
		USB_OTG_MODIFY_REG32(&pdev->regs.GREGS->GINTMSK, 0, intmsk.d32);
		return 1;
	}

	uint32_t bytesRemaining = pURB->length - pURB->transferred;
	uint8_t* pDataProc = pURB->pData + pURB->transferred;

	switch(grxsts.b.pktsts)
	{
		case GRXSTS_PKTSTS_IN:
			/* Read the data into the host buffer. */
			if((grxsts.b.bcnt > 0) && (pDataProc != (void *)0))
			{
				if(pURB == 0)
				{
					USB_OTG_ReadDropPacket(pdev, grxsts.b.bcnt);
				}
				else
				{
					if(grxsts.b.bcnt > bytesRemaining)
					{
						USB_OTG_ReadPacket(pdev, pDataProc, bytesRemaining);
						USB_OTG_ReadDropPacket(pdev, grxsts.b.bcnt-bytesRemaining);
					}
					else
					{
						USB_OTG_ReadPacket(pdev, pDataProc, grxsts.b.bcnt);
					}
				}

				/*manage multiple Xfer */
				pDataProc += grxsts.b.bcnt;
				pURB->transferred += grxsts.b.bcnt;

//				count = pdev->host.hc[channelnum].xfer_count;
//				pdev->host.XferCnt[channelnum] = count;

				hctsiz.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[channelnum]->HCTSIZ);
				if(hctsiz.b.pktcnt > 0)
				{
					/* re-activate the channel when more packets are expected */
					hcchar.b.chen = 1;
					hcchar.b.chdis = 0;
					USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[channelnum]->HCCHAR, hcchar.d32);
				}
			}
			break;

		case GRXSTS_PKTSTS_IN_XFER_COMP:

		case GRXSTS_PKTSTS_DATA_TOGGLE_ERR:
		case GRXSTS_PKTSTS_CH_HALTED:
		default:
			break;
	}

	/* Enable the Rx Status Queue Level interrupt */
	intmsk.b.rxstsqlvl = 1;
	USB_OTG_MODIFY_REG32(&pdev->regs.GREGS->GINTMSK, 0, intmsk.d32);
	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_port_ISR 
 *         This function determines which interrupt conditions have occurred
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_port_ISR(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_HPRT0_TypeDef hprt0;
	USB_OTG_HPRT0_TypeDef hprt0_dup;
	USB_OTG_HCFG_TypeDef hcfg;
	uint32_t do_reset = 0;
	uint32_t retval = 0;

	hcfg.d32 = 0;
	hprt0.d32 = 0;
	hprt0_dup.d32 = 0;

	hprt0.d32 = USB_OTG_READ_REG32(pdev->regs.HPRT0);
	hprt0_dup.d32 = USB_OTG_READ_REG32(pdev->regs.HPRT0);

	/* Clear the interrupt bits in GINTSTS */

	hprt0_dup.b.prtena = 0;
	hprt0_dup.b.prtconndet = 0;
	hprt0_dup.b.prtenchng = 0;
	hprt0_dup.b.prtovrcurrchng = 0;

	/* Port Connect Detected */
	if(hprt0.b.prtconndet)
	{

		hprt0_dup.b.prtconndet = 1;

		if(pdev->host.hcdCb && pdev->host.hcdCb->DevConnected)
			pdev->host.hcdCb->DevConnected(pdev);

		retval |= 1;
	}

	/* Port Enable Changed */
	if(hprt0.b.prtenchng)
	{
		hprt0_dup.b.prtenchng = 1;

		if(hprt0.b.prtena == 1)
		{

			if(pdev->host.hcdCb && pdev->host.hcdCb->DevConnected)
				pdev->host.hcdCb->DevConnected(pdev);

			if((hprt0.b.prtspd == HPRT0_PRTSPD_LOW_SPEED) || (hprt0.b.prtspd == HPRT0_PRTSPD_FULL_SPEED))
			{

				hcfg.d32 = USB_OTG_READ_REG32(&pdev->regs.HREGS->HCFG);

				if(hprt0.b.prtspd == HPRT0_PRTSPD_LOW_SPEED)
				{
					USB_OTG_WRITE_REG32(&pdev->regs.HREGS->HFIR, 6000);
					if(hcfg.b.fslspclksel != HCFG_6_MHZ)
					{
						if(pdev->cfg.phy_itface == USB_OTG_EMBEDDED_PHY)
						{
							USB_OTG_InitFSLSPClkSel(pdev, HCFG_6_MHZ);
						}
						do_reset = 1;
					}
				}
				else
				{

					USB_OTG_WRITE_REG32(&pdev->regs.HREGS->HFIR, 48000);
					if(hcfg.b.fslspclksel != HCFG_48_MHZ)
					{
						USB_OTG_InitFSLSPClkSel(pdev, HCFG_48_MHZ);
						do_reset = 1;
					}
				}
			}
			else
			{
				do_reset = 1;
			}
		}
	}

	/* Overcurrent Change Interrupt */
	if(hprt0.b.prtovrcurrchng)
	{
		hprt0_dup.b.prtovrcurrchng = 1;
		retval |= 1;
	}

	if(do_reset)
	{
		USB_OTG_ResetPort(pdev);
	}

	/* Clear Port Interrupts */
	USB_OTG_WRITE_REG32(pdev->regs.HPRT0, hprt0_dup.d32);

	return retval;
}

/**
 * @brief  USB_OTG_USBH_handle_hc_n_Out_ISR 
 *         Handles interrupt for a specific Host Channel
 * @param  pdev: Selected device
 * @param  hc_num: Channel number
 * @retval status 
 *
 * IRQL: USB IRQ
 */
uint32_t USB_OTG_USBH_handle_hc_n_Out_ISR(USB_OTG_CORE_HANDLE *pdev, uint32_t num)
{
	USB_OTG_HCINTn_TypeDef hcint;
	USB_OTG_HCINTMSK_TypeDef hcintmsk;
	USB_OTG_HC_REGS *hcreg;
	USB_OTG_HCCHAR_TypeDef hcchar;

	hcreg = pdev->regs.HC_REGS[num];
	hcint.d32 = USB_OTG_READ_REG32(&hcreg->HCINT);
	hcintmsk.d32 = USB_OTG_READ_REG32(&hcreg->HCINTMSK);
	hcint.d32 = hcint.d32 & hcintmsk.d32;

	hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[num]->HCCHAR);

	if(hcint.b.ahberr)
	{
		CLEAR_HC_INT(hcreg, ahberr);
		UNMASK_HOST_INT_CHH(num);
	}
	else if(hcint.b.ack)
	{
		CLEAR_HC_INT(hcreg, ack);
	}
	else if(hcint.b.frmovrun)
	{
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, frmovrun);
	}
	else if(hcint.b.xfercompl)
	{
		pdev->host.hc[num].errCnt = 0;
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, xfercompl);
		pdev->host.hc[num].hcStatus = HC_XFRC;
	}

	else if(hcint.b.stall)
	{
		CLEAR_HC_INT(hcreg, stall);
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		pdev->host.hc[num].hcStatus = HC_STALL;
	}

	else if(hcint.b.nak)
	{
		pdev->host.hc[num].errCnt = 0;
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, nak);
		pdev->host.hc[num].hcStatus = HC_NAK;
	}

	else if(hcint.b.xacterr)
	{
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		pdev->host.hc[num].errCnt++;
		pdev->host.hc[num].hcStatus = HC_XACTERR;
		CLEAR_HC_INT(hcreg, xacterr);
	}
	else if(hcint.b.nyet)
	{
		pdev->host.hc[num].errCnt = 0;
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, nyet);
		pdev->host.hc[num].hcStatus = HC_NYET;
	}
	else if(hcint.b.datatglerr)
	{

		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, nak);
		pdev->host.hc[num].hcStatus = HC_DATATGLERR;

		CLEAR_HC_INT(hcreg, datatglerr);
	}
	else if(hcint.b.chhltd)
	{
		MASK_HOST_INT_CHH(num);

		if(pdev->host.hc[num].pURB == 0)
		{
			CLEAR_HC_INT(hcreg, chhltd);
			return 1;
		}

		uint16_t num_packets = (pdev->host.hc[num].pURB->length + pdev->host.hc[num].max_packet - 1) / pdev->host.hc[num].max_packet;

		if(pdev->host.hc[num].hcStatus == HC_XFRC)
		{
			pdev->host.hc[num].pURB->state = URB_DONE;

			if(hcchar.b.eptype == EP_TYPE_BULK || hcchar.b.eptype == EP_TYPE_INTR)
			{
				if(num_packets % 2 == 1)
					pdev->host.hc[num].pid ^= 1;
//				pdev->host.hc[num].toggle_out ^= 1;
			}
		}
		else if(pdev->host.hc[num].hcStatus == HC_NAK)
		{
			pdev->host.hc[num].pURB->state = URB_NOTREADY;
		}
		else if(pdev->host.hc[num].hcStatus == HC_NYET)
		{
			if(pdev->host.hc[num].do_ping == 1)
			{
				USB_OTG_HC_DoPing(pdev, num);
			}

			pdev->host.hc[num].pURB->state = URB_NOTREADY;
		}
		else if(pdev->host.hc[num].hcStatus == HC_STALL)
		{
			pdev->host.hc[num].pURB->state = URB_STALL;
		}
		else if(pdev->host.hc[num].hcStatus == HC_XACTERR)
		{
			if(pdev->host.hc[num].errCnt == 3)
			{
				pdev->host.hc[num].pURB->state = URB_ERROR;
				pdev->host.hc[num].errCnt = 0;
			}
		}

		int32_t yield = pdFALSE;
		xQueueSendFromISR(pdev->host.hEventQueue, &num, &yield);
		if(yield == pdTRUE)
			taskYIELD();

		CLEAR_HC_INT(hcreg, chhltd);
	}

	return 1;
}

/**
 * @brief  USB_OTG_USBH_handle_hc_n_In_ISR 
 *         Handles interrupt for a specific Host Channel
 * @param  pdev: Selected device
 * @param  hc_num: Channel number
 * @retval status 
 *
 * IRQL: USB IRQ
 */
uint32_t USB_OTG_USBH_handle_hc_n_In_ISR(USB_OTG_CORE_HANDLE *pdev, uint32_t num)
{
	USB_OTG_HCINTn_TypeDef hcint;
	USB_OTG_HCINTMSK_TypeDef hcintmsk;
	USB_OTG_HCCHAR_TypeDef hcchar;
	USB_OTG_HCTSIZn_TypeDef hctsiz;
	USB_OTG_HC_REGS *hcreg;

	hcreg = pdev->regs.HC_REGS[num];
	hcint.d32 = USB_OTG_READ_REG32(&hcreg->HCINT);
	hcintmsk.d32 = USB_OTG_READ_REG32(&hcreg->HCINTMSK);
	hcint.d32 = hcint.d32 & hcintmsk.d32;
	hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[num]->HCCHAR);
	hcintmsk.d32 = 0;

	if(hcint.b.ahberr)
	{
		CLEAR_HC_INT(hcreg, ahberr);
		UNMASK_HOST_INT_CHH(num);
	}
	else if(hcint.b.ack)
	{
		CLEAR_HC_INT(hcreg, ack);
	}

	else if(hcint.b.stall)
	{
		UNMASK_HOST_INT_CHH(num);
		pdev->host.hc[num].hcStatus = HC_STALL;
		CLEAR_HC_INT(hcreg, nak);
		/* Clear the NAK Condition */
		CLEAR_HC_INT(hcreg, stall);
		/* Clear the STALL Condition */
		hcint.b.nak = 0; /* NOTE: When there is a 'stall', reset also nak, 
		 else, the pdev->host.HC_Status = HC_STALL
		 will be overwritten by 'nak' in code below */
		USB_OTG_HC_Halt(pdev, num);
	}
	else if(hcint.b.datatglerr)
	{
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, nak);
		pdev->host.hc[num].hcStatus = HC_DATATGLERR;
		CLEAR_HC_INT(hcreg, datatglerr);
	}

	if(hcint.b.frmovrun)
	{
		UNMASK_HOST_INT_CHH(num);
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, frmovrun);
	}

	else if(hcint.b.xfercompl)
	{

		if(pdev->cfg.dma_enable == 1)
		{
			hctsiz.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[num]->HCTSIZ);
			if(pdev->host.hc[num].pURB)
				pdev->host.hc[num].pURB->transferred = pdev->host.hc[num].pURB->length - hctsiz.b.xfersize;

//			pdev->host.XferCnt[num] = pdev->host.hc[num].xfer_len - hctsiz.b.xfersize;
		}

		pdev->host.hc[num].hcStatus = HC_XFRC;
		pdev->host.hc[num].errCnt = 0;
		CLEAR_HC_INT(hcreg, xfercompl);

		uint16_t num_packets = 1;
		if(pdev->host.hc[num].pURB)
			num_packets = (pdev->host.hc[num].pURB->length + pdev->host.hc[num].max_packet - 1) / pdev->host.hc[num].max_packet;

		if((hcchar.b.eptype == EP_TYPE_CTRL) || (hcchar.b.eptype == EP_TYPE_BULK))
		{
			UNMASK_HOST_INT_CHH(num);
			USB_OTG_HC_Halt(pdev, num);
			CLEAR_HC_INT(hcreg, nak);
			if(num_packets % 2 == 1)
				pdev->host.hc[num].pid ^= 1;
//			pdev->host.hc[num].toggle_in ^= 1;

		}
		else if(hcchar.b.eptype == EP_TYPE_INTR)
		{
			hcchar.b.oddfrm = 1;
			USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[num]->HCCHAR, hcchar.d32);
			if(num_packets % 2 == 1)
				pdev->host.hc[num].pid ^= 1;

			if(pdev->host.hc[num].pURB)
				pdev->host.hc[num].pURB->state = URB_DONE;

			int32_t yield = pdFALSE;
			xQueueSendFromISR(pdev->host.hEventQueue, &num, &yield);
			if(yield == pdTRUE)
				taskYIELD();
		}

	}
	else if(hcint.b.chhltd)
	{
		MASK_HOST_INT_CHH(num);

		if(pdev->host.hc[num].pURB == 0)
		{
			CLEAR_HC_INT(hcreg, chhltd);
			return 1;
		}

		if(pdev->host.hc[num].hcStatus == HC_XFRC)
		{
			pdev->host.hc[num].pURB->state = URB_DONE;
		}
		else if(pdev->host.hc[num].hcStatus == HC_STALL)
		{
			pdev->host.hc[num].pURB->state = URB_STALL;
		}
		else if((pdev->host.hc[num].hcStatus == HC_XACTERR) || (pdev->host.hc[num].hcStatus == HC_DATATGLERR))
		{
			pdev->host.hc[num].errCnt = 0;
			pdev->host.hc[num].pURB->state = URB_ERROR;
		}
		else if(pdev->host.hc[num].hcStatus == HC_NAK)
		{
			pdev->host.hc[num].errCnt = 0;
			pdev->host.hc[num].pURB->state = URB_ERROR;
		}
		else if(hcchar.b.eptype == EP_TYPE_INTR)
		{
//			pdev->host.hc[num].pid ^= 1;
//			pdev->host.hc[num].toggle_in ^= 1;
		}

		if(hcchar.b.eptype != EP_TYPE_INTR)
		{
			int32_t yield = pdFALSE;
			xQueueSendFromISR(pdev->host.hEventQueue, &num, &yield);
			if(yield == pdTRUE)
				taskYIELD();
		}

		CLEAR_HC_INT(hcreg, chhltd);

	}
	else if(hcint.b.xacterr)
	{
		UNMASK_HOST_INT_CHH(num);
		pdev->host.hc[num].errCnt++;
		pdev->host.hc[num].hcStatus = HC_XACTERR;
		USB_OTG_HC_Halt(pdev, num);
		CLEAR_HC_INT(hcreg, xacterr);

	}
	else if(hcint.b.nak)
	{
		pdev->host.hc[num].errCnt++;

		if(hcchar.b.eptype == EP_TYPE_INTR)
		{
			UNMASK_HOST_INT_CHH(num);
			USB_OTG_HC_Halt(pdev, num);
		}
		else if((hcchar.b.eptype == EP_TYPE_CTRL) || (hcchar.b.eptype == EP_TYPE_BULK))
		{
			/* re-activate the channel  */
			hcchar.b.chen = 1;
			hcchar.b.chdis = 0;
			USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[num]->HCCHAR, hcchar.d32);
		}
		pdev->host.hc[num].hcStatus = HC_NAK;
		CLEAR_HC_INT(hcreg, nak);

		if(pdev->host.hc[num].errCnt > 20000)
		{
			UNMASK_HOST_INT_CHH(num);
			USB_OTG_HC_Halt(pdev, num);
		}
	}

	return 1;

}

/**
 * @brief  USB_OTG_USBH_handle_IncompletePeriodicXfer_ISR 
 *         Handles the incomplete Periodic transfer Interrupt
 * @param  pdev: Selected device
 * @retval status 
 *
 * IRQL: USB IRQ
 */
static uint32_t USB_OTG_USBH_handle_IncompletePeriodicXfer_ISR(USB_OTG_CORE_HANDLE *pdev)
{

	USB_OTG_GINTSTS_TypeDef gintsts;
	USB_OTG_HCCHAR_TypeDef hcchar;

	hcchar.d32 = USB_OTG_READ_REG32(&pdev->regs.HC_REGS[0]->HCCHAR);
	hcchar.b.chen = 1;
	hcchar.b.chdis = 1;
	USB_OTG_WRITE_REG32(&pdev->regs.HC_REGS[0]->HCCHAR, hcchar.d32);

	gintsts.d32 = 0;
	/* Clear interrupt */
	gintsts.b.incomplisoout = 1;
	USB_OTG_WRITE_REG32(&pdev->regs.GREGS->GINTSTS, gintsts.d32);

	return 1;
}
