/**
 ******************************************************************************
 * @file    usbh_msc_core.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   This file implements the MSC class driver functions
 *          ===================================================================
 *                                MSC Class  Description
 *          ===================================================================
 *           This module manages the MSC class V1.0 following the "Universal
 *           Serial Bus Mass Storage Class (MSC) Bulk-Only Transport (BOT) Version 1.0
 *           Sep. 31, 1999".
 *           This driver implements the following aspects of the specification:
 *             - Bulk-Only Transport protocol
 *             - Subclass : SCSI transparent command set (ref. SCSI Primary Commands - 3 (SPC-3))
 *
 *  @endverbatim
 *
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/

#include "usbh_msc.h"
#include "usbh_core.h"

#include "usbh_msc.h"

#include "fatfs/ff.h"
#include "util/console.h"

USBH_MSC usbhMsc;

static FATFS fatfs;

#define USB_REQ_BOT_RESET                0xFF
#define USB_REQ_GET_MAX_LUN              0xFE

static USBH_Status USBH_MSC_InterfaceInit(USB_OTG_CORE_HANDLE *pdev, void *phost);
static void USBH_MSC_InterfaceDeInit(USB_OTG_CORE_HANDLE *pdev, void *phost);
static USBH_Status USBH_MSC_Handle(USB_OTG_CORE_HANDLE *pdev, void *phost);
static int16_t USBH_MSC_GETMaxLUN(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost, uint8_t* pMaxLun);
static void USBH_MSC_ErrorHandle(uint8_t status);

/**
 * @brief  USBH_MSC_Init
 *         Initializes the mass storage parameters
 * @param  None
 * @retval None
 */
void USBH_MSC_Init(USB_OTG_CORE_HANDLE *pdev)
{
	if(HCD_IsDeviceConnected(pdev))
	{
		usbhMsc.bot.CBWData.field.CBWSignature = USBH_MSC_BOT_CBW_SIGNATURE;
		usbhMsc.bot.CBWData.field.CBWTag = USBH_MSC_BOT_CBW_TAG;
		usbhMsc.bot.CBWData.field.CBWLUN = 0; /*Only one LUN is supported*/
	}

	usbhMsc.errorCount = 0;

	usbhMsc.callbacks.Init = USBH_MSC_InterfaceInit;
	usbhMsc.callbacks.DeInit = USBH_MSC_InterfaceDeInit;
	usbhMsc.callbacks.Machine = USBH_MSC_Handle;
}

/**
 * @brief  USBH_MSC_InterfaceInit
 *         Interface initialization for MSC class.
 * @param  pdev: Selected device
 * @param  hdev: Selected device property
 * @retval USBH_Status : Status of class request handled.
 */
static USBH_Status USBH_MSC_InterfaceInit(USB_OTG_CORE_HANDLE *pdev, void *phost)
{
	USBH_HOST *pphost = phost;

	usbhMsc.state = USBH_MSC_BOT_INIT_STATE;

	if((pphost->device_prop.Itf_Desc[0].bInterfaceClass == MSC_CLASS)
			&& (pphost->device_prop.Itf_Desc[0].bInterfaceProtocol == MSC_PROTOCOL))
	{
		if(pphost->device_prop.Ep_Desc[0][0].bEndpointAddress & 0x80)
		{
			usbhMsc.setup.bulkInEp = (pphost->device_prop.Ep_Desc[0][0].bEndpointAddress);
			usbhMsc.setup.bulkInEpSize = pphost->device_prop.Ep_Desc[0][0].wMaxPacketSize;
		}
		else
		{
			usbhMsc.setup.bulkOutEp = (pphost->device_prop.Ep_Desc[0][0].bEndpointAddress);
			usbhMsc.setup.bulkOutEpSize = pphost->device_prop.Ep_Desc[0][0].wMaxPacketSize;
		}

		if(pphost->device_prop.Ep_Desc[0][1].bEndpointAddress & 0x80)
		{
			usbhMsc.setup.bulkInEp = (pphost->device_prop.Ep_Desc[0][1].bEndpointAddress);
			usbhMsc.setup.bulkInEpSize = pphost->device_prop.Ep_Desc[0][1].wMaxPacketSize;
		}
		else
		{
			usbhMsc.setup.bulkOutEp = (pphost->device_prop.Ep_Desc[0][1].bEndpointAddress);
			usbhMsc.setup.bulkOutEpSize = pphost->device_prop.Ep_Desc[0][1].wMaxPacketSize;
		}

		usbhMsc.setup.hc_num_out = USBH_Alloc_Channel(pdev, usbhMsc.setup.bulkOutEp);
		usbhMsc.setup.hc_num_in = USBH_Alloc_Channel(pdev, usbhMsc.setup.bulkInEp);

		/* Open the new channels */
		USBH_Open_Channel(pdev, usbhMsc.setup.hc_num_out, pphost->device_prop.address, pphost->device_prop.speed, EP_TYPE_BULK,
				usbhMsc.setup.bulkOutEpSize);

		USBH_Open_Channel(pdev, usbhMsc.setup.hc_num_in, pphost->device_prop.address, pphost->device_prop.speed, EP_TYPE_BULK,
				usbhMsc.setup.bulkInEpSize);

	}
	else
	{
		pphost->usr_cb->DeviceNotSupported();
	}

	return USBH_OK;

}

/**
 * @brief  USBH_MSC_InterfaceDeInit
 *         De-Initialize interface by freeing host channels allocated to interface
 * @param  pdev: Selected device
 * @param  hdev: Selected device property
 * @retval None
 */
void USBH_MSC_InterfaceDeInit(USB_OTG_CORE_HANDLE *pdev, void *phost)
{
	(void)phost;

	f_mount(0, 0);

	if(usbhMsc.setup.hc_num_out)
	{
		USB_OTG_HC_Halt(pdev, usbhMsc.setup.hc_num_out);
		USBH_Free_Channel(pdev, usbhMsc.setup.hc_num_out);
		usbhMsc.setup.hc_num_out = 0; /* Reset the Channel as Free */
	}

	if(usbhMsc.setup.hc_num_in)
	{
		USB_OTG_HC_Halt(pdev, usbhMsc.setup.hc_num_in);
		USBH_Free_Channel(pdev, usbhMsc.setup.hc_num_in);
		usbhMsc.setup.hc_num_in = 0; /* Reset the Channel as Free */
	}
}

/**
 * @brief  USBH_MSC_Handle
 *         MSC state machine handler
 * @param  pdev: Selected device
 * @param  hdev: Selected device property
 * @retval USBH_Status
 */

static USBH_Status USBH_MSC_Handle(USB_OTG_CORE_HANDLE *pdev, void *phost)
{
	USBH_Status status = USBH_OK;
	uint8_t mscStatus = USBH_MSC_BUSY;

	static uint8_t maxLunExceed = FALSE;

	if(HCD_IsDeviceConnected(pdev))
	{
		switch(usbhMsc.state)
		{
			case USBH_MSC_BOT_INIT_STATE:
				USBH_MSC_Init(pdev);
				usbhMsc.state = USBH_MSC_BOT_RESET;
				break;

			case USBH_MSC_BOT_RESET:
				/* Issue BOT RESET request */
				status = USBH_MSC_BOTReset(pdev, phost);
				if(status == USBH_OK)
				{
					usbhMsc.state = USBH_MSC_GET_MAX_LUN;
				}

				if(status == USBH_NOT_SUPPORTED)
				{
					/* If the Command has failed, then we need to move to Next State, after
					 STALL condition is cleared by Control-Transfer */
					usbhMsc.stateBkp = USBH_MSC_GET_MAX_LUN;

					/* a Clear Feature should be issued here */
					usbhMsc.state = USBH_MSC_CTRL_ERROR_STATE;
				}
				break;

			case USBH_MSC_GET_MAX_LUN:
				/* Issue GetMaxLUN request */
				status = USBH_MSC_GETMaxLUN(pdev, phost, usbhMsc.setup.buff);

				if(status == USBH_OK)
				{
					usbhMsc.setup.maxLun = *(usbhMsc.setup.buff);

					/* If device has more that one logical unit then it is not supported */
					if((usbhMsc.setup.maxLun > 0) && (maxLunExceed == FALSE))
					{
						maxLunExceed = TRUE;
						status = USBH_NOT_SUPPORTED;

						break;
					}
					usbhMsc.state = USBH_MSC_TEST_UNIT_READY;
				}

				if(status == USBH_NOT_SUPPORTED)
				{
					/* If the Command has failed, then we need to move to Next State, after
					 STALL condition is cleared by Control-Transfer */
					usbhMsc.stateBkp = USBH_MSC_TEST_UNIT_READY;

					/* a Clear Feature should be issued here */
					usbhMsc.state = USBH_MSC_CTRL_ERROR_STATE;
				}
				break;

			case USBH_MSC_CTRL_ERROR_STATE:
				/* Issue Clearfeature request */
				status = USBH_ClrFeature(pdev, phost, 0x00);
				if(status == USBH_OK)
				{
					/* If GetMaxLun Request not support, assume Single LUN configuration */
					usbhMsc.setup.maxLun = 0;

					usbhMsc.state = usbhMsc.stateBkp;
				}
				break;

			case USBH_MSC_TEST_UNIT_READY:
				/* Issue SCSI command TestUnitReady */
				mscStatus = USBH_MSC_TestUnitReady(pdev, phost);

				if(mscStatus == USBH_MSC_OK)
				{
					usbhMsc.state = USBH_MSC_READ_CAPACITY10;
					usbhMsc.errorCount = 0;
					status = USBH_OK;
				}
				else
				{
					USBH_MSC_ErrorHandle(mscStatus);
				}
				break;

			case USBH_MSC_READ_CAPACITY10:
				/* Issue READ_CAPACITY10 SCSI command */
				mscStatus = USBH_MSC_ReadCapacity10(pdev, phost);
				if(mscStatus == USBH_MSC_OK)
				{
					usbhMsc.state = USBH_MSC_MODE_SENSE6;
					usbhMsc.errorCount = 0;
					status = USBH_OK;
				}
				else
				{
					USBH_MSC_ErrorHandle(mscStatus);
				}
				break;

			case USBH_MSC_MODE_SENSE6:
				/* Issue ModeSense6 SCSI command for detecting if device is write-protected */
				mscStatus = USBH_MSC_ModeSense6(pdev, phost);
				if(mscStatus == USBH_MSC_OK)
				{
					usbhMsc.state = USBH_MSC_INIT_FILESYSTEM;
					usbhMsc.errorCount = 0;
					status = USBH_OK;
				}
				else
				{
					USBH_MSC_ErrorHandle(mscStatus);
				}
				break;

			case USBH_MSC_REQUEST_SENSE:
				/* Issue RequestSense SCSI command for retreiving error code */
				mscStatus = USBH_MSC_RequestSense(pdev, phost);
				if(mscStatus == USBH_MSC_OK)
				{
					usbhMsc.state = usbhMsc.stateBkp;
					status = USBH_OK;
				}
				else
				{
					USBH_MSC_ErrorHandle(mscStatus);
				}
				break;

			case USBH_MSC_BOT_USB_TRANSFERS:
				/* Process the BOT state machine */
//				USBH_MSC_HandleBOTXfer(pdev, phost);
				break;

			case USBH_MSC_INIT_FILESYSTEM:
			{
				if(f_mount(0, &fatfs) != FR_OK)
				{
					ConsolePrint("> Cannot initialize File System.\r\n");
					usbhMsc.state = USBH_MSC_UNRECOVERED_STATE;
					return 0;
				}

				ConsolePrint("> File System initialized.\r\n");
				ConsolePrint("> Disk capacity : %dMB\r\n", (usbhMsc.storageParams.capacity / (1024 / usbhMsc.storageParams.pageLength)) / 1024);

				if(usbhMsc.storageParams.writeProtect)
				{
					ConsolePrint("> Disk is write-protected!\r\n");
				}

				// Get drive information and free clusters
				uint32_t fre_clust, tot_kbytes, fre_kbytes;
				FATFS *myfsptr = &fatfs; //this is necessary because f_getfree expects a pointer to a pointer to a file system object.
				FRESULT fsresult = f_getfree("/", &fre_clust, &myfsptr);
				if(fsresult == FR_OK)
				{
					// Get total bytes and free bytes
					tot_kbytes = ((fatfs.n_fatent - 2) * fatfs.csize) / (1024 / _MAX_SS);
					fre_kbytes = (fre_clust * fatfs.csize) / (1024 / _MAX_SS);
					// Print free space in unit of KB
					ConsolePrint("Free:  %8lu KB\r\n", fre_kbytes);
					ConsolePrint("Total: %8lu KB\r\n", tot_kbytes);
				}
				else
				{
					ConsolePrint("f_getfree failed. FRESULT Error code: %d.  See FATfs/ff.h for FRESULT code meaning.\r\n", fsresult);
					usbhMsc.state = USBH_MSC_UNRECOVERED_STATE;
					return 0;
				}

				// Determine FAT sub-type
				BYTE fssubtype = 42;
				switch(fatfs.fs_type)
				{
					case 1:
						fssubtype = 12;
						break;
					case 2:
						fssubtype = 16;
						break;
					case 3:
						fssubtype = 32;
						break;
					case 0:
						fssubtype = 0;
						break;
				}

				ConsolePrint("Type:   FAT%hhd\r\n", fssubtype);

				usbhMsc.state = USBH_MSC_IDLE;
			}
			break;

			case USBH_MSC_IDLE:
				break;

			case USBH_MSC_UNRECOVERED_STATE:

				status = USBH_UNRECOVERED_ERROR;

				break;

			default:
				break;

		}
	}
	return status;
}

/**
 * @brief  USBH_MSC_GETMaxLUN
 *         This request is used to reset the mass storage device and its
 *         associated interface. This class-specific request shall ready the
 *         device for the next CBW from the host.
 * @param  pdev: Selected device
 * @retval USBH_Status : USB ctl xfer status
 */
static int16_t USBH_MSC_GETMaxLUN(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost, uint8_t* pMaxLun)
{
	phost->Control.setup.b.bmRequestType = USB_D2H | USB_REQ_TYPE_CLASS | USB_REQ_RECIPIENT_INTERFACE;

	phost->Control.setup.b.bRequest = USB_REQ_GET_MAX_LUN;
	phost->Control.setup.b.wValue.w = 0;
	phost->Control.setup.b.wIndex.w = 0;
	phost->Control.setup.b.wLength.w = 1;

	return USBHCtrlReq(pdev, phost, &phost->Control.setup, pMaxLun);
}

/**
 * @brief  USBH_MSC_ErrorHandle
 *         The function is for handling errors occuring during the MSC
 *         state machine
 * @param  status
 * @retval None
 */

void USBH_MSC_ErrorHandle(uint8_t status)
{
	if(status == USBH_MSC_FAIL)
	{
		usbhMsc.errorCount++;
		if(usbhMsc.errorCount < USBH_MSC_ERROR_RETRY_LIMIT)
		{ /* Try MSC level error recovery, Issue the request Sense to get
		 Drive error reason  */
			usbhMsc.state = USBH_MSC_REQUEST_SENSE;
		}
		else
		{
			/* Error trials exceeded the limit, go to unrecovered state */
			usbhMsc.state = USBH_MSC_UNRECOVERED_STATE;
		}
	}
	else if(status == USBH_MSC_PHASE_ERROR)
	{
		/* Phase error, Go to Unrecoovered state */
		usbhMsc.state = USBH_MSC_UNRECOVERED_STATE;
	}
	else if(status == USBH_MSC_BUSY)
	{
		/*No change in state*/
	}
}
