/**
 ******************************************************************************
 * @file    usbh_usr.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   This file includes the usb host library user callbacks
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include <string.h>
#include "usbh_usr.h"
#include "fatfs/ff.h"       /* FATFS */
#include "usb.h"
#include "util/console.h"

#include "FreeRTOS.h"
#include "task.h"

#include "usbh_hid_core.h"

/*  Points to the DEVICE_PROP structure of current device */
/*  The purpose of this register is to speed up the execution */
USBH_Usr_cb_TypeDef USR_cb =
{	USBH_USR_Init,
	USBH_USR_DeInit,
	USBH_USR_DeviceAttached,
	USBH_USR_DeviceDisconnected,
	USBH_USR_OverCurrentDetected,
	USBH_USR_EnumerationDone,
	USBH_USR_ClassSetup,
	USBH_USR_DeviceNotSupported,
};

/**
 * @brief  USBH_USR_Init
 *         Displays the message on LCD for host lib initialization
 * @param  None
 * @retval None
 */
void USBH_USR_Init(void)
{
	ConsolePrint("> USB Host library started.\r\n");
}

/**
 * @brief  USBH_USR_DeviceAttached
 *         Displays the message on LCD on device attached
 * @param  None
 * @retval None
 */
void USBH_USR_DeviceAttached(void)
{
	ConsolePrint("> Device Attached\r\n");

	vTaskDelay(500/portTICK_RATE_MS);
}

/**
 * @brief  USBH_DisconnectEvent
 *         Device disconnect event
 * @param  None
 * @retval Staus
 */
void USBH_USR_DeviceDisconnected(void)
{
	ConsolePrint("> Device Disconnected\r\n");
}

/**
 * @brief  EnumerationDone
 *         User response request is displayed to ask application jump to class
 * @param  None
 * @retval None
 */
void USBH_USR_EnumerationDone(USB_OTG_CORE_HANDLE *pDev, USBH_HOST *pHost)
{
	(void)pDev;

	ConsolePrint("> Enumeration completed\r\n");

	switch(pHost->device_prop.speed)
	{
		case HPRT0_PRTSPD_HIGH_SPEED:
			ConsolePrint("> High speed device detected\r\n");
			break;
		case HPRT0_PRTSPD_FULL_SPEED:
			ConsolePrint("> Full speed device detected\r\n");
			break;
		case HPRT0_PRTSPD_LOW_SPEED:
			ConsolePrint("> Low speed device detected\r\n");
			break;
		default: break;
	}

	USBH_DevDesc_TypeDef *hs = &pHost->device_prop.Dev_Desc;
	ConsolePrint("--- DEVICE ---\r\n");
	ConsolePrint("VID :             0x%04hX\r\n", hs->idVendor);
	ConsolePrint("PID :             0x%04hX\r\n", hs->idProduct);
	ConsolePrint("Device Class:     0x%02hhX\r\n", hs->bDeviceClass);
	ConsolePrint("Device Sub Class: 0x%02hhX\r\n", hs->bDeviceSubClass);
	ConsolePrint("Device Protocol:  0x%02hhX\r\n", hs->bDeviceProtocol);
	ConsolePrint("Num Configs:      0x%02hhX\r\n", hs->bNumConfigurations);

	ConsolePrint("Manufacturer : %s\r\n", (char *)pHost->device_prop.manufacturerString);
	ConsolePrint("Product : %s\r\n", (char *)pHost->device_prop.productString);
	ConsolePrint("Serial Number : %s\r\n", (char *)pHost->device_prop.serialNumberString);

	vTaskDelay(50/portTICK_RATE_MS);

	USBH_CfgDesc_TypeDef * cfgDesc = &pHost->device_prop.Cfg_Desc;
	USBH_InterfaceDesc_TypeDef* itfDesc = pHost->device_prop.Itf_Desc;
	USBH_EpDesc_TypeDef* epDesc = pHost->device_prop.Ep_Desc[0];

	ConsolePrint("--- CFG ---\r\n");
	ConsolePrint("Num Interfaces: %hhd\r\n", cfgDesc->bNumInterfaces);
	ConsolePrint("Max Power:      %hdmA\r\n", cfgDesc->bMaxPower*2);

	for(uint8_t i = 0; i < cfgDesc->bNumInterfaces; i++)
	{
		ConsolePrint("--- INTERFACE ---\r\n");
		ConsolePrint("Class:         0x%02hhX\r\n", (itfDesc+i)->bInterfaceClass);
		ConsolePrint("Sub Class:     0x%02hhX\r\n", (itfDesc+i)->bInterfaceSubClass);
		ConsolePrint("Protocol:      0x%02hhX\r\n", (itfDesc+i)->bInterfaceProtocol);
		ConsolePrint("Num Endpoints: %hhd\r\n", (itfDesc+i)->bNumEndpoints);

		for(uint8_t j = 0; j < (itfDesc+i)->bNumEndpoints; j++)
		{
			ConsolePrint("--- ENDPOINT ---\r\n");
			ConsolePrint("Address:  0x%02hhX\r\n", (epDesc + i*USBH_MAX_NUM_INTERFACES+j)->bEndpointAddress);
			switch((epDesc + i*USBH_MAX_NUM_INTERFACES+j)->bmAttributes & EP_TYPE_MSK)
			{
				case EP_TYPE_CTRL: ConsolePrint("Type:     CTRL\r\n"); break;
				case EP_TYPE_BULK: ConsolePrint("Type:     BULK\r\n"); break;
				case EP_TYPE_ISOC: ConsolePrint("Type:     ISOC\r\n"); break;
				case EP_TYPE_INTR: ConsolePrint("Type:     INTR\r\n"); break;
			}
			ConsolePrint("Interval: %hhd\r\n", (epDesc + i*USBH_MAX_NUM_INTERFACES+j)->bInterval);

			vTaskDelay(50/portTICK_RATE_MS);
		}
	}

}

/**
 * @brief  USBH_USR_DeviceNotSupported
 *         Device is not supported
 * @param  None
 * @retval None
 */
void USBH_USR_DeviceNotSupported(void)
{
	ConsolePrint("> Device not supported.\r\n");
}

/**
 * @brief  USBH_USR_UserInput
 *         User Action for application state entry
 * @param  None
 * @retval USBH_USR_Status : User response for key button
 */
USBH_Class_cb_TypeDef* USBH_USR_ClassSetup(USB_OTG_CORE_HANDLE *pDev, struct _Host_TypeDef* pHost)
{
	(void)pDev;

	USBH_InterfaceDesc_TypeDef *id = pHost->device_prop.Itf_Desc;

	if(id->bInterfaceClass == 0x08)
	{
		ConsolePrint("> Mass storage device connected\r\n");

		return &usbhMsc.callbacks;
	}
	else if(id->bInterfaceClass == 0x03)
	{
		ConsolePrint("> HID device connected\r\n");

		return &HID_cb;
	}

	return 0;
}

/**
 * @brief  USBH_USR_OverCurrentDetected
 *         Over Current Detected on VBUS
 * @param  None
 * @retval Staus
 */
void USBH_USR_OverCurrentDetected(void)
{
	ConsolePrint("> Overcurrent detected.\r\n");
}

/**
 * @brief  USBH_USR_DeInit
 *         Deint User state and associated variables
 * @param  None
 * @retval None
 */
void USBH_USR_DeInit(void)
{
	ConsolePrint("> USB DeInit\r\n");
}
